# This file is part of libdrt.
#
# Written 2011-2014 by Martin Stigge, Pontus Ekberg.
#
# GPLv3 applies.

from __future__ import absolute_import

from drt.analysis.rbf import RbfEngine

from drt.util.statistics import Timer, stats

from drt.util.logging import ERROR, INFO, DEBUG, DEBUG2, DEBUG3, DEBUG4
from drt.util.logging import log, indent_log

# TODO: Used for general caching. Basically a hack. Make nicer.
GLOBAL_CACHE = dict()

class MBPEngine(object):
    """Computes a bound on the maximal busy period for a DRT task set. Uses an
    RbfEngine which computes the request-bound function essentially in the same
    incremental way as the demand-bound function is computed for feasibility in
    the RTAS'11 paper. Same framework is used.

    After creating the object, call the get_max_busy_period() method.
    """

    def __init__(self, taskset, increments=None,
            path_class = None,
            store_class = None):

        self._taskset = taskset
        self._increments = increments
        self._path_class = path_class
        self._store_class = store_class
        self.reset()

        if self._taskset in GLOBAL_CACHE:
            self._mbp = GLOBAL_CACHE[self._taskset]

    def reset(self):
        self._mbp = None

    def get_max_busy_period(self):
        """ Returns upper bound on busy window size, based on
            incremental rbf computation."""

        if self._mbp is None:
            timer = Timer()
            log(INFO, 0, ("Computing upper bound of maximal busy window size "+
                "for taskset '%s'.") % str(self._taskset))
            tasks = self._taskset.get_tasks()

            rbfengines = dict((task, RbfEngine(task, self._increments,
                    path_class = self._path_class,
                    store_class = self._store_class))
                for task in tasks)

            log(INFO, 1, 'Preparing rbf of task set...')
            indent_log(2)
            for task in tasks:
                rbfengines[task].prepare()
            indent_log(-2)
            log(INFO, 2, 'Done preparing rbf for task set.')

            t = 1
            while True:
                rbfts = [rbfengines[task].get_rbf(t-1) for task in tasks]
                rbft = sum(rbfts)
                log(DEBUG, 2, "Computed rbf(%d)=%d." % (t, rbft))
                if (rbft <= t):
                    log(INFO, 2, 'Done.')
                    log(DEBUG, 2, 'rbf(%d) = %d' % (t, rbft))
                    log(DEBUG2, 3, '= sum(%s)' % (rbfts))
                    break
                else:
                    log(DEBUG3, 2, 'rbf(%d) = %d' % (t, rbft))
                    log(DEBUG4, 3, '= sum(%s)' % (rbfts))
                    t = rbft

            log(INFO, 0, "Busy window size %d computed within %.2f s." % 
                (t, timer.stop()))

            self._mbp = t
            GLOBAL_CACHE[self._taskset] = t

        else:
            log(DEBUG, 0, "Using cached maximal busy window size %d." %
                self._mbp)

        return self._mbp
 
  

# vim:textwidth=79:shiftwidth=4:tabstop=4:cc=80:expandtab:autoindent

