# This file is part of libdrt.
#
# Written 2011-2014 by Martin Stigge, Pontus Ekberg.
#
# GPLv3 applies.

from __future__ import absolute_import

from drt.analysis.util.dbfpaths import DemandTriple
from drt.analysis.util.pathstores import DiscardingSortedDbfStore
from drt.analysis.util.pathstores import DiscardingSortedDbfStoreNoD
from drt.analysis.util.explorationengines import *

from drt.util.statistics import Timer, stats

from drt.util.logging import ERROR, INFO, DEBUG, DEBUG2, DEBUG3, log, indent_log


class DbfEngine(object):
    """Computes the demand-bound function of a task using the method from
    the RTAS'11 paper.
    
    For using an instance, first use the prepare() method to set up the engine.
    Without increments, the prepare() call will compute the dbf up to D.
    After that, retrieve the dbf point-wise using get_dbf().
    Random access is possible.

    If the engine is initialized with an increments not None, the engine will
    compute the dbf incrementally. The initial call to prepare() only computes
    the prefix of size given in increments. Future get_dbf() calls
    incrementally extend the known prefix in chunks. This allows a very
    efficient feasibility test if counterexamples exist early.

    Custom store_class and path_class can be provided which default
    to the most optimized variants available for constrained deadlines.

    CAUTION For arbitrary deadlines, this default (demand triple abstraction)
    is not safe. Use the DemandQuadruples path class instead.
    """


    def __init__(self, task, D, increments=None,
            store_class = None,
            path_class = None):
        self._task = task
        self._D = D
        self._increments = increments
        self._store_class = store_class
        self._path_class = path_class
        self._dbf = []
        self._witnesses = []
        if self._store_class is None:
            if increments is None:
                self._store_class = DiscardingSortedDbfStore
            else:
                self._store_class = DiscardingSortedDbfStoreNoD
        if self._path_class is None:
            self._path_class = DemandTriple

    def prepare(self):
        if self._increments is None:
            log(INFO, 0, 'Computing dbf for task %s...' % self._task)
            log(DEBUG, 1, 'D:           %d' % self._D)
            log(DEBUG, 1, 'Path class:  %s' % self._path_class.__name__)
            log(DEBUG, 1, 'Store class: %s' % self._store_class.__name__)

            # Explore task graph.
            indent_log(2)
            store = self._store_class(self._D)
            ee = ExplorationEngine(self._task, store, self._path_class)
            ee.run()
            indent_log(-2)

            # Generate dbf.
            (self._dbf, self._witnesses) = paths_to_dbf(store.get_paths(), 
                self._D)

        else:
            log(INFO, 0, ("Initialising incremental dbf computation " +
                "for task %s...") % self._task)
            log(DEBUG, 1, 'Path class:  %s' % self._path_class.__name__)
            log(DEBUG, 1, 'Store class: %s' % self._store_class.__name__)

            # Explore task graph.
            indent_log(2)
            # This expects a store without D
            store = self._store_class()
            ee = IncrementalExplorationEngine(self._task, store, 
                self._path_class)
            eerunner = ee.run(self._increments)
            indent_log(-2)
            log(DEBUG, 1, 'Init done.')

            self._dbf = []
            self._store = store
            self._eerunner = eerunner

    def get_dbf(self, t):
        """Return dbf(t) for this task. Method prepare() needs to be
        called beforehand. 
        
        For non-incremental computation, all paths have been explored there
        already.

        For incremental computation, we return the value if graph exploration
        has been run sufficiently long. Otherwise we run the engine until 
        the value is clear.
        """

        if len(self._dbf) <= t:
            # If we are not running incremental, there is some serious 
            # problem here.
            if self._increments is None:
                raise RuntimeError("Dbf not prepared in non-incremental test.")

            store = self._store
            eerunner = self._eerunner

            while store.current_limit < t:
                log(DEBUG, 2, ("Incremental dbf computation for task '%s' " +
                    "from %d...") % (self._task, store.current_limit))
                eerunner.next()

            (self._dbf, self._witnesses) = paths_to_dbf(store.get_paths(), 
                    store.current_limit)

        # Now we know that self.dbf is defined at t
        return self._dbf[t]


def paths_to_dbf(paths, D):
    log(DEBUG, 1, 'Generating dbf from paths...')
    dbf = [0] * (D + 1)
    witnesses = [None] * (D + 1) # Witness paths.
    for path in paths:
        e = path.get_e()
        d = path.get_d()
        t = d
        while t <= D and dbf[t] < e:
            dbf[t] = e
            witnesses[t] = path
            t += 1
    return (dbf, witnesses)


    

# vim:textwidth=79:shiftwidth=4:tabstop=4:cc=80:expandtab:autoindent

