# This file is part of libdrt.
#
# Written 2011-2014 by Martin Stigge, Pontus Ekberg.
#
# GPLv3 applies.

"""
Path abstractions for utilization computation.
"""

from drt.util.logging import ERROR, INFO, DEBUG, DEBUG2, DEBUG3, log, indent_log
from drt.analysis.util.paths import Path

class UtilizationPath(Path):
    """Template for path abstraction classes used to compute utilization."""

    def get_p(self):
        """Return the total duration of the path (i.e., the minimal time
        required to spawn all jobs in it).
        """
        raise NotImplementedError

    def get_e2(self):
        """Return total execution requirement of abstracted path
        without last node. (In the end, this is interesting for cycles!)
        """ 
        raise NotImplementedError

    def get_start(self):
        """Return the first job in the path."""
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError


class FullUtilizationPath(UtilizationPath):

    """Stores entire concrete paths (i.e., it performs no
    abstractions at all). Domination relation for utilization computations.
    """

    def get_p(self):
        if len(self.jobs) == 0:
            return 0
        else:
            task = self.jobs[0].get_task()
            edges = zip(self.jobs[:-1], self.jobs[1:])
            return sum(task.get_delay(edge) for edge in edges)

    def get_e2(self):
        return sum([job.get_e() for job in self.jobs[:-1]])

    def get_start(self):
        return self.jobs[0]

    def get_end(self):
        return self.jobs[-1]

    def get_extended(self, job):
        return [FullUtilizationPath(self.jobs[:] + [job])]

    def __len__(self):
        return len(self.jobs) - 1

    def __ge__(self, o):
        return (self.get_start() == o.get_start()
            and self.get_end() == o.get_end() 
            and self.get_e2() * o.get_p() >= o.get_e2() * self.get_p())
			
    def __gt__(self, o):
        return (self >= o 
            and (self.get_e2() * o.get_p() > o.get_e2() * self.get_p()))

    def __str__(self):
        return '[' + ' -> '.join(map(str, self.jobs)) + ']'


class UtilizationTriple(UtilizationPath):

    """UtilizationTriple stores paths abstracted as utilization triples, i.e.,
    as triples <e, p, (u, v)> of execution time (minus that of the first
    vertex, i.e., "e2"), duration and first and last vertices.
    """

    def __init__(self, jobs, e2=None, p=None, length=None):
        if e2 is None:
            self.e2 = sum([job.get_e() for job in jobs[:-1]])
        else:
            self.e2 = e2

        if p is None:
            task = jobs[0].get_task()
            edges = zip(jobs[:-1], jobs[1:])
            self.p = sum(task.get_delay(edge) for edge in edges)
        else:
            self.p = p
        
        self.first = jobs[0]
        self.last = jobs[-1]

        if length is None:
            self.length = len(jobs) - 1
        else:
            self.length = length

    def get_jobs(self):
        raise RuntimeError("Abstract path does not have jobs.")

    def get_p(self):
        return self.p

    def get_e2(self):
        return self.e2

    def get_start(self):
        return self.first

    def get_end(self):
        return self.last

    def get_extended(self, job):
        # TODO: Should this really be a method of a path?
        task = job.get_task()
        # Careful here: Don't add job.get_e(), but self.last.get_e() instead.
        e2 = self.e2 + self.last.get_e()
        p = self.p + task.get_delay((self.last, job))
        return [UtilizationTriple([self.first, job], e2, p, self.length + 1)]

    def __len__(self):
        return self.length

    def __ge__(self, o):
        return (self.get_start() == o.get_start()
            and self.get_end() == o.get_end() 
            and self.get_e2() * o.get_p() >= o.get_e2() * self.get_p())
			
    def __gt__(self, o):
        return (self >= o 
            and (self.get_e2() * o.get_p() > o.get_e2() * self.get_p()))

    def __str__(self):
        return '<%d, %d, (%s, %s)>' % (self.e2, self.p, self.first, self.last)

 


# vim:textwidth=79:shiftwidth=4:tabstop=4:cc=80:expandtab:autoindent

