# This file is part of libdrt.
#
# Written 2011-2014 by Martin Stigge, Pontus Ekberg.
#
# GPLv3 applies.

class Job(object):

    def __init__(self, e, d, task=None, name='J'):
        self._e = e
        self._d = d
        self._task = task
        self._name = name

    def __str__(self):
        return "%s(%d, %d)" % (str(self._name), self._e, self._d)

    def get_e(self):
        return self._e

    def get_d(self):
        return self._d

    def get_task(self):
        return self._task

    def set_task(self, task):
        self._task = task

    def dominates(self, j2, strict=False):
        # Let's not get into the business of overloading < and <=
        # or rely on default behavior
        if strict:
            return (((self._e, self._d) != (j2.get_e(), j2.get_d())) 
                and self.dominates(j2))
        else:
            return (self._e >= j2.get_e()) and (self._d <= j2.get_d())


class DRTTask(object):
    """A DRT task as defined in the original RTAS11 paper."""

    def __init__(self, jobs, edges=None, name='T'):
        """
        Args:
            jobs: Set of jobs.
            edges: Dictionary, for each pair (u, v) the edge label.
            name: String used for output.

        Will set the task of all jobs to the new object.
        """

        self._jobs = set(jobs)
        self._name = name
        if edges is not None:
            self._edges = edges
        else:
            self._edges = dict() 
        for job in self._jobs:
            job.set_task(self)

    def __str__(self):
        return str(self._name)

    def __len__(self):
        return len(self._jobs)

    def add_job(self, job):
        self._jobs.add(job)

    def remove_job(self, job):
        remove_edges = set(e for e in self._edges 
            if e[0] == job or e[1] == job)
        for edge in remove_edges:
            self.remove_edge(edge)
        if job in self._jobs:
            self._jobs.remove(job)


    def add_edge(self, edge, delay):
        if edge in self._edges:
            # We don't allow multiple edges for each pair; take minimum.
            self._edges[edge] = min(delay, self._edges[edge])
        else:
            self._edges[edge] = delay

    def remove_edge(self, edge):
        self._edges.pop(edge)

    def get_jobs(self):
        """Returns the set of jobs.

        WARNING: This is the actual object. If you change it afterwords,
        you will be messing up the task object."""
        return self._jobs

    def get_edges(self):
        """Returns the set of edges."""
        return self._edges.keys()

    def get_delay(self, edge):
        return self._edges[edge]

    def get_mindelay(self, job):
        edges = [(j1, j2) for (j1, j2) in self._edges.keys() if j1 == job]
        if len(edges) > 0:
            return min([self._edges[e] for e in edges])
        else:
            return job.get_d() # added by Morteza to prevent error when the job has no successor
            return float('Inf')

    def get_minprevdelay(self, job):
        edges = [(j1, j2) for (j1, j2) in self._edges.keys() if j2 == job]
        if len(edges) > 0:
            return min([self._edges[e] for e in edges])
        else:
            return float('Inf')

    def get_preds(self, job):
        """Given a job, return the set of predecessor jobs."""
        preds = set()
        for (j1, j2) in self._edges.keys():
            if j2 == job:
                preds.add(j1)
        return preds

    def get_succs(self, job):
        """Given a job, return the set of successor jobs."""
        succs = set()
        for (j1, j2) in self._edges.keys():
            if j1 == job:
                succs.add(j2)
        return succs


class TaskSet(object):

    def __init__(self, tasks=None, name='TS'):
        self._tasks = set(tasks) or set()
        self._name = name

    def __str__(self):
        return str(self._name)

    def __len__(self):
        return len(self._tasks)

    def get_tasks(self):
        return self._tasks

    def add_task(self, task):
        self._tasks.add(task)

    def remove_task(self, task):
        self._tasks.remove(task)


class SPTaskSet(TaskSet):

    def __init__(self, tasks=None, priolist=None, name='TS'):
        """
        Args:
            tasks: Either a set of tasks as usual, or a list, which
                is shortcut notation for replacing the priolist parameter,
                mainly for backwards compatibility.
            priolist: A list of tasks in priority order. Lower index means
                higher priority. Must contain exactly the same tasks
                as tasks parameter.
        """
        if isinstance(tasks, list) and priolist is None:
            self._priolist = tasks
        else:
            self._priolist = priolist
        return super(SPTaskSet, self).__init__(
            tasks=set(tasks), name=name)

    def get_priolist(self):
        return self._priolist

    def add_task(self, task, index=None):
        if index is None:
            self._priolist.append(task)
        else:
            self._priolist.insert(index, task)
        return super(SPTaskSet, self).add_task(task)

    def remove_task(self, task):
        self._priolist.remove(task)
        return super(SPTaskSet, self).remove_task(task)


class SJPTaskSet(TaskSet):

    def __init__(self, tasks, priolist=None, name='TS'):
        """
        Args:
            tasks: A set of tasks 
            priolist: A list of jobs in priority order. Lower index means
                higher priority. Must contain exactly the same jobs
                as all tasks contain.
                If None, only SJP feasibility can be checked.
        """
        self._priolist = priolist
        return super(SJPTaskSet, self).__init__(
            tasks=tasks, name=name)

    def get_priolist(self):
        return self._priolist

    def add_task(self, task):
        # TODO: When implementing this, provide some way to specify priolist.
        raise RuntimeError("Not implemented.")



# vim:textwidth=79:shiftwidth=4:tabstop=4:cc=80:expandtab:autoindent
