# This file is part of libdrt.
#
# Written 2014 by Martin Stigge.
#
# GPLv3 applies.

import unittest

import drt.util.logging
import drt.model
import drt.examples
import drt.analysis.utilization

class Test_DRT_Model(unittest.TestCase):

    def test_task_creation(self):
        v1 = drt.model.Job(2, 3, name="J1")
        v2 = drt.model.Job(2, 4, name="J2")
        v3 = drt.model.Job(2, 3, name="J3")
        edges = dict()
        edges[(v1, v2)] = 10
        edges[(v2, v1)] = 20
        edges[(v2, v3)] = 30
        T = drt.model.DRTTask([v1, v2, v3], edges, "T")

        self.assertEqual(v1.get_e(), 2)
        self.assertEqual(v1.get_d(), 3)
        self.assertEqual(v1.get_task(), T)

        self.assertTrue(v1.dominates(v2))
        self.assertTrue(v1.dominates(v2, strict=True))
        self.assertTrue(v1.dominates(v3))
        self.assertFalse(v1.dominates(v3, strict=True))
        self.assertFalse(v2.dominates(v1))

        self.assertEqual(len(T), 3)
        self.assertEqual(type(T.get_jobs()), set)

        self.assertTrue(T.get_delay((v1, v2)) == 10)
        self.assertTrue(T.get_mindelay(v2) == 20)
        self.assertTrue(T.get_mindelay(v3) == float('Inf'))
        self.assertTrue(T.get_minprevdelay(v3) == 30)
        self.assertTrue(T.get_preds(v1) == set([v2]))
        self.assertTrue(T.get_succs(v2) == set([v1, v3]))

    def test_task_manipulation(self):
        v1 = drt.model.Job(2, 3, name="J1")
        v2 = drt.model.Job(2, 4, name="J2")
        v3 = drt.model.Job(2, 3, name="J3")
        edges = dict()
        edges[(v1, v2)] = 10
        edges[(v2, v1)] = 20
        T = drt.model.DRTTask([v1, v2], edges, "T")

        self.assertTrue(len(T) == 2)
        self.assertTrue(T.get_delay((v1, v2)) == 10)

        T.add_job(v3)
        T.add_edge((v1, v3), 10)

        self.assertTrue(len(T) == 3)
        self.assertTrue(v3 in T.get_jobs())
        self.assertTrue((v1, v3) in T.get_edges())

        T.remove_edge((v2, v1))
        T.remove_job(v3)

        self.assertTrue((v1, v2) in T.get_edges())
        self.assertTrue((v2, v1) not in T.get_edges())
        self.assertTrue((v1, v3) not in T.get_edges())
        self.assertTrue(v1 in T.get_jobs())
        self.assertTrue(v2 in T.get_jobs())
        self.assertTrue(v3 not in T.get_jobs())
        self.assertTrue(len(T) == 2)

    def test_taskset(self):
        v1 = drt.model.Job(2, 3, name="J1")
        v2 = drt.model.Job(2, 4, name="J2")
        v3 = drt.model.Job(2, 3, name="J3")
        edges = dict()
        edges[(v1, v2)] = 10
        edges[(v2, v1)] = 20
        edges[(v2, v3)] = 30
        T1 = drt.model.DRTTask([v1, v2, v3], edges, "T1")

        T2 = drt.examples.PeriodicTask(1, 10, "T2")

        TS = drt.model.TaskSet([T1], "TS")
        self.assertTrue(len(TS) == 1)

        TS.add_task(T2)
        self.assertTrue(len(TS) == 2)
        self.assertTrue(T2 in TS.get_tasks())

        TS.remove_task(T1)
        self.assertTrue(len(TS) == 1)
        self.assertTrue(T1 not in TS.get_tasks())

        TS2 = drt.model.SPTaskSet([T2, T1])
        self.assertTrue(type(TS2.get_tasks()) == set)
        self.assertTrue(TS2.get_priolist() == [T2, T1])

        TS2.remove_task(T1)
        TS2.add_task(T1, 0)
        self.assertTrue(len(TS2) == 2)
        self.assertTrue(TS2.get_priolist() == [T1, T2])
    
        

if __name__ == "__main__":
    unittest.main()
