# This file is part of libdrt.
#
# Written 2011-2014 by Martin Stigge, Pontus Ekberg.
#
# GPLv3 applies.


#from __future__ import absolute_import

from drt.analysis.util.dbfpaths import DemandTriple
from drt.analysis.util.upaths import UtilizationTriple
from drt.analysis.util.pathstores import DiscardingSortedDbfStore
from drt.analysis.util.pathstores import DiscardingSortedDbfStoreNoD
from drt.analysis.util.pathstores import DiscardingUtilizationStore
from drt.analysis.util.explorationengines import *

from drt.analysis.utilization import UEngine
from drt.analysis.dbf import DbfEngine

from drt.util.statistics import Timer, stats

from drt.util.logging import ERROR, INFO, DEBUG, DEBUG2, DEBUG3, log, indent_log


class EDFSchedulabilityAnalyzer(object):
    """Tests a given taskset for EDF schedulability on a uniprocessor platform.
    This is equivalent to feasibility.

    Method is taken from the RTAS'11 article. First, a pseudo-polynomial bound
    D is computed based on the utilization. Second, the dbf for all tasks is
    computed up to the bound, testing for a t with dbf(t) > t. By default,
    the dbf is computed incrementally via the dbfengine instance.

    In order to get the result, first call the run() method after which the
    result can be retrieved via get_result().

    Custom path and store classes can be provided for both dbf and utilization
    computations. They will be used to initialize dbf and utilization engines.
    """


    def __init__(self, taskset, increments=1000,
            path_class = None,
            store_class = None,
            u_path_class = None,
            u_store_class = None):
        self._taskset = taskset
        self._increments = increments
        self._path_class = path_class
        self._store_class = store_class
        self._u_path_class = u_path_class
        self._u_store_class = u_store_class

        self.reset()
        log(INFO, 0, ("EDF Schedulability Analyzer set up for taskset '%s' " +
            "with increments = %s.") % (str(taskset), str(increments)))

    def reset(self):
        self._result = None
        self._utilizations = dict(zip(self._taskset.get_tasks(),
            [None]*len(self._taskset)))
        self._dbfs = dict(zip(self._taskset.get_tasks(),
            [None]*len(self._taskset)))
        stats[self] = dict()
        self._stats = stats[self]
        self._U = None

    def get_result(self):
        return self._result

    def get_U(self):
        return self._U

    def run(self):
        if self._result is not None:
            return

        timer_total = Timer()
        log(INFO, 0, "Checking EDF schedulability of taskset '%s'." %
            str(self._taskset))
        tasks = self._taskset.get_tasks()

        # Compute task utilizations
        log(INFO, 1, "Computing taskset utilization.")
        timer_u = Timer()
        for task in tasks:
            if self._utilizations[task] is not None:
                log(INFO, 1, "Using cached utilization for task '%s': %d" %
                    (str(task), self._utilizations[task]))
            else:
                indent_log(1)
                ue = UEngine(task,
                        path_class = self._u_path_class,
                        store_class = self._u_store_class)
                ue.run()
                self._utilizations[task] = ue.get_U()
                indent_log(-1)
        U = sum(self._utilizations.values())
        self._U = U
        self._stats["U"] = U

        timer_u.stop()
        log(INFO, 1, "Done computing utilization after %.2fs." %
            timer_u.get_t())
        self._stats["utime"] = timer_u.get_t()

        # Start the actual dbf test.
        if U >= 1:
            feasible = False
        else:
            e_sum = sum(sum(job.get_e() for job in task.get_jobs())
                for task in tasks)
            D = int(e_sum / (1 - U))
            log(INFO, 2, "Taskset's total execution: %d" % e_sum)
            log(INFO, 2, "Taskset's total utilization: %.5f" % U)
            log(INFO, 2, "Bound (D): %d" % D)

            dbfengines = dict((task, DbfEngine(task, D, self._increments,
                    path_class = self._path_class,
                    store_class = self._store_class))
                for task in tasks)

            log(INFO, 1, 'Preparing dbf of task set...')
            indent_log(2)
            for task in tasks:
                dbfengines[task].prepare()
            indent_log(-2)
            log(INFO, 2, 'Done preparing dbf for task set.')

            feasible = True
            for t in xrange(D + 1):
                dbfts = [dbfengines[task].get_dbf(t) for task in tasks]
                dbft = sum(dbfts)
                if (dbft > t) and feasible:
                    log(INFO, 2, '*** COUNTEREXAMPLE! ***')
                    log(DEBUG, 2, 'dbf(%d) = %d' % (t, dbft))
                    log(DEBUG2, 3, '= sum(%s)' % (dbfts))
                    feasible = False
                    break
                else:
                    log(DEBUG2, 2, 'dbf(%d) = %d' % (t, dbft))
                    log(DEBUG3, 3, '= sum(%s)' % (dbfts))

        if feasible:
            log(INFO, 1, ':-) Task set is feasible!')
        else:
            log(INFO, 1, ':-( Task set is not feasible!')        

        timer_total.stop()
        log(INFO, 0, "EDF Schedulability test completed after %.2fs." %
            timer_total.get_t())
        self._stats["totaltime"] = timer_total.get_t()
        self._stats["feasible"] = feasible
        self._stats["result"] = feasible

        self._result = feasible


def is_edf_schedulable(TS):
    """Shortcut for testing the given taskset's feasibility
    (equivalent to edf schedulability).
    """

    A = EDFSchedulabilityAnalyzer(TS)
    A.run()
    return A.get_result()

is_feasible = is_edf_schedulable
               

# vim:textwidth=79:shiftwidth=4:tabstop=4:cc=80:expandtab:autoindent

