# This file is part of libdrt.
#
# Written 2011-2014 by Martin Stigge, Pontus Ekberg.
#
# GPLv3 applies.


from __future__ import absolute_import

from drt.analysis.util.rbfpaths import RequestTriple
from drt.analysis.util.pathstores import DiscardingSortedRbfStoreNoD
from drt.analysis.util.explorationengines import *

from drt.util.statistics import Timer, stats

from drt.util.logging import ERROR, INFO, DEBUG, DEBUG2, DEBUG3, log, indent_log

# TODO: Used for general caching. Basically a hack. Make nicer.
GLOBAL_CACHE = dict()


class RbfEngine(object):
    """ Computes Request-Bound Function incrementally.
    """
    

    def __init__(self, task, increments=None,
            store_class = None,
            path_class = None):
        self._task = task
        if increments is None:
            max_delay = None
            for edge in task.get_edges():
                max_delay = max(max_delay, task.get_delay(edge))
            if max_delay is not None:
                self._increments = int(1.3 * max_delay)  # Heuristic
            else:
                self._increments = 1000
        else:
            self._increments = increments
        if store_class is None:
            self._store_class = DiscardingSortedRbfStoreNoD
        else:
            self._store_class = store_class
        if path_class is None:
            self._path_class = RequestTriple
        else:
            self._path_class = path_class

        self._rbf = []
        self._witnesses = []

        if self._task in GLOBAL_CACHE:
            self._rbf = GLOBAL_CACHE[self._task]

    def prepare(self):
        log(INFO, 0, ("Initializing incremental rbf computation " +
            "for task %s...") % self._task)
        log(DEBUG, 1, 'Path class:  %s' % self._path_class.__name__)
        log(DEBUG, 1, 'Store class: %s' % self._store_class.__name__)

        # Explore task graph.
        indent_log(2)
        # This expects a store without D
        store = self._store_class()
        ee = IncrementalExplorationEngine(self._task, store, 
            self._path_class)
        eerunner = ee.run(self._increments)
        indent_log(-2)
        log(DEBUG, 1, 'Init done.')

        self._rbf = []
        self._store = store
        self._eerunner = eerunner

    def get_rbf(self, t):
        """Return rbf(t) for this task. Method prepare() needs to be
        called beforehand. 
        
        We return the value if graph exploration has been run sufficiently
        long. Otherwise we run the engine until the value is known.

        Note: This is about time slots, i.e., rbf(0)>0.
        (In papers we usually define rbf(0)=0, i.e.,
            rbf_paper(t) = rbf_here(t-1).)
        """

        if len(self._rbf) <= t:

            store = self._store
            eerunner = self._eerunner

            while store.current_limit < t:
                log(DEBUG, 2, ("Incremental rbf computation for task '%s' " +
                    "from %d...") % (self._task, store.current_limit))
                eerunner.next()

            (self._rbf, self._witnesses) = paths_to_rbf(store.get_paths(), 
                    store.current_limit)

            GLOBAL_CACHE[self._task] = self._rbf

        # Now we know that self._rbf is defined at t
        return self._rbf[t]


def paths_to_rbf(paths, D):
    log(DEBUG, 1, 'Generating rbf from paths...')
    rbf = [0] * (D + 1)
    witnesses = [None] * (D + 1) # Witness paths.
    for path in paths:
        e = path.get_e()
        p = path.get_p()
        t = p
        while t <= D and rbf[t] < e:
            rbf[t] = e
            witnesses[t] = path
            t += 1
    return (rbf, witnesses)


    

# vim:textwidth=79:shiftwidth=4:tabstop=4:cc=80:expandtab:autoindent

