# This file is part of libdrt.
#
# Written 2013-2014 by Martin Stigge
# Modified 2015-2016 by Morteza Mohaqeqi
#
# GPLv3 applies.
#
# Modified version: the class SPSchedulabilityAnalyzer was removed
# other parts of the code is unchanged. Only new function get_mrf() added 

from __future__ import absolute_import

import itertools

from drt.analysis.util.explorationengines import *
from drt.analysis.util.rfpaths import FullRfPath
from drt.analysis.util.rfpaths import AbstractRfPath
from drt.analysis.util.rfpaths import DiscardingRfStore

from drt.util.statistics import Timer, stats

from drt.util.logging import ERROR, INFO, DEBUG, DEBUG2, DEBUG3, DEBUG4
from drt.util.logging import log, indent_log

import drt.model

class SPSchedulabilityAnalyzer(object):
    """Tests a given taskset for static priority (SP) schedulability on 
    a uniprocessor platform.

    Method is taken from the RTSS'13 article. For each task, a set of critical
    request functions is derived and organized in an abstraction tree. Each job
    is then tested for interference from tasks of higher priorities represented
    by their (abstract) request functions, taken from the abstraction trees.
    Only critical jobs are considered (unless cj_optimization is set to False).

    In order to get the result, first call the run() method after which the
    result can be retrieved via get_result().
    """


    def __init__(self, taskset, cj_optimization=True):
        """
        Args:
            taskset: Taskset to be tested. Instance of SPTaskSet, i.e., needs
                to provide get_priolist() method with a priority order.
            cj_optimization: Only test critical jobs. Safe optimization.

        Note:
            * Limited to constrainted deadlines.
        """
        self._taskset = taskset
        self._cj_optimization = cj_optimization

        self.reset()
        log(INFO, 0, "SP Schedulability Analyzer set up for taskset '%s'." %
            str(taskset))

    def reset(self):
        self._result = None
        stats[self] = dict()
        self._stats = stats[self]

    def get_result(self):
        return self._result

    def response_time(self, e, d):
        
        if self._cj_optimization:
            log(DEBUG, 1, "Building critical_jobs set.")

            cj = set()
            for t in self._taskset.get_tasks():  # modified by Morteza.get_priolist()
                # Critical jobs in current task
                cj_task = set()

                for job in t.get_jobs():
                    # Remove from cj all jobs that we dominate
                    cj.difference_update([j 
                        for j in cj if job.dominates(j, strict=True)])
                    # Remove from cj_task all jobs that we dominate
                    cj_task.difference_update([j 
                        for j in cj_task if job.dominates(j, strict=True)])
                    # Add ourselves to cj_task if we are not dominated 
                    # by anything in cj_task
                    if all([not j.dominates(job, strict=True) for j in cj_task]):
                        cj_task.add(job)
                cj = cj.union(cj_task)

            # Non-critical jobs; mainly for debugging
            ncj = set()
            ncj_reasons = dict()
            ncj_intra = dict()
            tasks = [t for t in self._taskset.get_tasks()]  # modified by Morteza  .get_priolist()
            for i in range(len(tasks)):
                t = tasks[i]
                for j in t.get_jobs():
                    if j not in cj:
                        ncj.add(j)
                        intra_task_dominators = cj.intersection(t.get_jobs())
                        inter_task_dominators = cj.intersection([j2 
                            for t2 in tasks[i+1:] for j2 in t2.get_jobs()])
                        ncj_reasons[j] = list(intra_task_dominators) + \
                            list(inter_task_dominators)
                        ncj_intra[j] = (len(intra_task_dominators) > 0)

            all_jobs = [j for t in self._taskset.get_tasks() 
                for j in t.get_jobs()]
            log(DEBUG2, 2, "%d of %d (%3.2f%%) jobs are critical." % 
                (len(cj), len(all_jobs), float(len(cj)*100)/len(all_jobs)))
            log(DEBUG2, 2, ("Of %d non-critical jobs, %d are intra-task " +
                "dominated.") % 
                (len(ncj), len([j for j in ncj_intra if ncj_intra[j]])))
            log(DEBUG3, 2, "Critical jobs: %s" % [str(j) for j in cj])
            log(DEBUG3, 2, "Non-critical jobs:")
            for j in ncj:
                log(DEBUG3, 3, "%s, dominated by at least %s." % 
                    (str(j), str(ncj_reasons[j][0])))

            self._stats["alljobs_count"] = len(all_jobs)
            self._stats["ncj_count"] = len(ncj)
            self._stats["ncj_intra_count"] = len([j for j in ncj_intra 
                if ncj_intra[j]])

        else:
            cj = set([j for T in self._taskset.get_tasks() 
                for j in T.get_jobs()])

        D = max([j.get_d() for j in cj])
        # D = max(D, d)   # added by Morteza
        
        indent_log(2)
        # These will be used in all interference testers, so generate them
        # here just once.
        rfpath_generators = dict()
        for task in self._taskset.get_tasks():
            rfpath_generators[task] = RfPathGenerator(task, D)
            rfpath_generators[task].prepare()
        indent_log(-2)
        log(DEBUG, 1, "Done preparing request functions.")

        # Go through priority list in reverse order, since it's most likely
        # to find a deadline violation at the lowest priority.
        # TODO: Some testing whether this is actually true
        indent_log(1)
        tasks = [t for t in self._taskset.get_tasks()]  # modified by Motrteza tasks = self._taskset.get_priolist()
        rspt = response_time(e, D, tasks, rfpath_generators)
        return rspt

    def run(self):
        timer_total = Timer()
        log(INFO, 0, "Checking SP schedulability of taskset '%s'." %
            str(self._taskset))

        if self._cj_optimization:
            log(DEBUG, 1, "Building critical_jobs set.")

            cj = set()
            for t in self._taskset.get_priolist():
                # Critical jobs in current task
                cj_task = set()

                for job in t.get_jobs():
                    # Remove from cj all jobs that we dominate
                    cj.difference_update([j 
                        for j in cj if job.dominates(j, strict=True)])
                    # Remove from cj_task all jobs that we dominate
                    cj_task.difference_update([j 
                        for j in cj_task if job.dominates(j, strict=True)])
                    # Add ourselves to cj_task if we are not dominated 
                    # by anything in cj_task
                    if all([not j.dominates(job, strict=True) for j in cj_task]):
                        cj_task.add(job)
                cj = cj.union(cj_task)

            # Non-critical jobs; mainly for debugging
            ncj = set()
            ncj_reasons = dict()
            ncj_intra = dict()
            tasks = self._taskset.get_priolist()
            for i in range(len(tasks)):
                t = tasks[i]
                for j in t.get_jobs():
                    if j not in cj:
                        ncj.add(j)
                        intra_task_dominators = cj.intersection(t.get_jobs())
                        inter_task_dominators = cj.intersection([j2 
                            for t2 in tasks[i+1:] for j2 in t2.get_jobs()])
                        ncj_reasons[j] = list(intra_task_dominators) + \
                            list(inter_task_dominators)
                        ncj_intra[j] = (len(intra_task_dominators) > 0)

            all_jobs = [j for t in self._taskset.get_tasks() 
                for j in t.get_jobs()]
            log(DEBUG2, 2, "%d of %d (%3.2f%%) jobs are critical." % 
                (len(cj), len(all_jobs), float(len(cj)*100)/len(all_jobs)))
            log(DEBUG2, 2, ("Of %d non-critical jobs, %d are intra-task " +
                "dominated.") % 
                (len(ncj), len([j for j in ncj_intra if ncj_intra[j]])))
            log(DEBUG3, 2, "Critical jobs: %s" % [str(j) for j in cj])
            log(DEBUG3, 2, "Non-critical jobs:")
            for j in ncj:
                log(DEBUG3, 3, "%s, dominated by at least %s." % 
                    (str(j), str(ncj_reasons[j][0])))

            self._stats["alljobs_count"] = len(all_jobs)
            self._stats["ncj_count"] = len(ncj)
            self._stats["ncj_intra_count"] = len([j for j in ncj_intra 
                if ncj_intra[j]])

        else:
            cj = set([j for T in self._taskset.get_tasks() 
                for j in T.get_jobs()])

        D = max([j.get_d() for j in cj])
        log(DEBUG, 1, "Preparing request functions for D=%d." % D)
        indent_log(2)
        # These will be used in all interference testers, so generate them
        # here just once.
        rfpath_generators = dict()
        for task in self._taskset.get_tasks():
            rfpath_generators[task] = RfPathGenerator(task, D)
            rfpath_generators[task].prepare()
        indent_log(-2)
        log(DEBUG, 1, "Done preparing request functions.")

        # Go through priority list in reverse order, since it's most likely
        # to find a deadline violation at the lowest priority.
        # TODO: Some testing whether this is actually true
        indent_log(1)
        idx_list = reversed(range(len(self._taskset.get_tasks())))
        tasks = self._taskset.get_priolist()
        for i in idx_list:
            checktask = tasks[i]
            log(DEBUG, 0, "Checking lowest-priority feasibility of task '%s'." %
                str(checktask))
            
            log(DEBUG2, 1, "Interfering tasks: '%s'" %
                "', '".join([str(tasks[j]) for j in range(i)]))

            lpengine = LPEngine(checktask, tasks[:i], rfpath_generators,
                critical_jobs=cj)
            lpengine.run()
            res = lpengine.get_result()
            if res:
                log(DEBUG, 0, "Task ok.") 
                #print "Task ok. " + str(checktask)  # Morteza
            else:
                log(DEBUG, 0, "Task can't have this priority.")
                #print "Task can't have this priority. " + str(checktask)  # Morteza
                break
        indent_log(-1)
        if res:
            log(INFO, 1, ":-) Taskset schedulable with this priority order!")
        else:
            log(INFO, 1, ":-( Taskset unschedulable with this priority order!")
  
        timer_total.stop()
        log(INFO, 0, "SP Schedulability test completed after %.2fs." %
            timer_total.get_t())
        self._stats["totaltime"] = timer_total.get_t()
        self._stats["result"] = res

        self._result = res

    # Added by Morteza, 
    #  used for response time calc.
    def is_lowest_priority_schedulable(self, checktask):
        timer_total = Timer()
        log(INFO, 0, "Checking SP schedulability of taskset '%s'." %
            str(self._taskset))

        if self._cj_optimization:
            log(DEBUG, 1, "Building critical_jobs set.")

            cj = set()
            for t in self._taskset.get_priolist():
                # Critical jobs in current task
                cj_task = set()

                for job in t.get_jobs():
                    # Remove from cj all jobs that we dominate
                    cj.difference_update([j 
                        for j in cj if job.dominates(j, strict=True)])
                    # Remove from cj_task all jobs that we dominate
                    cj_task.difference_update([j 
                        for j in cj_task if job.dominates(j, strict=True)])
                    # Add ourselves to cj_task if we are not dominated 
                    # by anything in cj_task
                    if all([not j.dominates(job, strict=True) for j in cj_task]):
                        cj_task.add(job)
                cj = cj.union(cj_task)

            # Non-critical jobs; mainly for debugging
            ncj = set()
            ncj_reasons = dict()
            ncj_intra = dict()
            tasks = self._taskset.get_priolist()
            for i in range(len(tasks)):
                t = tasks[i]
                for j in t.get_jobs():
                    if j not in cj:
                        ncj.add(j)
                        intra_task_dominators = cj.intersection(t.get_jobs())
                        inter_task_dominators = cj.intersection([j2 
                            for t2 in tasks[i+1:] for j2 in t2.get_jobs()])
                        ncj_reasons[j] = list(intra_task_dominators) + \
                            list(inter_task_dominators)
                        ncj_intra[j] = (len(intra_task_dominators) > 0)

            all_jobs = [j for t in self._taskset.get_tasks() 
                for j in t.get_jobs()]
            log(DEBUG2, 2, "%d of %d (%3.2f%%) jobs are critical." % 
                (len(cj), len(all_jobs), float(len(cj)*100)/len(all_jobs)))
            log(DEBUG2, 2, ("Of %d non-critical jobs, %d are intra-task " +
                "dominated.") % 
                (len(ncj), len([j for j in ncj_intra if ncj_intra[j]])))
            log(DEBUG3, 2, "Critical jobs: %s" % [str(j) for j in cj])
            log(DEBUG3, 2, "Non-critical jobs:")
            for j in ncj:
                log(DEBUG3, 3, "%s, dominated by at least %s." % 
                    (str(j), str(ncj_reasons[j][0])))

            self._stats["alljobs_count"] = len(all_jobs)
            self._stats["ncj_count"] = len(ncj)
            self._stats["ncj_intra_count"] = len([j for j in ncj_intra 
                if ncj_intra[j]])

        else:
            cj = set([j for T in self._taskset.get_tasks() 
                for j in T.get_jobs()])

        D = max([j.get_d() for j in cj])
        log(DEBUG, 1, "Preparing request functions for D=%d." % D)
        indent_log(2)
        # These will be used in all interference testers, so generate them
        # here just once.
        rfpath_generators = dict()
        for task in self._taskset.get_tasks():
            rfpath_generators[task] = RfPathGenerator(task, D)
            rfpath_generators[task].prepare()
        indent_log(-2)
        log(DEBUG, 1, "Done preparing request functions.")

        # Go through priority list in reverse order, since it's most likely
        # to find a deadline violation at the lowest priority.
        # TODO: Some testing whether this is actually true
        
        
        # Analysing the schedulability of the checktask:
        indent_log(1)
        hp_tasks = self._taskset.get_priolist()
        hp_tasks = [t for t in hp_tasks if t._name != checktask._name]
        
        lpengine = LPEngine(checktask, hp_tasks, rfpath_generators,
        critical_jobs=cj)
        lpengine.run()
        res = lpengine.get_result()
        if res:
            log(DEBUG, 0, "Task ok.") 
            # print "Task ok. " + str(checktask)  # Morteza
        else:
            log(DEBUG, 0, "Task can't have this priority.")
            # print "Task can't have this priority. " + str(checktask)  # Morteza
            # break
            
        indent_log(-1)
        if res:
            log(INFO, 1, ":-) Taskset schedulable with this priority order!")
        else:
            log(INFO, 1, ":-( Taskset unschedulable with this priority order!")
        
        timer_total.stop()
        log(INFO, 0, "SP Schedulability test completed after %.2fs." %
            timer_total.get_t())
        self._stats["totaltime"] = timer_total.get_t()
        self._stats["result"] = res
        
        self._result = res


class SPFeasibilityAnalyzer(object):
    """Tests a given taskset for static priority (SP) feasibility on 
    a uniprocessor platform.

    Method is taken from the RTSS'13 article and works similar to
    SPSchedulabilityAnalizer (see there). The difference is that the
    feasibility test runs Audsley's algorithm in order to determine a feasible
    priority order bottom-up. While doing so, the set of critical jobs may
    shrink.

    In order to get the result, first call the run() method after which the
    result can be retrieved via get_result(). This result is the synthesized
    priority order in case of a feasible input.
    """


    def __init__(self, taskset, cj_optimization=True):
        """
        Args:
            taskset: Taskset to be tested. Instance of TaskSet, i.e., 
                does NOT need to provide a priority order.
            cj_optimization: Only test critical jobs. Safe optimization.

        Note:
            * Limited to constrainted deadlines.
        """
        self._taskset = taskset
        self._cj_optimization = cj_optimization

        self.reset()
        log(INFO, 0, "SP Feasibility Analyzer set up for taskset '%s'." %
            str(taskset))

    def reset(self):
        self._rfpath_generators = dict()
        self._result = None
        stats[self] = dict()
        self._stats = stats[self]

    def get_result(self):
        return self._result

    def run(self):
        timer_total = Timer()
        log(INFO, 0, "Checking SP feasibility of taskset '%s'." %
            str(self._taskset))

        all_jobs = set([j for T in self._taskset.get_tasks()
                for j in T.get_jobs()])

        # For critical jobs optimization, remove set of 
        # intra-task non-critical jobs.
        if self._cj_optimization:
            log(DEBUG, 1, "Building initial set of critical jobs.")

            # Intra-task dominance: j1 >= j2 if j1.e >= j2.e and j1.d <= j2.d
            cj = set()
            for t in self._taskset.get_tasks():
                # Critical jobs in current task
                cj_task = set()

                for job in t.get_jobs():
                    # Remove from cj_task all jobs that we dominate
                    cj_task.difference_update([j 
                        for j in cj_task if job.dominates(j, strict=True)])
                    # Add ourselves to cj_task if we are not dominated 
                    # by anything in cj_task
                    if all([not j.dominates(job, strict=True) for j in cj_task]):
                        cj_task.add(job)
                cj = cj.union(cj_task)

            log(DEBUG2, 2, "%d of %d (%.2f%%) jobs are intra-task critical." % 
                (len(cj), len(all_jobs), float(len(cj)*100)/len(all_jobs)))
            log(DEBUG3, 2, "Critical jobs: %s" % [str(j) for j in cj])
            
            self._stats["alljobs_count"] = len(all_jobs)
            self._stats["ncj_intra_count"] = len(all_jobs) - len(cj)
            self._stats["ncj_count"] = len(all_jobs) - len(cj)

            self._jobs_to_test = cj
        else:
            self._jobs_to_test = all_jobs

        D = max([j.get_d() for j in self._jobs_to_test])
        log(DEBUG, 1, "Preparing request functions for D=%d." % D)
        timer_rf = Timer()
        indent_log(2)
        # These will be used in all interference testers, so generate them
        # here just once.
        rfpath_generators = dict()
        for task in self._taskset.get_tasks():
            rfpath_generators[task] = RfPathGenerator(task, D)
            rfpath_generators[task].prepare()
        self._rfpath_generators = rfpath_generators
        indent_log(-2)
        timer_rf.stop()
        self._stats["rftime"] = timer_rf.get_t()
        log(DEBUG, 1, "Done preparing request functions in %.2fs." %
            timer_rf.get_t())
        
        indent_log(1)
        res = self.get_prio_order(self._taskset.get_tasks())
        indent_log(-1)

        if res is not None:
            log(INFO, 1, ":-) Taskset is SP feasible!")
            prio_order = res
            log(DEBUG, 1, "Priority order: '%s'" %
                "', '".join([str(task) for task in prio_order]))
        else:
            log(INFO, 1, ":-( Taskset is not SP feasible.")
  
        timer_total.stop()
        log(INFO, 0, "SP Feasibility test completed after %.2fs." %
            timer_total.get_t())
        self._stats["totaltime"] = timer_total.get_t()
        self._stats["sp-feasible"] = res is not None
        self._stats["result"] = res

        self._result = res


    def get_prio_order(self, tasks):

        if len(tasks) == 0:
            if self._cj_optimization:
                log(INFO, 0, "Done. %d jobs were critical." % 
                    len(self._jobs_to_test))
            else:
                log(INFO, 0, "Done.")
            return []

        log(INFO, 0, "Determining priority order for tasks %s." 
            % str([str(T) for T in tasks]))

#        D = max([j.get_d() for j in self._jobs_to_test.intersection(
#            set([j for T in tasks for j in T.get_jobs()]))])
#        log(DEBUG, 1, "Upper bound for scheduling window size: D=%d" % D)

        # Test all tasks until one is found that can have lowest priority. 
        indent_log(1)
        for lp_task in tasks:
               
            hp_tasks = tasks.copy()
            hp_tasks.remove(lp_task)
            log(DEBUG, 0, "Checking lowest-priority feasibility of task %s." 
                % str(lp_task))
            
            log(DEBUG2, 1, "Interfering tasks: '%s'" %
                "', '".join([str(t) for t in hp_tasks]))

            lpengine = LPEngine(lp_task, hp_tasks, self._rfpath_generators,
                critical_jobs=self._jobs_to_test)
            lpengine.run()
            res = lpengine.get_result()
            
            if res:
                log(DEBUG, 1, "Task is lp-feasible.")

                if self._cj_optimization:
                    # Update set of jobs to be tested;
                    # More jobs could be inter-task non-critical.
                    # Inter-task dominance: T1.j1 >= T2.j2 if
                    #    j1 >= j2 and T1 >= T2 
                    new_ncj = set()
                    for job in lp_task.get_jobs():
                        hp_jobs = set(j for t in hp_tasks 
                            for j in t.get_jobs())
                        hp_to_test = self._jobs_to_test.intersection(hp_jobs)
                        new_ncj_job = [j for j in hp_to_test
                            if job.dominates(j, strict=True)]
                        log(DEBUG, 2, "New non-critical jobs from '%s': '%s'" %
                            (str(job),
                             "', '".join([str(j) for j in new_ncj_job])))
                        new_ncj.update(new_ncj_job)
                    old_cj = self._jobs_to_test.copy()
                    self._jobs_to_test.difference_update(new_ncj)
                    log(DEBUG, 2, ("Now, we have %d critical jobs left. " +
                        "(Before: %d)") % 
                            (len(self._jobs_to_test), len(old_cj)))
                    self._stats["ncj_count"] += len(new_ncj)

                break
            else:
                log(DEBUG, 1, "Task can't have this priority.")
        indent_log(-1)


        # Done checking one task. Now either recurse or abort
        if res:
            log(DEBUG, 0, "Found LP task '%s'." % str(lp_task))
            res = self.get_prio_order(hp_tasks)
            if res is None:
                return None
            else:
                res.append(lp_task)
                return res
        else:
            log(DEBUG, 0, "No task can have lowest priority. Infeasible.")
            return None

    def get_mrf(self, maxTime):
        log(INFO, 0, "Checking SP feasibility of taskset '%s'." %
            str(self._taskset))

        all_jobs = set([j for T in self._taskset.get_tasks()
                for j in T.get_jobs()])

        # For critical jobs optimization, remove set of 
        # intra-task non-critical jobs.
        if self._cj_optimization:
            log(DEBUG, 1, "Building initial set of critical jobs.")

            # Intra-task dominance: j1 >= j2 if j1.e >= j2.e and j1.d <= j2.d
            cj = set()
            for t in self._taskset.get_tasks():
                # Critical jobs in current task
                cj_task = set()

                for job in t.get_jobs():
                    # Remove from cj_task all jobs that we dominate
                    cj_task.difference_update([j 
                        for j in cj_task if job.dominates(j, strict=True)])
                    # Add ourselves to cj_task if we are not dominated 
                    # by anything in cj_task
                    if all([not j.dominates(job, strict=True) for j in cj_task]):
                        cj_task.add(job)
                cj = cj.union(cj_task)

            log(DEBUG2, 2, "%d of %d (%.2f%%) jobs are intra-task critical." % 
                (len(cj), len(all_jobs), float(len(cj)*100)/len(all_jobs)))
            log(DEBUG3, 2, "Critical jobs: %s" % [str(j) for j in cj])
            
            self._stats["alljobs_count"] = len(all_jobs)
            self._stats["ncj_intra_count"] = len(all_jobs) - len(cj)
            self._stats["ncj_count"] = len(all_jobs) - len(cj)

            self._jobs_to_test = cj
        else:
            self._jobs_to_test = all_jobs

        timer_rf = Timer()
        indent_log(2)
        # These will be used in all interference testers, so generate them
        # here just once.
        rfpath_generators = dict()
        for task in self._taskset.get_tasks():
            rfpath_generators[task] = RfPathGenerator(task, maxTime)
            rfpath_generators[task].prepare()
        self._rfpath_generators = rfpath_generators
        indent_log(-2)
        timer_rf.stop()
        self._stats["rftime"] = timer_rf.get_t()
        log(DEBUG, 1, "Done preparing request functions in %.2fs." %
            timer_rf.get_t())
        
        # Changes by Morteza starts here:
        mrf_list = []
        for task in self._taskset.get_tasks():
            rfroot = rfpath_generators[task].get_rfroot(maxTime)
            mrf_list.append(rfroot.get_rf())
        
        return mrf_list

class LPEngine(object):

    """Checks lowest-priority feasibility of given task, with set of hp_tasks
    as interfering tasks of higher priority.
    A dict of generators for rfpaths needs to be supplied and thus created
    beforehand. Only the set of critical jobs will be tested, which, if none is
    supplied, is the set of all jobs.

    After the engine is run with run(), the result can be obtained with
    get_result(). It's either
    * True, iff the task can tolerate any interference from hp_tasks
    * False, iff it cannot, or
    * None, iff the engine wasn't run (successfully) yet since the last reset.
    """

    def __init__(self, task, hp_tasks, rfpath_generators, critical_jobs=None):
        self._task = task
        self._hp_tasks = hp_tasks
        self._rfpath_generators = rfpath_generators
        if critical_jobs is None:
            self._critical_jobs = self._task.get_jobs()
        else:
            self._critical_jobs = critical_jobs
        self.reset()

    def reset(self):
        self._result = None

    def get_result(self):
        return self._result

    def run(self):
        if len(self._hp_tasks) == 0:
            log(DEBUG, 0, "Trivially LP feasible.")
            self._result = True
            return

        jobs_to_test = set(self._task.get_jobs()).intersection(
            self._critical_jobs)
        jobs_not_to_test = set(self._task.get_jobs()).difference(
            self._critical_jobs)
        log(DEBUG2, 0, "Jobs to test: %s" % 
            ", ".join([str(j) for j in jobs_to_test]))
        log(DEBUG2, 0, "Jobs skipped: %s" % 
            ", ".join([str(j) for j in jobs_not_to_test]))

        # This is in case jobs_to_test is empty
        res = False
        for j in jobs_to_test:
            log(DEBUG, 0, "Checking job '%s'..." % str(j))
            indent_log(1)
            res = interference_test(j.get_e(), j.get_d(), self._hp_tasks,
                self._rfpath_generators)
            indent_log(-1)
            if res:
                log(DEBUG, 1, "Job is NOT lowest-priority feasible.")
                break
            else:
                log(DEBUG, 1, "Job is lowest-priority feasible.")
        self._result = not res



def interference_test(e, d, hp_tasks, rfpath_generators, 
        stats_dict=None):
    """
    Checks whether hp_tasks can cause more than d-e interference
    in a window of d time units. That is, True if a job <e, d> could miss
    its deadline.

    Implementation using combinatorial abstraction refinement.
    For this, rfpath_generators must be supplied in order to generate
    rfpaths and abstraction trees.

    Arg:
        stats_dict: If not None, some statistics will be saved in this dict.
    """
    
    # Dict with key = task and value = set of rfpaths
    rfpaths = dict([
        (task, rfpath_generators[task].get_rfpaths(d).copy()) 
        for task in hp_tasks])
    rfroots = dict([
        (task, rfpath_generators[task].get_rfroot(d)) 
        for task in hp_tasks])

    for task in hp_tasks:
        log(DEBUG2, 0, "Root path for task '%s': %s" % 
            (str(task), str(rfroots[task])))

    start_tuple = list(rfroots.values())
    todo = [start_tuple]

    stats_total = reduce(lambda x,y:x*y, [len(paths) 
        for paths in rfpaths.values()])
    stats_tested = 0
    stats_pos = 0
    stats_neg = 0
    timer = Timer()

    while len(todo) > 0:
        rftuple = todo.pop()
        log(DEBUG3, 1, "Testing combination [%s]" % 
            ', '.join([p.get_name() for p in rftuple]))

        rfs = [p.get_rf() for p in rftuple]
        idle_instant = None
        t = 0
        while t < d:
            log(DEBUG4, 2, "Checking workload at t=%d..." % t)
            workload = sum([rf[t] for rf in rfs]) + e
            if workload <= t + 1:
                idle_instant = t
                break
            else:
                t = workload - 1
        stats_tested += 1

        if idle_instant is not None:
            log(DEBUG3, 2, "Found idle instant at t=%d." % t)
            stats_pos += 1
        else:
            log(DEBUG3, 2, "Found no idle instant.")
            stats_neg += 1
            
            # Check whether rftuple is abstract
            is_abstract = any([not rfpath.is_concrete() for rfpath in rftuple])

            if is_abstract:
                # Need to do more work.
                # Push concretization of some abstract path.
                log(DEBUG3, 3, "This combination was abstract. " + 
                    "Pushing concretization.")
                (max_spread, min_pos, min_rfpath) = max([
                    (rfpath.spread, i, rfpath) 
                    for (i, rfpath) in enumerate(rftuple) 
                    if len(rfpath.jobs) == 0])
                log(DEBUG4, 4, "Splitting path %s with spread %.5f." % 
                    (str(min_rfpath), max_spread))
                for rfpath in min_rfpath.childs:
                    new_rftuple = rftuple[:]
                    new_rftuple[min_pos] = rfpath
                    log(DEBUG4, 4, "Pushing combination [%s]" % 
                        ', '.join([p.get_name() for p in new_rftuple]))
                    todo.append(new_rftuple)
            else:
                # Done: Found deadline miss.
                log(DEBUG3, 3, "This combination was concrete. We are done.")
                break

    timer.stop()
    log(DEBUG, 0, "Done with %s after %.2fs. Summary:" % 
        ("SCHED" if idle_instant else "UNSCHED", timer.get_t()))
    log(DEBUG, 1, "Combinations tested / total: %d / %d (%.1f%%)" %
        (stats_tested, stats_total, float(stats_tested*100)/stats_total))

    if stats_dict is not None:
        stats_dict["total"] = stats_total
        stats_dict["tested"] = stats_tested
        stats_dict["pos"] = stats_pos
        stats_dict["neg"] = stats_neg
        stats_dict["result"] = (idle_instant is None)
        stats_dict["time"] = timer.get_t()

    return idle_instant is None
 
 

def response_time(e, d, hp_tasks, rfpath_generators, 
        stats_dict=None):
    """
    Checks whether hp_tasks can cause more than d-e interference
    in a window of d time units. That is, True if a job <e, d> could miss
    its deadline.

    Implementation using combinatorial abstraction refinement.
    For this, rfpath_generators must be supplied in order to generate
    rfpaths and abstraction trees.

    Arg:
        stats_dict: If not None, some statistics will be saved in this dict.
    """
    
    # Dict with key = task and value = set of rfpaths
    rfpaths = dict([
        (task, rfpath_generators[task].get_rfpaths(d).copy()) 
        for task in hp_tasks])
    rfroots = dict([
        (task, rfpath_generators[task].get_rfroot(d)) 
        for task in hp_tasks])

    for task in hp_tasks:
        log(DEBUG2, 0, "Root path for task '%s': %s" % 
            (str(task), str(rfroots[task])))

    start_tuple = list(rfroots.values())
    todo = [start_tuple]

    stats_total = reduce(lambda x,y:x*y, [len(paths) 
        for paths in rfpaths.values()])
    stats_tested = 0
    stats_pos = 0
    stats_neg = 0
    timer = Timer()

    while len(todo) > 0:
        rftuple = todo.pop()
        log(DEBUG3, 1, "Testing combination [%s]" % 
            ', '.join([p.get_name() for p in rftuple]))

        rfs = [p.get_rf() for p in rftuple]
        idle_instant = None
        t = 0
        while t < d:
            log(DEBUG4, 2, "Checking workload at t=%d..." % t)
            workload = sum([rf[t] for rf in rfs]) + e
            if workload <= t + 1:
                idle_instant = t
                break
            else:
                t = workload - 1
        stats_tested += 1

        if idle_instant is not None:
            log(DEBUG3, 2, "Found idle instant at t=%d." % t)
            stats_pos += 1
        else:
            log(DEBUG3, 2, "Found no idle instant.")
            stats_neg += 1
            
            # Check whether rftuple is abstract
            is_abstract = any([not rfpath.is_concrete() for rfpath in rftuple])

            if is_abstract:
                # Need to do more work.
                # Push concretization of some abstract path.
                log(DEBUG3, 3, "This combination was abstract. " + 
                    "Pushing concretization.")
                (max_spread, min_pos, min_rfpath) = max([
                    (rfpath.spread, i, rfpath) 
                    for (i, rfpath) in enumerate(rftuple) 
                    if len(rfpath.jobs) == 0])
                log(DEBUG4, 4, "Splitting path %s with spread %.5f." % 
                    (str(min_rfpath), max_spread))
                for rfpath in min_rfpath.childs:
                    new_rftuple = rftuple[:]
                    new_rftuple[min_pos] = rfpath
                    log(DEBUG4, 4, "Pushing combination [%s]" % 
                        ', '.join([p.get_name() for p in new_rftuple]))
                    todo.append(new_rftuple)
            else:
                # Done: Found deadline miss.
                log(DEBUG3, 3, "This combination was concrete. We are done.")
                break

    timer.stop()
    log(DEBUG, 0, "Done with %s after %.2fs. Summary:" % 
        ("SCHED" if idle_instant else "UNSCHED", timer.get_t()))
    log(DEBUG, 1, "Combinations tested / total: %d / %d (%.1f%%)" %
        (stats_tested, stats_total, float(stats_tested*100)/stats_total))

    if stats_dict is not None:
        stats_dict["total"] = stats_total
        stats_dict["tested"] = stats_tested
        stats_dict["pos"] = stats_pos
        stats_dict["neg"] = stats_neg
        stats_dict["result"] = (idle_instant is None)
        stats_dict["time"] = timer.get_t()

    if idle_instant is None:
        return -1
    else:
        return idle_instant + 1 

class RfPathGenerator(object):
    """Generates rfpaths from a task on domains of any size D <= maxD.
    When initializing/instantiating, supply the maximal domain size maxD that 
    will ever be encountered. Run prepare() once to generate all critical
    request functions on this domain. The object will cache the result of this
    run.

    Subsequent calls to get_rfroot() and get_rfpaths() take a D as argument and
    will derive the resulting set of rfpaths from the originally cached set.
    Note that smaller D potentially means a smaller result set since suddenly,
    new dominances might appear.

    Use get_rfpaths() to retrieve the full set of rfpaths.
    Use get_rfroot() for one rfpath which is potentially abstract since it is
    the root of the abstraction tree.
    """

    def __init__(self, task, maxD, 
            path_class = None,
            abstract_path_class = None,
            store_class = None):
        self._task = task
        self._maxD = maxD
        self._path_class = path_class
        self._abstract_path_class = abstract_path_class
        self._store_class = store_class
        if self._path_class is None:
            self._path_class = FullRfPath
        if self._abstract_path_class is None:
            self._abstract_path_class = AbstractRfPath
        if self._store_class is None:
            self._store_class = DiscardingRfStore

        # TODO: Replace by proper caching technique
        self._cache = dict()

    def prepare(self):
        D = self._maxD
        log(DEBUG, 0, "Computing request functions for task %s..." % 
            str(self._task))
        log(DEBUG2, 1, 'D:           %d' % D)
        log(DEBUG2, 1, 'Path class: %s' % self._path_class.__name__)
        log(DEBUG2, 1, 'Abstract Path class:  %s' % 
            self._abstract_path_class.__name__)
        log(DEBUG2, 1, 'Store class: %s' % self._store_class.__name__)

        timer = Timer()
        # Explore task graph.
        indent_log(2)
        store = self._store_class(D)
        path_creator = lambda jlist: self._path_class(jlist, max_horizon = D)
        ee = ExplorationEngine(self._task, store, path_creator)
        ee.run()
        indent_log(-2)
        log(DEBUG, 1, 'Done exploring task graph after %.2f s. %d paths.' % 
            (timer.stop(), len(store.get_paths())))

        timer = Timer()
        log(DEBUG, 1, "Filtering for maximal relevant paths...")
        maxpaths = set()
        add_to_maxpaths_rf(store.get_paths(), maxpaths, D)
        log(DEBUG, 1, 'Done after %.2f s. %d paths left.' % (timer.stop(),
            len(maxpaths)))

        for i, p in enumerate(maxpaths):
            p.set_name("%s_D%d_p%d" % (str(self._task), D, i))

        self._cache[(D, None)] = maxpaths

    def get_rfroot(self, D):
        """Returns abstraction tree roots."""

        def rfs_max(paths, maxD):
            return [max([p.get_rf()[i] for p in paths]) for i in range(maxD)]

        if D > self._maxD:
            raise RuntimeError("Too large D.")

        if not (D, 1) in self._cache:
            # Root is not in cache yet
            log(DEBUG2, 0, "Deriving rfroot for task %s for D=%d..." 
                % (str(self._task), D))
            indent_log(1)
            paths = self.get_rfpaths(D).copy()
            indent_log(-1)
            while len(paths) > 1:
                # Compute distance matrix
                distances = list()
                for ((i1, p1), (i2, p2)) in itertools.product(
                    enumerate(paths), repeat=2):
                    if i1 < i2:
                        distances.append((p1.distance_to(p2), i1, i2, p1, p2))
                for (d, _, _, p1, p2) in distances:
                    log(DEBUG4, 1, "%s to %s: %.3f" % (p1.get_name(), 
                        p2.get_name(), d))

                # Merge path pair with minimal distances
                (dist, _, _, p1, p2) = min(distances)
                log(DEBUG4, 1, "Minimum choice: %.3f via (%s, %s)" % (dist, 
                    p1.get_name(), p2.get_name()))
                maxf = rfs_max([p1, p2], D)
                maxpath = self._abstract_path_class(D, maxf,
                    childs = [p1, p2], spread = dist)
                maxpath.set_name("%s_A1_D%d_p%d" % (str(self._task), D,
                    len(paths)-2))
                log(DEBUG3, 1, "New path %s." % str(maxpath))

                # Replace merged paths with maximum
                paths.remove(p1)
                paths.remove(p2)
                paths.add(maxpath)
            # Now there is only one path left in paths. Use it as root.
            
            # added by Morteza to prevent error when the graph is not connected or paths is empty
            #if len(paths) == 0:
            #    empty_job = drt.model.Job(0, 10)
            #    paths = set(self._path_class([empty_job, D))
            
            self._cache[(D, 1)] = list(paths)[0]
        else:
            log(DEBUG2, 0, "Using cached rfroot for task %s for D=%d..." % 
                (str(self._task), D))
        return self._cache[(D, 1)]

    def get_rfpaths(self, D):
        """Returns concrete rf."""
        if D > self._maxD:
            raise RuntimeError("Too large D.")

        if not (D, None) in self._cache:
            # There is no cached path set for this D, but we can derive one.
            maxD = self._maxD
            log(DEBUG2, 0, "Deriving rfpaths for task %s for D=%d..." 
                % (str(self._task), D))
            log(DEBUG3, 1, "Original set for D=%d:" % maxD)
            for path in self._cache[(maxD, None)]:
                log(DEBUG3, 2, "Path %s." % str(path))

            newpaths = set([self._path_class(p.get_jobs(), D) 
                for p in self._cache[(maxD, None)]])
            log(DEBUG2, 1, "Deriving from %d paths." % len(newpaths))

            timer = Timer()
            log(DEBUG2, 1, "Filtering for maximal relevant paths...")
            maxpaths = set()
            add_to_maxpaths_rf(newpaths, maxpaths, D)
            log(DEBUG2, 1, 'Done after %.2f s. %d paths left.' % (timer.stop(),
                len(maxpaths)))

            log(DEBUG3, 1, "Resulting new set for D=%d:" % D)
            for i, p in enumerate(maxpaths):
                p.set_name("%s_D%d_p%d" % (str(self._task), D, i))
                log(DEBUG3, 2, "Path %s." % str(p))
            self._cache[(D, None)] = maxpaths
        else:
            log(DEBUG2, 0, "Using cached rfpaths for task %s for D=%d..." % 
                (str(self._task), D))
        return self._cache[(D, None)]




def add_to_maxpaths_rf(paths, maxpaths, D):
    for path in paths:

        log(DEBUG3, 2, "Processing: %s" % path)

        # Check whether too short
        if path.get_horizon() < D:
            log(DEBUG3, 3, "Too short.")
            continue

        # Check whether we are dominated by some processed path
        dominated = False
        new_rf = path.get_rf()
        for p in maxpaths:
            rf = p.get_rf()
            dominated = all([rf[i] >= new_rf[i] for i in range(D)])
            if dominated:
                log(DEBUG3, 3, "Dominated by %s." % p)
                break
        if dominated:
            continue
        log(DEBUG3, 3, "New. Will add.")

        # Remove all paths dominated by us
        remove_paths = set()
        for p in maxpaths:
            rf = p.get_rf()
            dominated = all([rf[i] <= new_rf[i] for i in range(D)])
            if dominated:
                log(DEBUG3, 4, "Dominates: %s" % p)
                remove_paths.add(p)
        maxpaths.difference_update(remove_paths)

        # Done. Add finally.
        maxpaths.add(path)



def is_sp_feasible(TS):
    """Shortcut for testing the given taskset's SP feasibility.
    """
    A = SPFeasibilityAnalyzer(TS)
    A.run()
    return A.get_result()

def get_mrf(TS, maxTime):
    """Shortcut for testing the given taskset's SP feasibility.
    """
    A = SPFeasibilityAnalyzer(TS)
    return A.get_mrf(maxTime)

def calc_response_time(hp_tasks, task, maxD):
    A = SPSchedulabilityAnalyzer(hp_tasks)
    res = []
    for job in task.get_jobs():
        e = job.get_e()
        # rt = A.response_time(e, maxD)
        rt = A.run()
        res.append(rt)
        
    return res

def is_sp_schedulable(TS):
    """Shortcut for testing the given taskset's SP schedulability
    with given task priority order.
    """
    A = SPSchedulabilityAnalyzer(TS)
    A.run()
    return A.get_result()

# TS is the list of higher priority tasks
# returns True is task is SP schedulable
def is_sp_lowestpriority_schedulable(task, hpTS):
    """Shortcut for testing the given taskset's SP schedulability
    with given task priority order.
    """
    A = SPSchedulabilityAnalyzer(hpTS)
    A.is_lowest_priority_schedulable(task)
    return A.get_result()


# vim:textwidth=79:shiftwidth=4:tabstop=4:cc=80:expandtab:autoindent