# This file is part of libdrt.
#
# Written 2011-2014 by Martin Stigge, Pontus Ekberg.
#
# GPLv3 applies.


"""
Path abstractions for dbf computations.
"""

from drt.util.logging import ERROR, INFO, DEBUG, DEBUG2, DEBUG3, log, indent_log
from drt.analysis.util.paths import Path

class DbfPath(Path):
    
    """Template for path abstraction classes used to compute the dbf."""

    def get_e(self):
        """Return the total execution requirement for the path."""
        raise NotImplementedError

    def get_d(self):
        """Return the deadline of the path (i.e., the latest deadline
        relative to the spawning of the first job).
        """
        raise NotImplementedError


class FullDbfPath(DbfPath):

    """Stores entire concrete paths (i.e., it performs no
    abstractions at all). Domination relation for dbf computations.
    """

    def get_e(self):
        return sum([job.get_e() for job in self.jobs])

    def get_d(self):
        if len(self.jobs) == 0:
            return 0
        else:
            task = self.jobs[0].get_task()
            edges = zip(self.jobs[:-1], self.jobs[1:])
            return (sum(task.get_delay(edge) for edge in edges)
                + self.jobs[-1].get_d())

    def get_end(self):
        return self.jobs[-1]

    def get_extended(self, job):
        return [FullDbfPath(self.jobs[:] + [job])]

    def __len__(self):
        return len(self.jobs) - 1

    def __ge__(self, o):
        return (self.get_end() == o.get_end() 
            and self.get_e() >= o.get_e() 
            and self.get_d() <= o.get_d())
			
    def __gt__(self, o):
        return (self >= o and (self.get_e() != o.get_e() 
            or self.get_d() != o.get_d()))

    def __str__(self):
        return '[' + ' -> '.join(map(str, self.jobs)) + ']'


class DemandTriple(DbfPath):

    """DemandTriple stores paths abstracted as demand triples, i.e.,
    as triples <e, d, v> of execution time, deadline and last vertex.
    """

    def __init__(self, jobs, e=None, d=None):
        if e is None:
            self.e = sum([job.get_e() for job in jobs])
        else:
            self.e = e

        if d is None:
            task = jobs[0].get_task()
            edges = zip(jobs[:-1], jobs[1:])
            self.d = (sum(task.get_delay(edge) for edge in edges)
                + jobs[-1].get_d())
        else:
            self.d = d

        self.last = jobs[-1]
    
    def get_jobs(self):
        raise RuntimeError("Abstract path does not have jobs.")

    def get_e(self):
        return self.e

    def get_d(self):
        return self.d

    def get_end(self):
        return self.last

    def get_extended(self, job):
        task = job.get_task()
        e = self.e + job.get_e()
        d = (self.d - self.last.get_d() + task.get_delay((self.last, job)) 
            + job.get_d())
        return [DemandTriple([job], e, d)]

    def __ge__(self, o):
        return (self.last == o.last and self.e >= o.e and self.d <= o.d)

    def __gt__(self, o):
        return (self >= o and (self.get_e() != o.get_e()
            or self.get_d() != o.get_d()))


    def __str__(self):
        return '<%d, %d, %s>' % (self.e, self.d, self.last)



class DemandQuadruple(DbfPath):

    """DemandQuadruple stores paths abstracted as demand quadruples, i.e.,
    as tuples <e, d, p, v> of execution time, deadline, duration
    and last vertex.
    """

    def __init__(self, jobs, e=None, d=None, p=None):
        if e is None:
            self.e = sum([job.get_e() for job in jobs])
        else:
            self.e = e

        if d is None:
            task = jobs[0].get_task()
            edges = zip(jobs[:-1], jobs[1:])
            self.d = (sum(task.get_delay(edge) for edge in edges)
                + jobs[-1].get_d())
        else:
            self.d = d

        if p is None:
            task = jobs[0].get_task()
            edges = zip(jobs[:-1], jobs[1:])
            self.p = sum(task.get_delay(edge) for edge in edges)
        else:
            self.p = p

        self.last = jobs[-1]
    
    def get_jobs(self):
        raise RuntimeError("Abstract path does not have jobs.")

    def get_e(self):
        return self.e

    def get_p(self):
        return self.p

    def get_d(self):
        return self.d

    def get_end(self):
        return self.last

    def get_extended(self, job):
        """ Demand quadruples are abstractions of *marked* paths, so there are
        two extensions possible: marked and non-marked"""
        task = job.get_task()

        e = self.e + job.get_e()
        d = max(self.d, self.p + task.get_delay((self.last, job)) +
            + job.get_d())
        p = self.p + task.get_delay((self.last, job))
        dq1 = DemandQuadruple([job], e, d, p)

        e = self.e
        d = self.d
        p = self.p + task.get_delay((self.last, job))
        dq2 = DemandQuadruple([job], e, d, p)

        return [dq1, dq2]

    def __ge__(self, o):
        """ Domination relation unclear for demand quadruples. """
        raise NotImplementedError

    def __gt__(self, o):
        """ Domination relation unclear for demand quadruples. """
        raise NotImplementedError


    def __str__(self):
        return '<%d, %d, %d, %s>' % (self.e, self.d, self.p, self.last)



# vim:textwidth=79:shiftwidth=4:tabstop=4:cc=80:expandtab:autoindent

