# This file is part of libdrt.
#
# Written 2011-2014 by Martin Stigge, Pontus Ekberg.
#
# GPLv3 applies.


"""
Path abstractions for rbf computations.
"""

from drt.util.logging import ERROR, INFO, DEBUG, DEBUG2, DEBUG3, log, indent_log
from drt.analysis.util.paths import Path

class RbfPath(Path):
    
    """Template for path abstraction classes used to compute the rbf."""

    def get_e(self):
        """Return the total execution requirement for the path."""
        raise NotImplementedError

    def get_p(self):
        """Return the duration of the path (i.e., difference between
        first and last release)
        """
        raise NotImplementedError


class FullRbfPath(RbfPath):

    """Stores entire concrete paths (i.e., it performs no
    abstractions at all). Domination relation for rbf computations.
    """

    def get_e(self):
        return sum([job.get_e() for job in self.jobs])

    def get_p(self):
        if len(self.jobs) == 0:
            return 0
        else:
            task = self.jobs[0].get_task()
            edges = zip(self.jobs[:-1], self.jobs[1:])
            return sum(task.get_delay(edge) for edge in edges)

    def get_end(self):
        return self.jobs[-1]

    def get_extended(self, job):
        return [FullRbfPath(self.jobs[:] + [job])]

    def __len__(self):
        return len(self.jobs) - 1

    def __ge__(self, o):
        return (self.get_end() == o.get_end() 
            and self.get_e() >= o.get_e() 
            and self.get_p() <= o.get_p())
			
    def __gt__(self, o):
        return (self >= o and (self.get_e() != o.get_e() 
            or self.get_p() != o.get_p()))

    def __str__(self):
        return '[' + ' -> '.join(map(str, self.jobs)) + ']'


class RequestTriple(RbfPath):

    """RequestTriple stores paths abstracted as request triples, i.e.,
    as triples <e, p, v> of execution time, duration and last vertex.

    Note: Semantics is different from Demand Triples. p=10 for a request triple
    means that *in time slot 10*, the request is e, i.e., effective at t=11
    if we focus on integers.
    d=10 for a demand triple means that *for 10 time slots*.
    So, rbf(11) >= e but dbf(10) >= e.

    (Implicitly, rbf(0) == 0.)
    """

    def __init__(self, jobs, e=None, p=None):
        if e is None:
            self.e = sum([job.get_e() for job in jobs])
        else:
            self.e = e

        if p is None:
            task = jobs[0].get_task()
            edges = zip(jobs[:-1], jobs[1:])
            self.p = sum(task.get_delay(edge) for edge in edges)
        else:
            self.p = p

        self.last = jobs[-1]
    
    def get_jobs(self):
        raise RuntimeError("Abstract path does not have jobs.")

    def get_e(self):
        return self.e

    def get_p(self):
        return self.p

    def get_end(self):
        return self.last

    def get_extended(self, job):
        task = job.get_task()
        e = self.e + job.get_e()
        p = self.p + task.get_delay((self.last, job))
        return [RequestTriple([job], e, p)]

    def __ge__(self, o):
        return (self.last == o.last and self.e >= o.e and self.p <= o.p)

    def __gt__(self, o):
        return (self >= o and (self.get_e() != o.get_e()
            or self.get_p() != o.get_p()))

    def __str__(self):
        return '<%d, %d, %s>' % (self.e, self.p, self.last)



# vim:textwidth=79:shiftwidth=4:tabstop=4:cc=80:expandtab:autoindent

