# This file is part of libdrt.
#
# Written 2014 by Martin Stigge.
#
# GPLv3 applies.

import unittest

import drt.util.logging
import drt.examples
import drt.analysis.util

from drt.model import TaskSet
from drt.analysis.utilization import UEngine
from drt.analysis.dbf import DbfEngine
from drt.analysis.edf import EDFSchedulabilityAnalyzer, is_edf_schedulable


class Test_DRT_EDF(unittest.TestCase):

    def setUp(self):
        drt.util.logging._threshold = 0

    def test_U(self):
        T = drt.examples.rtas11()

        ue = UEngine(T)
        ue.run()
        result = ue.get_U()
        self.assertEqual(result, 1.0/6)

        ue = UEngine(T,
            path_class = drt.analysis.util.upaths.FullUtilizationPath)
        ue.run()
        result = ue.get_U()
        self.assertEqual(result, 1.0/6)

    def test_dbf(self):
        T = drt.examples.rtas11()

        dbfengine = drt.analysis.dbf.DbfEngine(T, 50, increments=None)              
        dbfengine.prepare()                                                         
        self.assertEqual(
            [dbfengine.get_dbf(t) for t in range(50)],
            [0, 0, 0, 0, 0, 2, 2, 2, 3, 3, 
             5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
             6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
             6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 
             8, 8, 8, 9, 9, 9, 9, 9, 9, 9])


        # Compare increment and non-increment versions
        D = 23456

        dbfengine1 = drt.analysis.dbf.DbfEngine(T, D, increments=None)
        dbfengine1.prepare()
        dbf1 = []
        for t in xrange(D):
            dbf1.append(dbfengine1.get_dbf(t))

        dbfengine2 = drt.analysis.dbf.DbfEngine(T, D, increments=1111)
        dbfengine2.prepare()
        dbf2 = []
        for t in xrange(D):
            dbf2.append(dbfengine2.get_dbf(t))
        
        self.assertEqual(dbf1, dbf2)


    def test_dbf_arbitrary(self):
        T = drt.examples.rtas11_arbitrary()

        dbfengine = drt.analysis.dbf.DbfEngine(T, 20, increments=None,
            path_class = drt.analysis.util.dbfpaths.DemandQuadruple,
            store_class = drt.analysis.util.pathstores.PathStorePD)
        dbfengine.prepare()                                                         
        self.assertEqual(
            [dbfengine.get_dbf(t) for t in range(20)],
            [0, 0, 0, 0, 0, 2, 2, 5, 5, 5, 
             5, 5, 5, 5, 5, 5, 7, 7, 8, 8])


    def test_edf_schedulability(self):
        T1 = drt.examples.rtas11()
        T2 = drt.examples.PeriodicTask(5, 10, 20)
        T3 = drt.examples.PeriodicTask(6, 10, 20)

        TS1 = TaskSet([T1, T2])
        TS2 = TaskSet([T1, T3])

        # Default test with DemandTriples
        A = EDFSchedulabilityAnalyzer(TS1)
        A.run()
        self.assertTrue(A.get_result())

        # Full path (i.e. no) abstraction
        A = EDFSchedulabilityAnalyzer(TS1,
                path_class = drt.analysis.util.dbfpaths.FullDbfPath,
                u_path_class = drt.analysis.util.upaths.FullUtilizationPath)
        A.run()
        self.assertTrue(A.get_result())

        # Also test shortcut
        self.assertFalse(is_edf_schedulable(TS2))
   
if __name__ == "__main__":
    unittest.main()
