# This file is part of libdrt.
#
# Written 2014 by Martin Stigge.
#
# GPLv3 applies.

import unittest

import drt.util.logging
import drt.examples
import drt.analysis.util

from drt.model import TaskSet
from drt.analysis.rbf import RbfEngine
from drt.analysis.busywindow import MBPEngine
from drt.analysis.sjp import RfPathGenerator, RfSfxPathGenerator
from drt.analysis.sjp import JLPEngine, interference_test
from drt.analysis.sjp import SJPSchedulabilityAnalyzer, SJPFeasibilityAnalyzer
from drt.analysis.sjp import is_sjp_schedulable, is_sjp_feasible


class Test_DRT_SJP(unittest.TestCase):

    def setUp(self):
        drt.util.logging._threshold = 0

    def test_rbf(self):
        T = drt.examples.rtas11()
        rbfengine = RbfEngine(T, increments=None)
        rbfengine.prepare()
        self.assertEqual([rbfengine.get_rbf(t) for t in range(50)],
                         [ 5,  5,  5,  5,  5,  5,  5,  5,  5,  5,
                           6,  6,  6,  6,  6,  6,  6,  6,  6,  6,
                           6,  6,  6,  6,  6,  6,  6,  6,  6,  6,
                           8,  8,  8,  8,  8,  9,  9,  9,  9,  9,
                          11, 11, 11, 11, 11, 11, 11, 11, 11, 11])

    def test_busywindow(self):
        TS = drt.examples.RTS_SI_RTSS13_Example()
        mbpengine = MBPEngine(TS)
        mbp = mbpengine.get_max_busy_period()
        self.assertEqual(mbp, 18)

    def test_minimize(self):
        T = drt.examples.non_minimal()
        nonzero_jobs = set(j for j in T.get_jobs() if j.get_e() > 0)
        T2 = drt.analysis.sjp.create_restricted_task(T, nonzero_jobs)
        self.assertEqual(len(T2), 2)
        jobs = dict((j._name, j) for j in T2.get_jobs())
        self.assertEqual(jobs['_v1'].get_e(), 7)
        self.assertEqual(jobs['_v1'].get_d(), 8)
        self.assertEqual(jobs['_v2'].get_e(), 20)
        self.assertEqual(jobs['_v2'].get_d(), 30)
        self.assertEqual(T2.get_delay((jobs['_v1'], jobs['_v2'])), 8)
        self.assertEqual(T2.get_delay((jobs['_v2'], jobs['_v1'])), 42)

        T = drt.examples.non_minimal2()
        nonzero_jobs = set(j for j in T.get_jobs() if j.get_e() > 0)
        T2 = drt.analysis.sjp.create_restricted_task(T, nonzero_jobs)
        self.assertEqual(len(T2), 5)
        jobs = dict((j._name, j) for j in T2.get_jobs())
        self.assertEqual(jobs['_v1'].get_e(), 1)
        self.assertEqual(jobs['_v1'].get_d(), 10)
        self.assertEqual(jobs['_v2'].get_e(), 2)
        self.assertEqual(jobs['_v2'].get_d(), 20)
        self.assertEqual(jobs['_v3'].get_e(), 3)
        self.assertEqual(jobs['_v3'].get_d(), 30)
        self.assertEqual(jobs['_v4'].get_e(), 4)
        self.assertEqual(jobs['_v4'].get_d(), 40)
        self.assertEqual(jobs['_v5'].get_e(), 5)
        self.assertEqual(jobs['_v5'].get_d(), 50)
        self.assertEqual(T2.get_delay((jobs['_v1'], jobs['_v3'])), 40)
        self.assertEqual(T2.get_delay((jobs['_v2'], jobs['_v5'])), 70)
        self.assertEqual(T2.get_delay((jobs['_v2'], jobs['_v4'])), 60)
        self.assertEqual(T2.get_delay((jobs['_v1'], jobs['_v4'])), 50)
        self.assertEqual(T2.get_delay((jobs['_v1'], jobs['_v5'])), 60)
        self.assertEqual(T2.get_delay((jobs['_v2'], jobs['_v3'])), 50)

    def test_interference(self):
        TS = drt.examples.RTS_SI_RTSS13_Example()
        (Thigh,) = [T for T in TS.get_tasks() if T._name == "T1"]
        (T,) = [T for T in TS.get_tasks() if T._name == "T2"]
        (v0,) = [v for v in T.get_jobs() if v._name == "v_0"]
        (v1,) = [v for v in T.get_jobs() if v._name == "v_1"]
        
        rfgenerator = RfPathGenerator(Thigh, 30)
        rfgenerator.prepare()
        rfreversegenerator = RfSfxPathGenerator(T, v1, 30)
        rfreversegenerator.prepare()

        res = drt.analysis.sjp.interference_test(
            d = v1.get_d(), 
            L = 18,
            hp_tasks = set([Thigh]),
            rfpath_generators = {Thigh: rfgenerator},
            rfsfx_generator = rfreversegenerator)
        self.assertTrue(res)
     

        TS = drt.examples.precision_demo()
        (Thigh,) = [T for T in TS.get_tasks() if T._name == "T1"]
        (T,) = [T for T in TS.get_tasks() if T._name == "T2"]
        (v0,) = [v for v in T.get_jobs() if v._name == "x_T2"]
        
        rfgenerator = RfPathGenerator(Thigh, 30)
        rfgenerator.prepare()
        rfreversegenerator = RfSfxPathGenerator(T, v0, 30)
        rfreversegenerator.prepare()

        res = drt.analysis.sjp.interference_test(
            d = v0.get_d(), 
            L = 10,
            hp_tasks = set([Thigh]),
            rfpath_generators = {Thigh: rfgenerator},
            rfsfx_generator = rfreversegenerator)
        self.assertFalse(res)


    def test_jlpengine(self):
        # Try infeasble version
        TS = drt.examples.RTS_SI_RTSS13_Example()
        (Thigh,) = [T for T in TS.get_tasks() if T._name == "T1"]
        (T,) = [T for T in TS.get_tasks() if T._name == "T2"]
        (v0,) = [v for v in T.get_jobs() if v._name == "v_0"]
        (v1,) = [v for v in T.get_jobs() if v._name == "v_1"]
        
        hp_jobs = set(Thigh.get_jobs())
        hp_jobs.add(v0)

        jlpengine = JLPEngine(v1, hp_jobs)
        jlpengine.run()
        self.assertFalse(jlpengine.get_result())

        # Try feasible version
        TS = drt.examples.RTS_SI_RTSS13_Example2()
        (Thigh,) = [T for T in TS.get_tasks() if T._name == "T1"]
        (T,) = [T for T in TS.get_tasks() if T._name == "T2"]
        (v0,) = [v for v in T.get_jobs() if v._name == "v_0"]
        (v1,) = [v for v in T.get_jobs() if v._name == "v_1"]
        
        hp_jobs = set(Thigh.get_jobs())
        hp_jobs.add(v0)

        jlpengine = JLPEngine(v1, hp_jobs)
        jlpengine.run()
        self.assertTrue(jlpengine.get_result())

    def test_sjp_schedulability(self):
        # Try feasible version
        TS = drt.examples.RTS_SI_RTSS13_Example2()
        (Thigh,) = [T for T in TS.get_tasks() if T._name == "T1"]
        (vhigh,) = Thigh.get_jobs()
        (T,) = [T for T in TS.get_tasks() if T._name == "T2"]
        (v0,) = [v for v in T.get_jobs() if v._name == "v_0"]
        (v1,) = [v for v in T.get_jobs() if v._name == "v_1"]
        
        TS._priolist = [v0, vhigh, v1]
        A = drt.analysis.sjp.SJPSchedulabilityAnalyzer(TS)
        A.run()
        self.assertTrue(A.get_result())

        # Try infeasible version
        TS = drt.examples.RTS_SI_RTSS13_Example()
        self.assertTrue(is_sjp_feasible(TS) is None)

 

if __name__ == "__main__":
    unittest.main()
