# This file is part of libdrt.
#
# Written 2014 by Martin Stigge.
#
# GPLv3 applies.

import unittest

import drt.util.logging
import drt.examples
import drt.analysis.util

from drt.model import TaskSet
from drt.analysis.sp import RfPathGenerator, interference_test, LPEngine
from drt.analysis.sp import SPSchedulabilityAnalyzer, SPFeasibilityAnalyzer
from drt.analysis.sp import is_sp_schedulable, is_sp_feasible


class Test_DRT_SP(unittest.TestCase):

    def setUp(self):
        drt.util.logging._threshold = 0

    def test_rfpath(self):
        T = drt.examples.rtss13()

        V = dict([(job._name, job) for job in T.get_jobs()])
        self.assertEqual([T.get_mindelay(v) for v in V.values()],
                         [13, 29, 10, 12, 50])

        p = drt.analysis.util.rfpaths.FullRfPath(
            [V['v1'], V['v2'], V['v3']], max_horizon=50)
        self.assertEqual(p.get_rf(),
                         [ 6,  6,  6,  6,  6,  6,  6,  6,  6,  6,
                           6,  6,  6, 11, 11, 11, 11, 11, 11, 11, 
                          11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 
                          11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 
                          11, 11, 12, 12, 12, 12, 12, 12, 12, 12])

    def test_rfgenerator(self):
        T = drt.examples.thesis_refinement()
        rfgenerator = RfPathGenerator(T, 30)
        rfgenerator.prepare()
        rfpaths = rfgenerator.get_rfpaths(30)
        rfs = set(tuple(p.get_rf()[:30]) for p in rfpaths)

        self.assertEqual(len(rfs), 3)
        self.assertTrue((6,  6,  6,  6,  6,  6,  6,  6,  6,  6, 
                         6,  6,  9,  9,  9,  9,  9,  9,  9,  9, 
                         9,  9, 13, 13, 13, 13, 13, 13, 13, 13) in rfs)

        self.assertTrue(( 4,  4,  4,  4,  4,  4,  4,  4, 10, 10, 
                         10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 
                         13, 13, 13, 13, 13, 13, 13, 13, 13, 13) in rfs)

        self.assertTrue((4, 4, 4, 4, 4, 4, 4, 4,  4,  4, 
                         4, 4, 4, 4, 4, 4, 4, 4,  4,  4, 
                         8, 8, 8, 8, 8, 8, 8, 8, 14, 14) in rfs)

    def test_interference(self):
        T = drt.examples.interference_demo()
        rfgenerator = RfPathGenerator(T, 20)
        rfgenerator.prepare()
        # Taskset causes interference of 12 in a 20 time units interval
        self.assertFalse(interference_test(8, 20, [T], {T: rfgenerator}))
        self.assertTrue(interference_test(9, 20, [T], {T: rfgenerator}))

    def test_lpengine(self):
        TS = drt.examples.sp_feasibility_demo(mode=1)
        rfgenerators = dict()
        hp_tasks = TS.get_priolist()[:-1] 
        for T in hp_tasks:
            rfgenerators[T] = RfPathGenerator(T, 100)
            rfgenerators[T].prepare()
        lpengine = LPEngine(TS.get_priolist()[-1],
            hp_tasks=hp_tasks,
            rfpath_generators=rfgenerators)
        lpengine.run()
        self.assertTrue(lpengine.get_result())

        TS = drt.examples.sp_feasibility_demo(mode=2)
        rfgenerators = dict()
        hp_tasks = TS.get_priolist()[:-1] 
        for T in hp_tasks:
            rfgenerators[T] = RfPathGenerator(T, 100)
            rfgenerators[T].prepare()
        lpengine = LPEngine(TS.get_priolist()[-1],
            hp_tasks=hp_tasks,
            rfpath_generators=rfgenerators)
        lpengine.run()
        self.assertFalse(lpengine.get_result())

    def test_sp_schedulability(self):
        TS = drt.examples.sp_feasibility_demo(mode=1)
        A = SPSchedulabilityAnalyzer(TS)
        A.run()
        self.assertTrue(A.get_result())

        TS = drt.examples.sp_feasibility_demo(mode=2)
        self.assertFalse(is_sp_schedulable(TS))

    def test_sp_feasibility(self):
        TS = drt.examples.sp_feasibility_demo(mode=2)
        [T1, T2, T4, T3] = TS.get_priolist()
        A = SPFeasibilityAnalyzer(TS)
        A.run()
        res = A.get_result()
        self.assertEqual(res, [T1, T2, T3, T4])

        TS = drt.examples.sp_feasibility_demo(mode=3)
        self.assertFalse(is_sp_feasible(TS))

    def test_sp_hardness(self):
        # Positive 3-PARTITION instance from the paper
        TS = drt.examples.ecrts12_transform([15, 16, 12, 16, 14, 
                                             11, 15, 12, 15], 42, gmf=True)
        self.assertFalse(is_sp_feasible(TS))

        # Negative 3-PARTITION instance (3 odd numbers can't add up to 100)
        TS = drt.examples.ecrts12_transform([27, 43, 27, 29, 39, 35], 100, gmf=True)
        self.assertTrue(is_sp_feasible(TS))
        


if __name__ == "__main__":
    unittest.main()
