# This file is part of libdrt.
#
# Written 2014 by Martin Stigge
#
# GPLv3 applies.

from __future__ import absolute_import

import itertools

from drt.model import Job, TaskSet, DRTTask
from drt.analysis.util.explorationengines import *

from drt.analysis.util.rfpaths import FullRfPath
from drt.analysis.util.rfpaths import AbstractRfPath
from drt.analysis.util.rfpaths import DiscardingRfStore

from drt.analysis.util.rfreversepaths import FullRfReversePath
from drt.analysis.util.rfreversepaths import AbstractRfReversePath
from drt.analysis.util.rfreversepaths import DiscardingRfReverseStore

from drt.analysis.busywindow import MBPEngine

from drt.util.statistics import Timer, stats

from drt.util.logging import ERROR, INFO, DEBUG, DEBUG2, DEBUG3, DEBUG4
from drt.util.logging import log, indent_log

# TODO: Used for general caching. Basically a hack. Make nicer.
GLOBAL_CACHE = {"D": dict(), "deactivate": False}


class SJPSchedulabilityAnalyzer(object):
    """Tests a given taskset for static job-type priority (SJP) schedulability 
    on a uniprocessor platform.

    Method is taken from the RTS Special Issue article from 2015. For each job
    that is to be tested, all possibly interfering jobs of higher priority are
    represented with request functions. Each function represents all jobs from
    the same task which have a higher priority than the tested job. A busy
    window extension technique is applied in order to obtain safe results.

    In order to get the result, first call the run() method after which the
    result can be retrieved via get_result().
    """

    def __init__(self, taskset):
        """
        Args:
            taskset: Taskset to be tested. Instance of SJPTaskSet, i.e., needs
                to provide get_priolist() method with a priority order.

        Note:
            * Limited to constrainted deadlines.
        """
        if taskset.get_priolist() is None:
            raise RuntimeError("SJP Schedulability test needs priority list.")

        self._taskset = taskset

        self.reset()
        log(INFO, 0, "SJP Schedulability Analyzer set up for taskset '%s'." %
            str(taskset))

    def reset(self):
        self._result = None
        stats[self] = dict()
        self._stats = stats[self]

    def get_result(self):
        return self._result

    def run(self):
        timer_total = Timer()
        log(INFO, 0, "Checking SJP schedulability of taskset '%s'." %
            str(self._taskset))

        jobs = self._taskset.get_priolist()

        indent_log(1)
        idx_list = reversed(range(len(jobs)))
        for i in idx_list:
            checkjob = jobs[i]
            log(DEBUG, 0, "Checking lowest-priority feasibility of job '%s'." %
                str(checkjob))
            
            log(DEBUG2, 1, "Interfering jobs: '%s'" %
                "', '".join([str(jobs[j]) for j in range(i)]))

            jlpengine = JLPEngine(checkjob, jobs[:i])
            jlpengine.run()
            res = jlpengine.get_result()

            if res:
                log(DEBUG, 0, "Job ok.")
            else:
                log(DEBUG, 0, "Job can't have this priority.")
                break
        indent_log(-1)
        if res:
            log(INFO, 1, ":-) Taskset schedulable with this priority order!")
        else:
            log(INFO, 1, ":-( Taskset unschedulable with this priority order!")
  
        timer_total.stop()
        log(INFO, 0, "SJP Schedulability test completed after %.2fs." %
            timer_total.get_t())
        self._stats["totaltime"] = timer_total.get_t()
        self._stats["result"] = res

        self._result = res



class SJPFeasibilityAnalyzer(object):
    """Tests a given taskset for static job-type priority (SJP) feasibility on 
    a uniprocessor platform.

    Method is taken from the RTS Special Issue journal paper from 2015.
    Works similar to SJPSchedulabilityAnalizer (see there). 
    The difference is that the feasibility test runs Audsley's algorithm in 
    order to determine a feasible priority order bottom-up. 

    In order to get the result, first call the run() method after which the
    result can be retrieved via get_result(). This result is the synthesized
    priority order in case of a feasible input.
    """


    def __init__(self, taskset):
        """
        Args:
            taskset: Taskset to be tested. Instance of TaskSet, i.e., 
                does NOT need to provide a job priority order.

        Note:
            * Limited to constrainted deadlines.
        """
        self._taskset = taskset

        self.reset()
        log(INFO, 0, "SJP Feasibility Analyzer set up for taskset '%s'." %
            str(taskset))

    def reset(self):
        self._result = None
        stats[self] = dict()
        self._stats = stats[self]

    def get_result(self):
        return self._result

    def run(self):
        timer_total = Timer()
        log(INFO, 0, "Checking SJP feasibility of taskset '%s'." %
            str(self._taskset))

        tasks = self._taskset.get_tasks() 

        #alljobs = set([job for task in tasks for job in task.get_jobs()])
        alljobs = sorted([job for task in tasks for job in task.get_jobs()],
            key = lambda job: job.get_d(), reverse = True)


        indent_log(1)
        res = self.get_prio_order(alljobs)
        indent_log(-1)

        if res is not None:
            log(INFO, 1, ":-) Taskset is SJP feasible!")
            prio_order = res
            log(DEBUG, 1, "Priority order: '%s'" %
                "', '".join([str(job) for job in prio_order]))
        else:
            log(INFO, 1, ":-( Taskset is not SJP feasible.")
  
        timer_total.stop()
        log(INFO, 0, "SJP Feasibility test completed after %.2fs." %
            timer_total.get_t())
        self._stats["totaltime"] = timer_total.get_t()
        self._stats["sjp-feasible"] = res is not None
        self._stats["result"] = res

        self._result = res


    def get_prio_order(self, jobs):

        if len(jobs) == 0:
            log(INFO, 0, "Done.")
            return []

        log(INFO, 0, "Determining job priority order for jobs %s." 
            % str([str(job) for job in jobs]))

        # Test all jobs until one is found that can have lowest priority. 
        indent_log(1)
        for lp_job in jobs:
               
#            hp_jobs = jobs.copy()
            hp_jobs = list(jobs)
            hp_jobs.remove(lp_job)
            log(DEBUG, 0, "Checking lowest-priority feasibility of job %s." 
                % str(lp_job))
            
            log(DEBUG2, 1, "Interfering jobs: '%s'" %
                "', '".join([str(job) for job in hp_jobs]))

            jlpengine = JLPEngine(lp_job, hp_jobs)
            jlpengine.run()
            res = jlpengine.get_result()
            
            if res:
                log(DEBUG, 1, "Job is lp-feasible.")
                break
            else:
                log(DEBUG, 1, "Job can't have this priority.")
        indent_log(-1)


        # Done checking one job. Now either recurse or abort.
        if res:
            log(DEBUG, 0, "Found LP job '%s'." % str(lp_job))
            res = self.get_prio_order(hp_jobs)
            if res is None:
                return None
            else:
                res.append(lp_job)
                return res
        else:
            log(DEBUG, 0, "No job can have lowest priority. Infeasible.")
            return None


class JLPEngine(object):
    """Checks lowest-priority feasibility of given job, with set of hp_jobs
    as interfering jobs of higher priority.

    Task information is pulled directly from the job objects and the rfpath
    generators etc. are created based on that. No caching, that's tricky to
    implement (and of unclear gain).

    After the engine is run with run(), the result can be obtained with
    get_result(). It's either
    * True, iff the job can tolerate any interference from hp_jobs
    * False, iff it cannot, or
    * None, iff the engine wasn't run (successfully) yet since the last reset.
    """


    def __init__(self, job, hp_jobs):
        self._job = job
        self._hp_jobs = hp_jobs
        self.reset()

    def reset(self):
        self._result = None

    def get_result(self):
        return self._result

    def run(self):
        if len(self._hp_jobs) == 0:
            log(DEBUG, 0, "Trivially JLP feasible (no hp jobs).")
            self._result = True
            return

        old_job = self._job

        # Create all minimized interfering tasks

        old_hp_tasks = set(j.get_task() for j in self._hp_jobs)
        old_own_task = old_job.get_task()
        if old_own_task in old_hp_tasks:
            old_hp_tasks.remove(old_own_task)

        old_jobs_keep = set(self._hp_jobs)  # Copy this one
        old_jobs_keep.add(old_job)

        hp_tasks = set()
        for task in old_hp_tasks:
            T = create_restricted_task(task, old_jobs_keep)
            if len(T) > 0: 
                hp_tasks.add(T)
        if len(hp_tasks) == 0:
            log(DEBUG, 0, "Trivially JLP feasible (no hp tasks).")
            self._result = True
            return

        (own_task, d) = create_restricted_task(old_own_task, old_jobs_keep,
            get_mapping = True)
        if d is not None:
            job = d[old_job]
        else:
            job = old_job

        # Compute domain D+L (deadline + busy window extension)
        D = job.get_d()
        mbpengine = MBPEngine(TaskSet(hp_tasks.union([own_task])))
        L = mbpengine.get_max_busy_period()

        # Create rfpath generators for interfering tasks
        log(DEBUG, 1, "Preparing request functions for D+L=%d." % (D+L))
        indent_log(2)
        rfpath_generators = dict()
        for task in hp_tasks:
            rfpath_generators[task] = RfPathGenerator(task, D+L)
            rfpath_generators[task].prepare()
        indent_log(-2)
        log(DEBUG, 1, "Done preparing request functions.")

        # Create own rfsfx
        log(DEBUG, 1, "Preparing sfx request function for D+L=%d." % (D+L))
        indent_log(2)
        rfsfx_generator = RfSfxPathGenerator(own_task, job, D+L)
        rfsfx_generator.prepare()
        indent_log(-2)
        log(DEBUG, 1, "Done preparing sfx request function.")

#        indent_log(1)
        res = interference_test(job.get_d(), L, hp_tasks,
            rfpath_generators, rfsfx_generator)
#        indent_log(-1)
        self._result = not res


def create_restricted_task(task, jobs_keep, get_mapping=False):
    """Creates a new task from <task> which reduces all jobs NOT in <jobs_keep>
    to an execution time of 0 and then minimizes the graph by merging
    edges going in and out of 0-wcet vertices.

    Returns <task> in case all its jobs are in <jobs_keep>.
    """

    if jobs_keep.issuperset(task.get_jobs()):
        if get_mapping:
            return (task, None)
        else:
            return task

    # There is a job in task which is not in jobs, i.e., should be removed.
    jobs_remove = task.get_jobs().difference(jobs_keep)

    newjobs = dict([(job, Job(job.get_e(), job.get_d(), name=("_%s" %
        job._name))) for job in task.get_jobs()])

    newtask = DRTTask(newjobs.values(), name=("_%s" % task._name))
    for edge in task.get_edges():
        (v1, v2) = edge
        newtask.add_edge( (newjobs[v1], newjobs[v2]),
            task.get_delay(edge))

    newjobs_remove = set(newjobs[j] for j in jobs_remove)

    # Minimize graph
    while True:
        s = newjobs_remove.intersection(newtask.get_jobs())
        if len(s) == 0:
            break
        rj = s.pop()
        in_edges = [e for e in newtask.get_edges() if e[1] == rj]
        out_edges = [e for e in newtask.get_edges() if e[0] == rj]
        for (edge1, edge2) in itertools.product(in_edges, out_edges):
            (v1, _) = edge1
            (_, v2) = edge2
            d = newtask.get_delay(edge1) + newtask.get_delay(edge2)
            newtask.add_edge((v1, v2), d)
        newtask.remove_job(rj)  # Removes edges as well

    if get_mapping:
        return (newtask, newjobs)
    else:
        return newtask


def interference_test(d, L, hp_tasks, rfpath_generators, rfsfx_generator,
        stats_dict=None):
    """
    Checks whether hp_tasks can cause more than d-e interference
    in a window of d time units. That is, True if a job <e, d> could miss
    its deadline. Considers busy window extensions up to L.

    Implementation using combinatorial abstraction refinement.
    For this, rfpath_generators must be supplied in order to generate
    rfpaths and abstraction trees.
    Another generator rfsfx_generator is for paths in the tested graph.

    Arg:
        stats_dict: If not None, some statistics will be saved in this dict.
    """
    
    # Dict with key = task and value = set of rfpaths
    rfpaths = dict([
        (task, rfpath_generators[task].get_paths(d+L).copy()) 
        for task in hp_tasks])
    rfroots = dict([
        (task, rfpath_generators[task].get_root(d+L)) 
        for task in hp_tasks])

    for task in hp_tasks:
        log(DEBUG2, 0, "Root path for task '%s': %s" % 
            (str(task), str(rfroots[task])))

    rfsfxpaths = rfsfx_generator.get_paths(L)
    rfsfxroot = rfsfx_generator.get_root(L)
    log(DEBUG2, 0, "Root rfsfx path: %s" % str(rfsfxroot))

    # Tuples: Position 0 is own task, i.e., an rfreverse path
    start_tuple = [rfsfxroot] + list(rfroots.values())
    todo = [start_tuple]

    stats_total = reduce(lambda x,y:x*y, [len(paths) 
        for paths in rfpaths.values()])
    stats_total *= len(rfsfxpaths)
    stats_tested = 0
    stats_pos = 0
    stats_neg = 0
    timer = Timer()

    while len(todo) > 0:
        rftuple = todo.pop()
        log(DEBUG3, 1, "Testing combination [%s]" % 
            ', '.join([p.get_name() for p in rftuple]))

        rfsfx = rftuple[0].get_rfreverse()
        rfs = [p.get_rf() for p in rftuple[1:]]
        x = 0
        feasible = True
        while x < L and feasible:
            t = 0
            idle_instant = None
            while t < d:
                workload = (sum([rf[t+x] for rf in rfs]) + rfsfx[x])
                log(DEBUG4, 2, "Workload at x=%d, t=%d is %d." % (x, t,
                    workload))
                if workload <= x + t + 1:
                    idle_instant = t
                    log(DEBUG4, 2, "Found idle instant!")
                    break
                else:
                    t = workload - x - 1
            if idle_instant is None:
                feasible = False
                log(DEBUG3, 2, "Found no idle instant for x=%d, t=%d." % 
                    (x, t))

            # Increase x until rfsfx changes.
            Y = rfsfx[x]
            while x < L and rfsfx[x] == Y:
                x += 1
        stats_tested += 1

        if feasible:
            log(DEBUG3, 2, "Found idle instant for all x.")
            stats_pos += 1
        else:
            log(DEBUG3, 2, "Found no idle instant.")
            stats_neg += 1
            
            # Check whether rftuple is abstract
            is_abstract = any([not rfpath.is_concrete() for rfpath in rftuple])

            if is_abstract:
                # Need to do more work.
                # Push concretization of some abstract path.
                log(DEBUG3, 3, "This combination was abstract. " + 
                    "Pushing concretization.")
                (max_spread, min_pos, min_rfpath) = max([
                    (rfpath.spread, i, rfpath) 
                    for (i, rfpath) in enumerate(rftuple) 
                    if len(rfpath.jobs) == 0])
                log(DEBUG4, 4, "Splitting path %s with spread %.5f." % 
                    (str(min_rfpath), max_spread))
                for rfpath in min_rfpath.childs:
                    new_rftuple = rftuple[:]
                    new_rftuple[min_pos] = rfpath
                    log(DEBUG4, 4, "Pushing combination [%s]" % 
                        ', '.join([p.get_name() for p in new_rftuple]))
                    todo.append(new_rftuple)
            else:
                # Done: Found deadline miss.
                log(DEBUG3, 3, "This combination was concrete. We are done.")
                break

    timer.stop()
    log(DEBUG, 0, "Done with %s after %.2fs. Summary:" % 
        ("SCHED" if idle_instant else "UNSCHED", timer.get_t()))
    log(DEBUG, 1, "Combinations tested / total: %d / %d (%.1f%%)" %
        (stats_tested, stats_total, float(stats_tested*100)/stats_total))

    if stats_dict is not None:
        stats_dict["total"] = stats_total
        stats_dict["tested"] = stats_tested
        stats_dict["pos"] = stats_pos
        stats_dict["neg"] = stats_neg
        stats_dict["result"] = (idle_instant is None)
        stats_dict["time"] = timer.get_t()

    return idle_instant is None
 

class PathGenerator(object):

    def __init__(self, task, maxD, path_class, abstract_path_class,
            store_class, start_jobs, backwards, 
            paths_max, add_to_maxpaths, pathtype_str):
        self._task = task
        self._maxD = maxD
        self._path_class = path_class
        self._abstract_path_class = abstract_path_class
        self._store_class = store_class
        self._start_jobs = start_jobs
        self._backwards = backwards
        self._paths_max = paths_max
        self._add_to_maxpaths = add_to_maxpaths
        self._pathtype_str = pathtype_str

        # TODO: Replace by proper caching technique
        self._cache = dict()

    def prepare(self):
        D = self._maxD

        log(DEBUG, 0, "Computing %ss for task %s on domain %d..." % 
            (self._pathtype_str, str(self._task), D))
        log(DEBUG2, 1, 'D:           %d' % D)
        log(DEBUG2, 1, 'Path class: %s' % self._path_class.__name__)
        log(DEBUG2, 1, 'Abstract Path class:  %s' % 
            self._abstract_path_class.__name__)
        log(DEBUG2, 1, 'Store class: %s' % self._store_class.__name__)

        timer = Timer()
        # Explore task graph.
        indent_log(2)
        store = self._store_class(D)
        path_creator = lambda jlist: self._path_class(jlist, max_horizon = D)
        ee = ExplorationEngine(self._task, store, path_creator,
            start_jobs = self._start_jobs, backwards = self._backwards)
        ee.run()
        indent_log(-2)
        log(DEBUG, 1, 'Done exploring task graph after %.2f s. %d paths.' % 
            (timer.stop(), len(store.get_paths())))

        timer = Timer()
        log(DEBUG, 1, "Filtering for maximal relevant paths...")
        maxpaths = set()
        self._add_to_maxpaths(store.get_paths(), maxpaths, D)
        log(DEBUG, 1, 'Done after %.2f s. %d paths left.' % (timer.stop(),
            len(maxpaths)))

        for i, p in enumerate(maxpaths):
            p.set_name("%s_D%d_p%d" % (str(self._task), D, i))

        self._cache[(D, None)] = maxpaths


    def get_root(self, D):
        """Returns abstraction tree roots."""

        if D > self._maxD:
            raise RuntimeError("Too large D.")

        if not (D, 1) in self._cache:
            # Root is not in cache yet
            log(DEBUG2, 0, "Deriving root for task %s for D=%d..." 
                % (str(self._task), D))
            indent_log(1)
            paths = self.get_paths(D).copy()
            indent_log(-1)
            while len(paths) > 1:
                # Compute distance matrix
                distances = list()
                for ((i1, p1), (i2, p2)) in itertools.product(
                    enumerate(paths), repeat=2):
                    if i1 < i2:
                        distances.append((p1.distance_to(p2), i1, i2, p1, p2))
                for (d, _, _, p1, p2) in distances:
                    log(DEBUG4, 1, "%s to %s: %.3f" % (p1.get_name(), 
                        p2.get_name(), d))

                # Merge path pair with minimal distances
                (dist, _, _, p1, p2) = min(distances)
                log(DEBUG4, 1, "Minimum choice: %.3f via (%s, %s)" % (dist, 
                    p1.get_name(), p2.get_name()))
                maxf = self._paths_max([p1, p2], D)
                maxpath = self._abstract_path_class(D, maxf,
                    childs = [p1, p2], spread = dist)
                maxpath.set_name("%s_A1_D%d_p%d" % (str(self._task), D,
                    len(paths)-2))
                log(DEBUG3, 1, "New path %s." % str(maxpath))

                # Replace merged paths with maximum
                paths.remove(p1)
                paths.remove(p2)
                paths.add(maxpath)
            # Now there is only one path left in paths. Use it as root.
            self._cache[(D, 1)] = list(paths)[0]
        else:
            log(DEBUG2, 0, "Using cached root for task %s for D=%d..." % 
                (str(self._task), D))
        return self._cache[(D, 1)]

    def get_paths(self, D):
        """Returns concrete paths."""
        if D > self._maxD:
            raise RuntimeError("Too large D.")

        if not (D, None) in self._cache:
            # There is no cached path set for this D, but we can derive one.
            maxD = self._maxD
            log(DEBUG2, 0, "Deriving concrete paths for task %s for D=%d..." 
                % (str(self._task), D))
            log(DEBUG3, 1, "Original set for D=%d:" % maxD)
            for path in self._cache[(maxD, None)]:
                log(DEBUG3, 2, "Path %s." % str(path))

            newpaths = set([self._path_class(p.get_jobs(), D) 
                for p in self._cache[(maxD, None)]])
            log(DEBUG2, 1, "Deriving from %d paths." % len(newpaths))

            timer = Timer()
            log(DEBUG2, 1, "Filtering for maximal relevant paths...")
            maxpaths = set()
            self._add_to_maxpaths(newpaths, maxpaths, D)
            log(DEBUG2, 1, 'Done after %.2f s. %d paths left.' % (timer.stop(),
                len(maxpaths)))

            log(DEBUG3, 1, "Resulting new set for D=%d:" % D)
            for i, p in enumerate(maxpaths):
                p.set_name("%s_D%d_p%d" % (str(self._task), D, i))
                log(DEBUG3, 2, "Path %s." % str(p))
            self._cache[(D, None)] = maxpaths
        else:
            log(DEBUG2, 0, "Using cached concrete paths for task %s for D=%d..." % 
                (str(self._task), D))
        return self._cache[(D, None)]



class RfPathGenerator(PathGenerator):

    def __init__(self, task, maxD, 
            path_class = None,
            abstract_path_class = None,
            store_class = None):
        self._task = task
        self._maxD = maxD
        self._path_class = path_class
        self._abstract_path_class = abstract_path_class
        self._store_class = store_class
        if self._path_class is None:
            self._path_class = FullRfPath
        if self._abstract_path_class is None:
            self._abstract_path_class = AbstractRfPath
        if self._store_class is None:
            self._store_class = DiscardingRfStore

        def rfs_max(paths, maxD):
            return [max([p.get_rf()[i] for p in paths]) for i in range(maxD)]

        return super(RfPathGenerator, self).__init__(
            task = self._task,
            maxD = self._maxD,
            path_class = self._path_class,
            abstract_path_class = self._abstract_path_class,
            store_class = self._store_class,
            start_jobs = self._task.get_jobs(),
            backwards = False,
            paths_max = rfs_max,
            add_to_maxpaths = add_to_maxpaths_rf,
            pathtype_str = "rf function")

    def prepare(self):

        D = self._maxD
        cindex = self._task
        if cindex in GLOBAL_CACHE["D"]:
            gcD = GLOBAL_CACHE["D"][cindex]
            if gcD >= D:
                self._cache[(gcD, None)] = GLOBAL_CACHE[cindex]
                log(DEBUG, 0, "CACHE HIT: %ss for task %s on domain %d..." % 
                    (self._pathtype_str, str(self._task), D))
                self._maxD = gcD
                return

        res = super(RfPathGenerator, self).prepare()

        if not GLOBAL_CACHE["deactivate"]:
            GLOBAL_CACHE["D"][cindex] = D
            GLOBAL_CACHE[cindex] = self._cache[(D, None)]

        return res

 

class RfSfxPathGenerator(PathGenerator):

    def __init__(self, task, end_vertex, maxD, 
            path_class = None,
            abstract_path_class = None,
            store_class = None):
        self._task = task
        self._end_vertex = end_vertex
        self._maxD = maxD
        self._path_class = path_class
        self._abstract_path_class = abstract_path_class
        self._store_class = store_class
        if self._path_class is None:
            self._path_class = FullRfReversePath
        if self._abstract_path_class is None:
            self._abstract_path_class = AbstractRfReversePath
        if self._store_class is None:
            self._store_class = DiscardingRfReverseStore

        def rfreverses_max(paths, maxD):
            return [max([p.get_rfreverse()[i] for p in paths]) 
                for i in range(maxD)]

        return super(RfSfxPathGenerator, self).__init__(
            task = self._task,
            maxD = self._maxD,
            path_class = self._path_class,
            abstract_path_class = self._abstract_path_class,
            store_class = self._store_class,
            start_jobs = [self._end_vertex],
            backwards = True,
            paths_max = rfreverses_max,
            add_to_maxpaths = add_to_maxpaths_rfreverse,
            pathtype_str = "rfsfx function")
            
    def prepare(self):

        D = self._maxD
        cindex = (self._task, self._end_vertex)
        if cindex in GLOBAL_CACHE["D"]:
            gcD = GLOBAL_CACHE["D"][cindex]
            if gcD >= D:
                self._cache[(gcD, None)] = GLOBAL_CACHE[cindex]
                log(DEBUG, 0, "CACHE HIT: %ss for (task %s, v %s) for D=%d..." % 
                    (self._pathtype_str, str(self._task),
                    str(self._end_vertex), D))
                self._maxD = gcD
                return

        res = super(RfSfxPathGenerator, self).prepare()

        if not GLOBAL_CACHE["deactivate"]:
            GLOBAL_CACHE["D"][cindex] = D
            GLOBAL_CACHE[cindex] = self._cache[(D, None)]

        return res



def add_to_maxpaths_rf(paths, maxpaths, D):
    for path in paths:

        log(DEBUG3, 2, "Processing: %s" % path)

        # Check whether too short
        if path.get_horizon() < D:
            log(DEBUG3, 3, "Too short.")
            continue

        # Check whether we are dominated by some processed path
        dominated = False
        new_rf = path.get_rf()
        for p in maxpaths:
            rf = p.get_rf()
            dominated = all([rf[i] >= new_rf[i] for i in range(D)])
            if dominated:
                log(DEBUG3, 3, "Dominated by %s." % p)
                break
        if dominated:
            continue
        log(DEBUG3, 3, "New. Will add.")

        # Remove all paths dominated by us
        remove_paths = set()
        for p in maxpaths:
            rf = p.get_rf()
            dominated = all([rf[i] <= new_rf[i] for i in range(D)])
            if dominated:
                log(DEBUG3, 4, "Dominates: %s" % p)
                remove_paths.add(p)
        maxpaths.difference_update(remove_paths)

        # Done. Add finally.
        maxpaths.add(path)


def add_to_maxpaths_rfreverse(paths, maxpaths, D):
    for path in paths:

        log(DEBUG3, 2, "Processing: %s" % path)

        # Check whether too short
        if path.get_horizon() < D:
            log(DEBUG3, 3, "Too short.")
            continue

        # Check whether we are dominated by some processed path
        dominated = False
        new_rfreverse = path.get_rfreverse()
        for p in maxpaths:
            rfreverse = p.get_rfreverse()
            dominated = all([rfreverse[i] >= new_rfreverse[i] 
                for i in range(D)])
            if dominated:
                log(DEBUG3, 3, "Dominated by %s." % p)
                break
        if dominated:
            continue
        log(DEBUG3, 3, "New. Will add.")

        # Remove all paths dominated by us
        remove_paths = set()
        for p in maxpaths:
            rfreverse = p.get_rfreverse()
            dominated = all([rfreverse[i] <= new_rfreverse[i] 
                for i in range(D)])
            if dominated:
                log(DEBUG3, 4, "Dominates: %s" % p)
                remove_paths.add(p)
        maxpaths.difference_update(remove_paths)

        # Done. Add finally.
        maxpaths.add(path)


def is_sjp_schedulable(TS):
    """Shortcut for testing the given taskset's SJP schedulability
    with given task priority order.
    """
    A = SJPSchedulabilityAnalyzer(TS)
    A.run()
    return A.get_result()

def is_sjp_feasible(TS):
    """Shortcut for testing the given taskset's SJP feasibility.
    """
    A = SJPFeasibilityAnalyzer(TS)
    A.run()
    return A.get_result()




# vim:textwidth=79:shiftwidth=4:tabstop=4:cc=80:expandtab:autoindent

