# This file is part of libdrt.
#
# Written 2011-2014 by Martin Stigge, Pontus Ekberg.
#
# GPLv3 applies.


from drt.util.logging import ERROR, INFO, DEBUG, DEBUG2, DEBUG3, log, indent_log


class ExplorationEngine(object):

    """The ExplorationEngine generates all paths in a Task (using the
    abstracton of a given Path class) and stores them in a given
    path store.
    """

    def __init__(self, task, store, path_creator, 
            start_jobs = None, backwards = False):
        """path_creator is a function taking a list of jobs and returning
        a path object. This can be a class (so the function is implicitly the 
        constructor) or a wrapper function.

        start_jobs: Which jobs to start at. (Default: All.)
        backwards: If True, extend paths in the front. (Default: False)
        """
        self.task = task
        self.store = store

        # TODO: path_creator is only used for creating initial paths from
        # start_jobs. Better would be to just supply the initial set of
        # paths with default to all 0-paths created from all jobs via
        # some default (or supplied?) path class
        self.path_creator = path_creator
        if start_jobs is None:
            self.start_jobs = task.get_jobs()
        else:
            self.start_jobs = start_jobs
        self.backwards = backwards

    def run(self):
        """Run the ExplorationEngine until completion."""
        log(DEBUG, 0, 'ExplorationEngine starting...')

        # Add all vertices as paths of length 0.
        log(DEBUG2, 1, 'Adding 0-paths to path store...')
        paths = [self.path_creator([job]) for job in self.start_jobs]
        indent_log(2)
        self.store.add_paths(paths)
        indent_log(-2)
        log(DEBUG2, 1, 'Added %d 0-paths to store.' % 
            self.store.get_fresh_num())

        # Build the rest of the paths for the path store one by one.
        log(DEBUG2, 1, 'Generating k-paths...')
        while True:
            log(DEBUG3, 2, 'Currently %d stored, %d fresh.' %
                (self.store.get_stored_num(), self.store.get_fresh_num()))
            path = self.store.pop_fresh()
            if path is None:
                break
            log(DEBUG3, 3, 'Extending %s...' % path)
            if self.backwards:
                u = path.get_start()
                extension_vertices = self.task.get_preds(u)
            else:
                u = path.get_end()
                extension_vertices = self.task.get_succs(u)
            for v in extension_vertices:
                log(DEBUG3, 3, '... with %s. Candidates:' % v)
                indent_log(4)
                ext_paths = path.get_extended(v)
                self.store.add_paths(ext_paths)
                indent_log(-4)
                # * What if we haven't stored any extension, shouldn't
                # the original path be stored then? We might lose a critical
                # path otherwise. (?)
                #   - Well, if all extensions are non-critical, how can the
                #   path itself be critical?
                #     - Depends on what type of path abstraction we are 
                #       deadling with. This is regulated via the store. 
                #       -> For demand triples, the pop_fresh() method always 
                #       saves the popped path directly in the store, 
                #       so it's preserved.
                #       -> For rfpaths, it doesn't since there is no reason to
                #       do so.

        log(DEBUG, 0, 'Done generating paths, %d paths stored in total' %
            self.store.get_stored_num())


class IncrementalExplorationEngine(object):

    """The IncrementalExplorationEngine generates all paths in a Task (using the
    abstracton of a given Path class) and stores them in a given
    path store. The run(increments) method yields after every increments steps
    with the guarantee that all paths up to this time-length are
    already created in the store.
    """

    def __init__(self, task, store, path_creator):
        """path_creator is a function taking a list of jobs and returning
        a path object. This can be a class (so the function is implicitly the 
        constructor) or a wrapper function.

        Note: store needs to support store.current_limit and is expected to
        pop paths in suitable increasing order.
        """

        self.task = task
        self.store = store
        self.path_creator = path_creator

    def run(self, increments):
        """Run the ExplorationEngine for increments steps at a time."""
        log(DEBUG2, 0, 'ExplorationEngine starting...')

        # Add all vertices as paths of length 0.
        log(DEBUG2, 1, 'Adding 0-paths to path store...')
        paths = [self.path_creator([job]) for job in self.task.get_jobs()]
        indent_log(2)
        self.store.add_paths(paths)
        indent_log(-2)
        log(DEBUG2, 2, 'Added %d 0-paths to store.' % 
            self.store.get_fresh_num())

        # Build the rest of the paths for the path store one by one.
        log(DEBUG2, 1, 'Generating k-paths...')
        current_limit = increments
        while True:
            log(DEBUG2, 2, 'Currently %d stored, %d fresh.' %
                (self.store.get_stored_num(), self.store.get_fresh_num()))
            if self.store.current_limit > current_limit:
                log(DEBUG, 2, "Store is complete up to %d (>%d). Yielding." %
                    (self.store.current_limit, current_limit))
                yield
                current_limit += increments
            path = self.store.pop_fresh()
            if path is None:
                break
            log(DEBUG3, 2, 'Extending %s...' % path)
            u = path.get_end()
            for v in self.task.get_succs(u):
                log(DEBUG3, 3, '... with %s. Candidates:' % v)
                indent_log(4)
                ext_paths = path.get_extended(v)
                self.store.add_paths(ext_paths)
                indent_log(-4)

        log(DEBUG2, 1, 'Done generating paths, %d paths stored in total' %
            self.store.get_stored_num())

# vim:textwidth=79:shiftwidth=4:tabstop=4:cc=80:expandtab:autoindent

