# This file is part of libdrt.
#
# Written 2013-2014 by Martin Stigge.
#
# GPLv3 applies.


"""
Path abstractions for dbf computations.
"""

from drt.util.logging import log, indent_log
from drt.util.logging import ERROR, INFO, DEBUG, DEBUG2, DEBUG3, DEBUG4
from drt.analysis.util.paths import Path
from drt.analysis.util.pathstores import PathStore


class RfPath(Path):
    
    """Template for path abstraction classes used to do fp analysis."""

    def get_rf(self):
        """Return request function of that path."""
        raise NotImplementedError

    def get_horizon(self):
        """Return rf horizon, i.e., how far the function is defined."""
        raise NotImplementedError

    def distance_to(self, o):
        """Return distance to path o. Distance is pointwise distance of
        request functions with exponentially decreasing weight."""
        rf1 = self.get_rf()
        rf2 = o.get_rf()
        base = 0.01**(1.0/self.get_horizon())
        return sum([abs(rf1[i] - rf2[i])*(0.8**i)
            for i in xrange(self.get_horizon())])


class FullRfPath(RfPath):
    """Stores entire concrete paths (i.e., it performs no
    abstractions at all). Domination relation on rf.
    
    max_horizon is an optional parameter used for domination relation.
    (Size of maximum rf prefix to be considered.)
    """

    def __init__(self, jobs, max_horizon = None, rf = None, horizon = None):
        super(FullRfPath, self).__init__(jobs)
        self.max_horizon = max_horizon
        self.rf = rf or self.recompute_rf()
        self.horizon = horizon or self.recompute_horizon()
        if self.horizon > self.max_horizon:
            self.horizon = self.max_horizon
        # Invariant: self.horizon <= self.max_horizon

    def is_concrete(self):
        return True

    def get_end(self):
        return self.jobs[-1]

    def get_extended(self, job):
        task = job.get_task()
#        return [FullRfPath(self.jobs[:] + [job], max_horizon = self.max_horizon)]
        newjobs = self.jobs[:] + [job]
        # TODO: Using self.rf[-1] assumes non-zero edge labels.
        newrf = (self.rf[:] + 
            [self.rf[-1]] * (task.get_delay((self.jobs[-1], job)) 
                - task.get_mindelay(self.jobs[-1])) +
            [self.rf[-1] + job.get_e()] * task.get_mindelay(job)
            )[:self.max_horizon]
        newhorizon = min(self.horizon + 
            task.get_delay((self.jobs[-1], job)) -
            task.get_mindelay(self.jobs[-1]) +
            task.get_mindelay(job), 
            self.max_horizon)
        return [FullRfPath(newjobs, max_horizon = self.max_horizon,
            rf = newrf, horizon = newhorizon)]
        
    def recompute_rf(self):
        task = self.jobs[0].get_task()
        log(DEBUG4, 1, "Recomputing RF.")
        rf = list()
        c = 0
        for (i, v) in enumerate(self.jobs):
            c += v.get_e()
            if i == len(self.jobs) - 1:
                rf.extend([c] * int(min(task.get_mindelay(v),
                                        self.max_horizon)))
            else:
                rf.extend([c] * task.get_delay((v, self.jobs[i + 1])))
        return rf[:self.max_horizon]

    def get_rf(self):
        return self.rf

    def recompute_horizon(self):   
        task = self.jobs[0].get_task()
        log(DEBUG4, 1, "Recomputing horizon.")
        # TODO: Why not just return min(len(self.rfreverse, self.max_horizon))
        h = 0
        for (i, v) in enumerate(self.jobs):
            if i == len(self.jobs) - 1:
                h += task.get_mindelay(v)
            else:
                h += task.get_delay((v, self.jobs[i + 1]))
        return min(h, self.max_horizon)

    def get_horizon(self):
        return self.horizon

    def __len__(self):
        return len(self.jobs) - 1

    def __ge__(self, o):
        # A (weakly) dominates B iff (1) OR (2) with
        # (1a) A.endvertex == B.endvertex, and
        # (1b) A.rfendpoint dominates B.rfendpoint, and
        # (1c) A >= B in all points up to A.horizon;
        # (2a) B.horizon >= max_horizon, and
        # (2b) A.endpoint (latest max_horizon-1) at least like B[max_horizon-1], and
        # (2c) A >= B in all points up to min(A.horizon,max_horizon).
        # Assumption: o.max_horizon >= self.max_horizon
        #H = min(self.get_horizon(), self.max_horizon)
        H = self.get_horizon()
        condition1 = (self.get_end() == o.get_end() and
            self.get_horizon() <= o.get_horizon() and
            self.get_rf()[-1] >= o.get_rf()[-1] and
            all([self.get_rf()[i] >= o.get_rf()[i] for i in xrange(H)]))
        condition2 = (self.max_horizon is not None and
            o.get_horizon() >= self.max_horizon and
            self.get_rf()[H-1] >= o.get_rf()[self.max_horizon-1] and
            all([self.get_rf()[i] >= o.get_rf()[i] for i in xrange(H)]))
        return condition1 or condition2
            
    def __gt__(self, o):
        # Like __ge__ except that at SOME point, A and B differ.
        #H = min(self.get_horizon(), self.max_horizon, o.get_horizon())
        H = min(self.get_horizon(), o.get_horizon())
        return (self >= o and
            not all([self.get_rf()[i] == o.get_rf()[i] for i in xrange(H)]))

    def __str__(self):
        return (self._name + 
            ' [' + ' -> '.join(map(str, self.jobs)) + ']' + ', ' + 
            str(self.get_rf()) + ", " + str(len(self.get_rf())))


class AbstractRfPath(RfPath):

    """Stores abstract rf paths, i.e., just rf.
    """

    def __init__(self, max_horizon = None, rf = None, horizon = None,
            childs = None, spread = None):
        if rf is None:
            raise RuntimeError("AbstractRfPath needs rf to be not None!")
        super(AbstractRfPath, self).__init__([])
        self.max_horizon = max_horizon
        self.rf = rf
        self.horizon = horizon or len(rf)
        self.childs = childs
        self.spread = spread

    def is_concrete(self):
        return False

    def get_jobs(self):
        raise RuntimeError("AbstractRfPath doesn't have jobs.")

    def get_end(self):
        raise RuntimeError("AbstractRfPath doesn't have jobs.")

    def get_extended(self, job):
        raise RuntimeError("AbstractRfPath doesn't have jobs.")
        
    def recompute_rf(self):
        raise RuntimeError("AbstractRfPath doesn't have jobs.")

    def get_rf(self):
        return self.rf

    def recompute_horizon(self):   
        raise RuntimeError("AbstractRfPath doesn't have jobs.")
         
    def get_horizon(self):
        return self.horizon

    def __len__(self):
        raise RuntimeError("AbstractRfPath doesn't have jobs.")

    def __ge__(self, o):
        # Abstract path A (weakly) dominates B iff (2) with
        # (2a) B.horizon >= max_horizon, and
        # (2b) A.endpoint (latest max_horizon-1) at least like B[max_horizon-1], and
        # (2c) A >= B in all points up to min(A.horizon,max_horizon).
        # Assumption: o.max_horizon >= self.max_horizon
        #H = min(self.get_horizon(), self.max_horizon)
        H = self.get_horizon()
        condition2 = (self.max_horizon is not None and
            o.get_horizon() >= self.max_horizon and
            self.get_rf()[min(H, self.max_horizon)-1] >= o.get_rf()[self.max_horizon-1] and
            all([self.get_rf()[i] >= o.get_rf()[i] for i in xrange(min(H, self.max_horizon))]))
        return condition2
            
    def __gt__(self, o):
        # Like __ge__ except that at SOME point, A and B differ.
        #H = min(self.get_horizon(), self.max_horizon, o.get_horizon())
        H = min(self.get_horizon(), o.get_horizon())
        return (self >= o and
            not all([self.get_rf()[i] == o.get_rf()[i] for i in xrange(H)]))

    def __str__(self):
        return (self._name + 
            '[ABSTRACT]' + ', ' + 
            str(self.get_rf()) + ", " + str(len(self.get_rf())))


class RfPathStore(PathStore):
    
    """A basic PathStore for Paths in rf computations. Only discards paths
    that have a horizon of at least D.
    """

    def __init__(self, D):
        self.D = D
        super(RfPathStore, self).__init__()

    def add_path(self, path):
        """Add the path to the store.
        If path does not exceed horizon yet, store it in fresh set.
        """
        if self.is_todo(path):
            self.fresh.add(path)
            log(DEBUG3, 0, "Stored todo %s." % str(path))
        else:
            self.stored.add(path)
            log(DEBUG3, 0, "Stored non-todo %s." % str(path))


    def is_todo(self, path):
        return path.get_horizon() < self.D



class DiscardingRfStore(RfPathStore):
    
    """A basic PathStore for Paths in rf computations. Discard all
    dominated paths.
    """

    def __init__(self, D):
        self.D = D
        self.stored = set()
        self.fresh = set()
        super(DiscardingRfStore, self).__init__(D)

    def pop_fresh(self):
        """Return one path on the 'frontier' of the explored paths, i.e.,
        one that has not previously been extended. Afterwards the path is
        discarded!
        """
        if len(self.fresh) > 0:
            path = self.fresh.pop()
            #self.stored.add(path)
            return path
        return None


    def add_path(self, path):
        """Add the path to the store.
        If path does not exceed horizon yet, store it in fresh set.
        If we are dominated by stored path, discard us.
        If already stored path is dominated, remove it.
        """
        # Check if path is dominated by a stored path
        dominated = False
        for p in self.fresh.union(self.stored):
            dominated = (p > path)
            if dominated:
                log(DEBUG2, 0, "Not storing %s." % str(path))
                log(DEBUG3, 1, "Dominated by already stored %s." % str(p))
                break
        if dominated:
            return

        # Remove all paths dominated by path
        # .. in fresh
        for p in tuple(self.fresh):
            log(DEBUG3, 1, "Checking for dominance in fresh: %s." % str(p))
            if path >= p:
                log(DEBUG3, 1, "Remove from fresh now dominated %s." % str(p))
                self.fresh.remove(p)

        # .. in stored
        for p in tuple(self.stored):
            log(DEBUG3, 1, "Checking for dominance in stored: %s." % str(p))
            if path >= p:
                log(DEBUG3, 1, "Remove from stored now dominated %s." % str(p))
                self.stored.remove(p)

        if self.is_todo(path):
            self.fresh.add(path)
            log(DEBUG2, 0, "Store in todo: %s" % str(path))
        else:
            self.stored.add(path)
            log(DEBUG2, 0, "Store in non-todo: %s" % str(path))


    def is_todo(self, path):
        return path.get_horizon() < self.D


# vim:textwidth=79:shiftwidth=4:tabstop=4:cc=80:expandtab:autoindent

