# This file is part of libdrt.
#
# Written 2013-2014 by Martin Stigge.
#
# GPLv3 applies.


"""
Path abstractions for dbf computations.
"""

from drt.util.logging import log, indent_log
from drt.util.logging import ERROR, INFO, DEBUG, DEBUG2, DEBUG3, DEBUG4
from drt.analysis.util.paths import Path
from drt.analysis.util.pathstores import PathStore
from drt.analysis.util.rfpaths import DiscardingRfStore


class RfReversePath(Path):
    
    """Template for path abstraction classes used to do fp analysis."""

    def get_rfreverse(self):
        """Return reverse request function of that path."""
        raise NotImplementedError

    def get_horizon(self):
        """Return rf horizon, i.e., how far the function is defined."""
        raise NotImplementedError

    def distance_to(self, o):
        """Return distance to path o. Distance is pointwise distance of
        request functions with exponentially decreasing weight."""
        rf1 = self.get_rfreverse()
        rf2 = o.get_rfreverse()
        base = 0.01**(1.0/self.get_horizon())
        return sum([abs(rf1[i] - rf2[i])*(0.8**i)
            for i in xrange(self.get_horizon())])


class FullRfReversePath(RfReversePath):
    """Stores entire concrete paths (i.e., it performs no
    abstractions at all). Domination relation on rfreverse.
    
    max_horizon is an optional parameter used for domination relation.
    (Size of maximum rf prefix to be considered.)
    """

    def __init__(self, jobs, max_horizon = None, rfreverse = None, 
            horizon = None):
        super(FullRfReversePath, self).__init__(jobs)
        self.max_horizon = max_horizon
        self.rfreverse = rfreverse or self.recompute_rfreverse()
        self.horizon = horizon or self.recompute_horizon()
        if self.horizon > self.max_horizon:
            self.horizon = self.max_horizon
        # Invariant: self.horizon <= self.max_horizon

    def is_concrete(self):
        return True

    def get_start(self):
        return self.jobs[0]

    def get_extended(self, job):
        task = job.get_task()
        newjobs = [job] + self.jobs[:]
        newrfreverse = (self.rfreverse[:] + 
            [self.rfreverse[-1]] * (task.get_delay((job, self.jobs[0])) 
                - task.get_minprevdelay(self.jobs[0])) +
            [self.rfreverse[-1] + job.get_e()] * task.get_minprevdelay(job)
            )[:self.max_horizon]
        newhorizon = min(len(newrfreverse), self.max_horizon)
        return [FullRfReversePath(newjobs, max_horizon = self.max_horizon,
            rfreverse = newrfreverse, horizon = newhorizon)]
        
    def recompute_rfreverse(self):
        task = self.jobs[0].get_task()
        log(DEBUG4, 1, "Recomputing RfReverse.")
        rfreverse = list()
        c = 0
        for (i, v) in reversed(list(enumerate(self.jobs))):
            c += v.get_e()
            if i > 0:
                rfreverse.extend([c] * task.get_delay((self.jobs[i - 1], v)))
            else:
                rfreverse.extend([c] * task.get_minprevdelay(v))
        return rfreverse[:self.max_horizon]

    def get_rfreverse(self):
        return self.rfreverse

    def recompute_horizon(self):   
        task = self.jobs[0].get_task()
        log(DEBUG4, 1, "Recomputing horizon.")
        return min(len(self.rfreverse), self.max_horizon)
         
    def get_horizon(self):
        return self.horizon

    def __len__(self):
        return len(self.jobs) - 1

    def __ge__(self, o):
        # A (weakly) dominates B iff (1) OR (2) with
        # (1a) A.endvertex == B.endvertex, and
        # (1b) A.rfendpoint dominates B.rfendpoint, and
        # (1c) A >= B in all points up to A.horizon;
        # (2a) B.horizon >= max_horizon, and
        # (2b) A.endpoint (latest max_horizon-1) at least like B[max_horizon-1], and
        # (2c) A >= B in all points up to min(A.horizon,max_horizon).
        # Assumption: o.max_horizon >= self.max_horizon
        #H = min(self.get_horizon(), self.max_horizon)
        H = self.get_horizon()
        condition1 = (self.get_start() == o.get_start() and
            self.get_horizon() <= o.get_horizon() and
            self.get_rfreverse()[-1] >= o.get_rfreverse()[-1] and
            all([self.get_rfreverse()[i] >= o.get_rfreverse()[i] 
                for i in xrange(H)]))
        condition2 = (self.max_horizon is not None and
            o.get_horizon() >= self.max_horizon and
            self.get_rfreverse()[H-1] >= o.get_rfreverse()[self.max_horizon-1] 
            and all([self.get_rfreverse()[i] >= o.get_rfreverse()[i] 
                for i in xrange(H)]))
        return condition1 or condition2
            
    def __gt__(self, o):
        # Like __ge__ except that at SOME point, A and B differ.
        #H = min(self.get_horizon(), self.max_horizon, o.get_horizon())
        H = min(self.get_horizon(), o.get_horizon())
        return (self >= o and
            not all([self.get_rfreverse()[i] == o.get_rfreverse()[i] 
                for i in xrange(H)]))

    def __str__(self):
        return (self._name + 
            ' [' + ' -> '.join(map(str, self.jobs)) + ']' + ', ' + 
            str(self.get_rfreverse()) + ", " + str(len(self.get_rfreverse())))


class AbstractRfReversePath(RfReversePath):

    """Stores abstract rfreverse paths, i.e., just rfreverse.
    """

    def __init__(self, max_horizon = None, rfreverse = None, horizon = None,
            childs = None, spread = None):
        if rfreverse is None:
            raise RuntimeError(("AbstractRfReversePath needs rfreverse to be" + 
                " not None!"))
        super(AbstractRfReversePath, self).__init__([])
        self.max_horizon = max_horizon
        self.rfreverse = rfreverse
        self.horizon = horizon or len(rfreverse)
        self.childs = childs
        self.spread = spread

    def is_concrete(self):
        return False

    def get_jobs(self):
        raise RuntimeError("AbstractRfReversePath doesn't have jobs.")

    def get_start(self):
        raise RuntimeError("AbstractRfReversePath doesn't have jobs.")

    def get_extended(self, job):
        raise RuntimeError("AbstractRfReversePath doesn't have jobs.")
        
    def recompute_rfreverse(self):
        raise RuntimeError("AbstractRfReversePath doesn't have jobs.")

    def get_rfreverse(self):
        return self.rfreverse

    def recompute_horizon(self):   
        raise RuntimeError("AbstractRfReversePath doesn't have jobs.")
         
    def get_horizon(self):
        return self.horizon

    def __len__(self):
        raise RuntimeError("AbstractRfReversePath doesn't have jobs.")

    def __ge__(self, o):
        # Abstract path A (weakly) dominates B iff (2) with
        # (2a) B.horizon >= max_horizon, and
        # (2b) A.endpoint (latest max_horizon-1) at least like B[max_horizon-1], and
        # (2c) A >= B in all points up to min(A.horizon,max_horizon).
        # Assumption: o.max_horizon >= self.max_horizon
        #H = min(self.get_horizon(), self.max_horizon)
        H = self.get_horizon()
        condition2 = (self.max_horizon is not None and
            o.get_horizon() >= self.max_horizon and
            self.get_rfreverse()[min(H, self.max_horizon)-1] >= 
                o.get_rfreverse()[self.max_horizon-1] and
            all([self.get_rfreverse()[i] >= o.get_rfreverse()[i] 
                for i in xrange(min(H, self.max_horizon))]))
        return condition2
            
    def __gt__(self, o):
        # Like __ge__ except that at SOME point, A and B differ.
        #H = min(self.get_horizon(), self.max_horizon, o.get_horizon())
        H = min(self.get_horizon(), o.get_horizon())
        return (self >= o and
            not all([self.get_rfreverse()[i] == o.get_rfreverse()[i] 
                for i in xrange(H)]))

    def __str__(self):
        return (self._name + 
            '[ABSTRACT]' + ', ' + 
            str(self.get_rfreverse()) + ", " + str(len(self.get_rfreverse())))



class DiscardingRfReverseStore(DiscardingRfStore):
    
    """A basic PathStore for Paths in rfreverse computations. Discard all
    dominated paths. Identical to DiscardingRfStore.
    """

    def __init__(self, D):
        super(DiscardingRfReverseStore, self).__init__(D)


# vim:textwidth=79:shiftwidth=4:tabstop=4:cc=80:expandtab:autoindent

