# This file is part of libdrt.
#
# Written 2011-2014 by Martin Stigge, Pontus Ekberg.
#
# GPLv3 applies.


#from __future__ import absolute_import

from drt.analysis.util.upaths import UtilizationTriple
from drt.analysis.util.pathstores import DiscardingUtilizationStore
from drt.analysis.util.explorationengines import *

from drt.util.statistics import Timer, stats

from drt.util.logging import ERROR, INFO, DEBUG, DEBUG2, DEBUG3, log, indent_log


class UEngine(object):
    """Computes the utilization of a task. For using an instance, first
    use the run() method and retrieve the result using get_U().
    Custom store_class and path_class can be provided which default
    to the most optimized variants available.
    """


    def __init__(self, task, store_class = None, path_class = None):
        self._task = task
        self._U = None
        self._store_class = store_class
        self._path_class = path_class
        if self._path_class is None:
            self._path_class = UtilizationTriple
        if self._store_class is None:
            self._store_class = DiscardingUtilizationStore

    def run(self):
        if self._U is not None:
            return

        log(INFO, 0, "Computing utilization for task '%s'..." %
            str(self._task))
        log(DEBUG, 1, 'Path class:  %s' % self._path_class.__name__)
        log(DEBUG, 1, 'Store class: %s' % self._store_class.__name__)

        store = self._store_class(len(self._task.get_jobs()))

        indent_log(1)
        ee = ExplorationEngine(self._task, store, self._path_class)
        ee.run()
        indent_log(-1)
        
        # Check cycles and compute U.
        log(DEBUG, 1, 'Computing utilization from paths...')
        e2_max = 0 # Start with dummy cycle (U = 0).
        p_max = 1
        witness = None

        for path in store.get_paths():
            # Check only cycles.
            if path.get_start() == path.get_end() and len(path) > 0:
                e2 = path.get_e2()
                p = path.get_p()
                log(DEBUG, 2, 'Found cycle %s with <%d, %d>.' % (path, e2, p))
      
                # Test if a new most dense cycle is found (avoid rounding).
                if e2 * p_max > e2_max * p:
                    e2_max = e2
                    p_max = p
                    witness = path
                    log(DEBUG, 2, '-> New candidate (U = %.5f): %s' % 
                        (float(e2_max) / p_max, witness))

        if witness is None:
            self._U = float(0)
            log(INFO, 1, "No cycle found. U = 0.")
        else:
            self._U = float(e2_max) / p_max
            log(INFO, 1, 'U = %.5f via %s' % (self._U, witness))

    def get_U(self):
        return self._U


def compute_utilization(TS):
    """Shortcut for computing the given taskset's utilization.
    """

    U = float(0)
    for T in TS.get_tasks():
        ue = UEngine(T)
        ue.run()
        U += ue.get_U()
    return U


# vim:textwidth=79:shiftwidth=4:tabstop=4:cc=80:expandtab:autoindent

