# This file is part of libdrt.
#
# Written 2011-2014 by Martin Stigge, Pontus Ekberg.
#
# GPLv3 applies.

import random

from drt.model import TaskSet, SPTaskSet, SJPTaskSet, DRTTask, Job

class rtas11(DRTTask):
    """
    Example Task from first original DRT paper "The Digraph Real-Time
    Task Model", Martin Stigge, Pontus Ekberg, Nan Guan and Wang Yi, RTAS'11
    """

    def __init__(self, name="RTAS11-Task"):

        v1 = Job(2, 5, name='v1')
        v2 = Job(1, 8, name='v2')
        v3 = Job(3, 8, name='v3')
        v4 = Job(5, 10, name='v4')
        v5 = Job(1, 5, name='v5')

        jobs = [v1, v2, v3, v4, v5]

        edges = dict()
        edges[(v1, v2)] = 10
        edges[(v1, v5)] = 20
        edges[(v2, v3)] = 15
        edges[(v2, v4)] = 20
        edges[(v3, v1)] = 11
        edges[(v4, v2)] = 20
        edges[(v5, v4)] = 10

        return super(rtas11, self).__init__(jobs, 
            edges=edges, name=name)


class rtas11_arbitrary(DRTTask):
    """
    Example Task with arbitrary deadlines from first original DRT paper "The
    Digraph Real-Time Task Model", Martin Stigge, Pontus Ekberg, Nan Guan and
    Wang Yi, RTAS'11
    """

    def __init__(self, name="RTAS11-Task"):

        v1 = Job(5, 7, name='v1')
        v2 = Job(1, 10, name='v2')
        v3 = Job(2, 5, name='v3')

        jobs = [v1, v2, v3]

        edges = dict()
        edges[(v1, v2)] = 8
        edges[(v2, v3)] = 3
        edges[(v3, v1)] = 1000

        return super(rtas11_arbitrary, self).__init__(jobs, 
            edges=edges, name=name)


class rtss13(DRTTask):
    """
    Example Task from SP Feasibility paper with combinatorial abstraction
    refinement: "Combinatorial Abstraction Refinement for Feasibility
    Analysis", Martin Stigge and Wang Yi, RTSS'13
    """

    def __init__(self, name="RTSS13-Task"):

        v1 = Job(6, 10, name='v1')
        v2 = Job(5, 25, name='v2')
        v3 = Job(1, 10, name='v3')
        v4 = Job(2, 12, name='v4')
        v5 = Job(10, 50, name='v5')

        jobs = [v1, v2, v3, v4, v5]

        edges = dict()
        edges[(v1, v2)] = 13
        edges[(v2, v3)] = 29
        edges[(v2, v4)] = 100
        edges[(v3, v4)] = 10
        edges[(v3, v5)] = 18
        edges[(v4, v1)] = 12
        edges[(v4, v2)] = 25
        edges[(v5, v5)] = 50

        return super(rtss13, self).__init__(jobs, 
            edges=edges, name=name)


class thesis_refinement(DRTTask):
    """
    Example Task from Martin Stigge's Ph.D. thesis, used in
    Refinement chapters.
    """

    def __init__(self, name="Thesis-Task"):

        v1 = Job(3, 10, name='v1')
        v2 = Job(4, 10, name='v2')
        v3 = Job(1,  3, name='v3')
        v4 = Job(6, 12, name='v4')
        v5 = Job(4,  8, name='v5')

        jobs = [v1, v2, v3, v4, v5]

        edges = dict()
        edges[(v1, v2)] = 10
        edges[(v1, v3)] = 15
        edges[(v2, v1)] = 15
        edges[(v2, v4)] = 17
        edges[(v3, v5)] = 30
        edges[(v4, v1)] = 12
        edges[(v5, v4)] = 8
        edges[(v5, v5)] = 20

        return super(thesis_refinement, self).__init__(jobs, 
            edges=edges, name=name)
        ldges[(v5, v4)] = 10


class interference_demo(DRTTask):
    """ This singleton task set causes interference of 12 in a 20 time units
    interval. Thus, interference_test(9, 20) will succeed, but reducing wcet
    to 8 will fail."""

    def __init__(self, name="I-Demo"):

        v1 = Job(1, 8, name="v1")
        v2 = Job(5, 7, name="v2")
        v3 = Job(2, 4, name="v3")

        jobs = [v1, v2, v3]

        edges = dict()
        edges[(v1, v2)] = 10
        edges[(v1, v3)] = 8
        edges[(v2, v3)] = 7
        edges[(v3, v3)] = 4
        edges[(v3, v1)] = 12

        return super(interference_demo, self).__init__(jobs,
            edges=edges, name=name)


class sp_feasibility_demo(SPTaskSet):
    """Example for fp schedulability and feasibility.
    In given priority order, task set is fp schedulable. (mode=1)
    In order [T1, T2, T4, T3] it is not. (mode=2)
    In any order, it is fp feasible.
    If you add a PeriodicTask(2, 25, 25, 'T5'), it is feasible but 
        not fp feasible. (mode=3)
    """

    def __init__(self, mode=1):

        T1 = PeriodicTask(1, 2, 10, name="T1", vertexpattern="u_%s")

        v1 = Job(1, 8, name="v1")
        v2 = Job(5, 7, name="v2")
        v3 = Job(2, 4, name="v3")
        edges = dict()
        edges[(v1, v2)] = 10
        edges[(v1, v3)] = 8
        edges[(v2, v3)] = 7
        edges[(v3, v3)] = 4
        edges[(v3, v1)] = 12
        T2 = DRTTask([v1, v2, v3], edges, "T2")

        w1 = Job(6, 20, name="w1")
        w2 = Job(1, 100, name="w2")
        edges = dict()
        edges[(w1, w2)] = 50
        edges[(w2, w1)] = 100
        T3 = DRTTask([w1, w2], edges, "T3")

        T4 = PeriodicTask(1, 100, 100, name="T4", vertexpattern="x_%s")
        
        T5 = PeriodicTask(2, 25, 25, name="T5", vertexpattern="y_%s")

        if mode == 1:
            tasks = [T1, T2, T3, T4]
        elif mode == 2:
            tasks = [T1, T2, T4, T3]
        elif mode == 3:
            tasks = [T1, T2, T4, T3, T5]

        return super(sp_feasibility_demo, self).__init__(tasks)


class precision_demo(SPTaskSet):
    """Example for effect of precision refinement.
    Fails with just rbf, succeeds after refinement.
    """

    def __init__(self):
        T1 = GMFTask([(1, 4, 4), (2, 6, 6), (4, 10, 10)], "T1")
        T2 = PeriodicTask(6, 12, 100, "T2", "x_%s")
        return super(precision_demo, self).__init__([T1, T2])


class PeriodicTask(DRTTask):

    def __init__(self, c, d, p, name="T", vertexpattern='v_%s'):
        v = Job(c, d, name = vertexpattern % name)
        edges = {(v, v): p}

        return super(PeriodicTask, self).__init__([v], 
            edges={(v, v): p}, name=name)


class GMFTask(DRTTask):
 
    def __init__(self, jtlist, name, namepattern='v_%d'):
        """
        Args:
            jtlist: List of elements (e, d, p) for a job type
        """

        vlist = []
        for (i, jt) in enumerate(jtlist):
            v = Job(jt[0], jt[1], name = namepattern % i)
            vlist.append((v, jt[2]))

        vertices = []
        edges = dict()
        for ((v1, delay), (v2, _)) in zip(vlist, vlist[1:] + vlist[:1]):
            vertices.append(v1)
            edges[(v1, v2)] = delay

        return super(GMFTask, self).__init__(vertices, edges, name=name)


class RandomTaskRTAS14(DRTTask):

    def __init__(self, size_min, size_max, fanout_min, fanout_max,
            p_min, p_max, ratioed_min, ratioed_max, connected=True, 
            d_implicit = False,
            name='RandomTask'):
        """New strategy:
        (1) Create all edges with uniformly random P
        (2) Create random D for all vertices, bounded by minimal outgoing p
        (3) Create random E for all vertices consistent with given ratio range

        Implicit: Deadlines implicitly constrained but random.
        """    
        size = random.randint(size_min, size_max)
        vertices = []
        edges = dict()

        # Create all job types without parameters for now
        fakejobs = []
        for i in xrange(size):
            fakejobs.append('%s_v%d' % (name, i + 1))
        
        # Connect job i to job i + 1 (to ensure fully connected graph).
        fakeedges = dict()
        if connected:
            for i in xrange(size):
                p = random.randint(p_min, p_max)
                fakeedges[(fakejobs[i], fakejobs[(i + 1) % size])] = p

        # Pick the rest of the successors randomly.
        for job in fakejobs:
            fanout = random.randint(fanout_min, fanout_max) - 1
            scount = 1 if connected else 0
            while fanout > 0 and len(fakejobs) > scount + 1:
                succ = random.choice(fakejobs)
                if (job, succ) in fakeedges:
                    continue
                p = random.randint(p_min, p_max)
                fakeedges[(job, succ)] = p
                scount += 1
                fanout -= 1

        # Set deadlines.
        deadlines = dict()
        for job in fakejobs:
            mindelay = min([fakeedges[(j1, j2)] for (j1, j2) in fakeedges 
                if j1 == job] + [p_max])
            if d_implicit:
                deadlines[job] = mindelay
            else:
                deadlines[job] = random.randint(mindelay/2, mindelay)

        # Set execution times.
        wcets = dict()
        for fj in fakejobs:
            wcets[fj] = random.randint(max(1, int(deadlines[fj] * ratioed_min)),
    	    	max(1, int(deadlines[fj] * ratioed_max)))

        # Create actual vertices
        realjobs = dict()
        for fj in fakejobs:
            v = Job(wcets[fj], deadlines[fj], name=fj)
            vertices.append(v)
            realjobs[fj] = v

        # Create actual edges
        for fe in fakeedges:
            e = (realjobs[fe[0]], realjobs[fe[1]])
            edges[e] = fakeedges[(fe[0], fe[1])]

        return super(RandomTaskRTAS14, self).__init__(vertices, edges, 
            name=name)


def gen_RTAS14(tid = 0, d_implicit = False):
    """Generates a randomly generated task, exactly as for RTAS14."""
    return RandomTaskRTAS14(size_min=5, size_max=10, 
        fanout_min=1, fanout_max=3, 
        p_min=100, p_max=300, 
        ratioed_min=0.001, ratioed_max=0.07,
#        d_implicit = d_implicit,
        name="T%d"%tid)


class RTS_SI_RTSS13_Example(SJPTaskSet):
    """
    Example for necessity of busy window extension.
    """

    def __init__(self):

        T1 = PeriodicTask(4, 10, 10, "T1", "x_%s")
        T2 = GMFTask([(5, 7, 7), (5, 10, 100)], "T2")

        return super(RTS_SI_RTSS13_Example, self).__init__([T1, T2])


class RTS_SI_RTSS13_Example2(SJPTaskSet):
    """
    Example for necessity of busy window extension. Feasible version.
    """

    def __init__(self):

        T1 = PeriodicTask(4, 10, 10, "T1", "x_%s")
        T2 = GMFTask([(5, 7, 7), (5, 11, 100)], "T2")

        return super(RTS_SI_RTSS13_Example2, self).__init__([T1, T2])


class non_minimal(DRTTask):
    """
    Example Task for trying minimizing of tasks.
    """

    def __init__(self, name="NM-Task"):

        v1 = Job(7,  8, name='v1')
        v2 = Job(20, 30, name='v2')
        v3 = Job(0, 10, name='v3')
        v4 = Job(0, 2, name='v4')

        jobs = [v1, v2, v3, v4]

        edges = dict()
        edges[(v1, v2)] = 8
        edges[(v2, v1)] = 60
        edges[(v2, v3)] = 30
        edges[(v3, v1)] = 20
        edges[(v3, v4)] = 10
        edges[(v4, v1)] = 2

        return super(non_minimal, self).__init__(jobs, 
            edges=edges, name=name)


class non_minimal2(DRTTask):
    """
    Another example task for trying minimizing of tasks.
    """

    def __init__(self, name="NM2-Task"):

        v0 = Job(0, 2, name='v0')
        v1 = Job(1, 10, name='v1')
        v2 = Job(2, 20, name='v2')
        v3 = Job(3, 30, name='v3')
        v4 = Job(4, 40, name='v4')
        v5 = Job(5, 50, name='v5')

        jobs = [v0, v1, v2, v3, v4, v5]

        edges = dict()
        edges[(v1, v0)] = 10
        edges[(v2, v0)] = 20
        edges[(v0, v3)] = 30
        edges[(v0, v4)] = 40
        edges[(v0, v5)] = 50

        return super(non_minimal2, self).__init__(jobs, 
            edges=edges, name=name)

def ecrts12_transform(s, B, gmf=False):
    """Given the 3-PARTITION instance (s, B), returns a DRT task set which is
    SP unschedulable (and infeasible) iff the input is a positive (True) or a
    negative (False) instance of 3-PARTITION. In fact, the produced task set is
    containing branching tasks, i.e., trees.

    Input details:
    * s is a list of natural numbers, representing element sizes of 3*m
      elements for some m.
    * B is a natural numer, representing the bound.

    Constraint: sum(s) must be m*B, that is, sum(s) == B*len(s)/3.

    (s, B) is a positive instance iff s can be partitioned in disjoined
    sublists P_1 to P_m such that sum(P_1) == B.
    
    The transformation follows the result in ECRTS'12 paper "Hardness Results
    for Static Priority Real-Time Scheduling", Martin Stigge and Wang Yi. Since
    3-PARTITION is NP-hard in the strong sense, this establishes coNP-hardness
    in the strong sense for the SP schedulability problem for DRT task systems.

    If gmf is True, the task set contains GMF instead of branching tasks.
    """

    assert(len(s) % 3 == 0)
    m = len(s) / 3
    assert(sum(s) == m*B)
    L = m**2 * B

    if gmf:
        tasks = list()
        for a in s:
            u = list()
            v = list()
            for j in range(m):
                u.append(Job(L, 3*m*L))
                v.append(Job((j+1) * a, 6*m*L))
            edges = dict()
            for j in range(m):
                edges[(u[j], v[j])] = 3*m*L + sum(k * B for k in range(j+1))
                edges[(v[j], u[(j+1)%m])] = 6*m*L
            tasks.append(DRTTask(u + v, edges))
        jlow = Job(1, 3*m*L + sum(j * B for j in range(m+1)))
        tasks.append(DRTTask([jlow], {(jlow, jlow): 6*m*L}))
    else:
        tasks = list()
        for a in s:
            u = Job(0, 0)
            v = list()
            edges = dict()
            for j in range(m):
                v.append(Job((j+1) * a, L))
                edges[(u, v[j])] = sum(k * B for k in range(j+1))
            tasks.append(DRTTask([u] + v, edges))
        tasks.append(DRTTask([Job(1, sum(j * B for j in range(m+1)))], dict()))

    return SPTaskSet(tasks)
        


# vim:textwidth=79:shiftwidth=4:tabstop=4:cc=80:expandtab:autoindent               

