import time
import itertools

import engine
# import path.fullpath
# import path.fullmarkedpath
import path.triple
# import path.quadruple

# import path.fullpathdiscard
import path.rffullpath

# import booleanformulas as bf

# Size of an exclusion constraint be to be considered, i.e., saved.
# This is size ABOVE number of tasks.
# Too low: Creates too many constraints that don't purge many combinations
# Too high: Overlooks too many interesting constraints, i.e., purging opportunities
xx_ec_threshold = 4
# Number of exclusion constraints to keep. The buffer is FIFO.
# Too low: Discards too many constraints, i.e., purging opportunities
# Too high: 
xx_ec_maxsize = 3
# Abstraction levels for (non-iterative) rf solver. 
# Legal values are just 0 and None. Method is only complete if None is in the list.
xx_abstraction_levels = [0, None]
#
xx_default_solver = 'rf_iterative'

# Record number of tested path combinations.
xx_stats_combinations_tested = 0
# Record number of skipped path combinations (because of exclusion constraint).
xx_stats_combinations_skipped = 0
# 
xx_stats_jobs_tested = 0
xx_stats_jobs_ncj = 0
#
xx_stats_al_testcounts = dict(zip(xx_abstraction_levels, 
    [0] * len(xx_abstraction_levels)))
xx_stats_al_nonzero_schedulable = 0

# List of triples (unschedulable, tested, total) for iterative interference test
xx_stats_combi = []
xx_stats_combi_MAXLEN = 1000000


class Job(object):

    """A basic representation of job types. Each job type has an
    execution time requirement, a deadline and any number of successor
    job types. Each successor is associated with a separation time
    (or period).
    """
    
    def __init__(self, E, D, name='J', task=None):
        self.E = E
        self.D = D
        self.name = name
        self.succs = {}
        self.mindelay = float('Inf')
    
    def add_succ(self, job, separation):
        self.succs[job] = separation
        self.mindelay = min(self.mindelay, separation)

    def get_e(self):
        return self.E

    def get_d(self):
        return self.D

    def get_name(self):
        return str(self.name)

    def get_mindelay(self):
        return self.mindelay

    def set_task(self, task):
        self.task = task

    def get_task(self):
        return self.task

    def __str__(self):
#        return str("%(name)s" % self.__dict__)
        return str("%(name)s(%(E)d, %(D)d)" % self.__dict__)

    def dominates(self, j2, strict=False):
        # Let's not get into the business of overloading < and <=
        # or rely on default behavior
        return (((self.E >= j2.E) and (self.D <= j2.D)) if not strict else
            self.dominates(j2) and ((self.E != j2.E) or (self.D != j2.D)))


class Task(object):

    """A Task is a graph of jobs (or job types), for which we can compute
    the utilization and the demand bound function (dbf).
    """

    def __init__(self, jobs, name='T'):
        """Create a TaskSet out of the collection of <jobs>, with an
        optional <name>. The path abstractions to be used can be specified, 
        otherwise the fullpath abstraction is used. (Not very abstract then.)
        """
        self.jobs = set(jobs)
        for j in jobs:
            j.set_task(self)
        self.name = name
        self.reset()

        # Determine if this is an arbitrary deadlines task
        self.constrained = True
        for j in self.jobs:
            for sep in j.succs.values():
                if sep < j.get_d():
                    self.constrained = False

        # Set default path abstractions accordingly
        if self.constrained:
            self.dbf_path_class = path.triple.DbfPath
            self.dbf_store_class = path.triple.DbfStore
            self.utilization_path_class = path.triple.UtilizationPath
            self.utilization_store_class = path.triple.UtilizationStore
            self.rf_path_class = path.rffullpath.FullRfPath
#            self.rf_store_class = path.RfPathStore
            self.rf_store_class = path.rffullpath.DiscardingRfStore
        else:
            self.dbf_path_class = path.quadruple.DbfPath
            self.dbf_store_class = path.quadruple.DbfStore
            self.utilization_path_class = path.quadruple.UtilizationPath
            self.utilization_store_class = path.quadruple.UtilizationStore

    def reset(self):
        self.U = None
        self.dbf = None
        self.rfpaths_cache = dict()
        self.dom_cache = dict()

    def set_dbf_path_abstraction(self, path_class):
        self.dbf_path_class = path_class
    
    def set_dbf_store_abstraction(self, store_class):
        self.dbf_store_class = store_class
    
    def set_utilization_path_abstraction(self, path_class):
        self.utilization_path_class = path_class
    
    def set_utilization_store_abstraction(self, store_class):
        self.utilization_store_class = store_class

    def get_jobs(self):
        """Returns a set containing all jobtypes in this task.
        """
        return self.jobs.copy()
    
    def get_e_sum(self):
        """Return the sum of the execution times of all the job (types)
        in the Task.
        """
        return sum([job.E for job in self.jobs])

    def reachable(self, jt1, jt2, t):
        """Returns True iff jt1 can reach jt2 within t time units.
        """
        # TODO: test more thoroughly

        min_durations = dict([(j, jt1.succs[j]) if j in jt1.succs else (j, "x")
            for j in self.jobs])
#        print ["%s: %s" % (j, min_durations[j]) for j in self.jobs]

        # At each step, update the dict with new step info. The minimal
        # value will strictly increase and eventually exceed t.
        for i in range(len(self.jobs)):
            # "x" compares larger than all integers
            new_durations = min_durations.copy()
            for job in [j for j in self.jobs if min_durations[j] < "x"]:
                for succ_job in job.succs.keys():
                    new_durations[succ_job] = min(
                        min_durations[job] + job.succs[succ_job],
                        new_durations[succ_job])
#                    dp(2, 2, "Reaching now %s within %d time units." % (succ_job, new_durations[succ_job]))
            min_durations = new_durations.copy()
#            print ["%s: %s" % (j, min_durations[j]) for j in self.jobs]
        return (min_durations[jt2] <= t)
     
    def get_max_releases(self, t):
        """Returns the maximal number of job releases by this task during
        any interval of size t.
        """
        releases = 0

        # This dict will contain for each job the minimal duration in
        # which the current number of jobs (counted in variable releases)
        # can be releases with this job being the last one.
        min_durations = dict.fromkeys(self.jobs, 0)

        # At each step, update the dict with new step info. The minimal
        # value will strictly increase and eventually exceed t.
        while min(min_durations.values()) <= t:
            releases += 1
#            dp(2, 2, "Step %d" % releases)
            # "x" compares larger than all integers
            new_durations = dict.fromkeys(self.jobs, "x")
            for job in filter(lambda j: min_durations[j] < "x", self.jobs):
                for succ_job in job.succs.keys():
                    new_durations[succ_job] = min(
                        min_durations[job] + job.succs[succ_job],
                        new_durations[succ_job])
#                    dp(2, 3, "Reaching %s from %s in time %d." % (succ_job, job, new_durations[succ_job]))
            min_durations = new_durations.copy()
        return releases

    def get_rfpaths(self, D, abstraction_level=None, force=False):
        """Return the request functions as lists for this task. Return only
        full functions with D prefix, i.e., defined on domain [0, D-1].

        Abstraction levels:
        None: Concrete paths
        0: One abstract rfpath which is the maximum over all concrete rfs
        1: Hierarchical abstractions from concrete paths. Still returns
            only one abstract rfpath but this one includes two children from
            which it abstracts. The leaves of this tree are all concrete paths.
        """
        
        if (D, None) > max(self.rfpaths_cache.keys() + [(-1, None)]) or force:
            self.prepare_rfpaths(D)
        elif not (D, None) in self.rfpaths_cache:
            # There is no cached path set for this D, but we can derive one.
            (maxD, _) = max(self.rfpaths_cache.keys())
            
            newrfpaths = set([path.rffullpath.FullRfPath(p.jobs, D) 
                for p in self.rfpaths_cache[(maxD, None)]])
            
            timer = time.time()
            maxpaths = set()
            add_to_maxpaths(newrfpaths, maxpaths, D)
            
            for i, p in enumerate(maxpaths):
                p.name = "%s_D%d_p%d" % (self.name, D, i)
            self.rfpaths_cache[(D, None)] = maxpaths

        # Here we have non-abstracted paths in the cache.
        # Take care of abstraction now.
        if not (D, abstraction_level) in self.rfpaths_cache.keys():
            if abstraction_level == 0:
                rfpaths = self.rfpaths_cache[(D, None)]
                maxrf = [max([p.get_rf()[i] for p in rfpaths])
                    for i in range(D)]
                maxrfpath = path.rffullpath.AbstractRfPath(D, maxrf)
                maxrfpath.name = "%s_A0_D%d" % (self.name, D)
                self.rfpaths_cache[(D, 0)] = set([maxrfpath])
            elif abstraction_level == 1:
                rfpaths = self.rfpaths_cache[(D, None)].copy()
                while len(rfpaths) > 1:
                    # Compute distance matrix
                    distances = list()
                    for ((i1, p1), (i2, p2)) in itertools.product(
                        enumerate(rfpaths), repeat=2):
                        if i1 < i2:
                            distances.append((p1.distance_to(p2), i1, i2, p1, p2))
                
                    # Merge path pair with minimal distances
                    (dist, _, _, p1, p2) = min(distances)
                    maxrf = [max(p1.get_rf()[i], p2.get_rf()[i]) for i in range(D)]
                    maxrfpath = path.rffullpath.AbstractRfPath(D, maxrf,
                        childs = [p1, p2], spread = dist)
                    maxrfpath.name = "%s_A1_D%d_p%d" % (self.name, D, len(rfpaths)-2)
                    
                    # Replace merged paths with maximum
                    rfpaths.remove(p1)
                    rfpaths.remove(p2)
                    rfpaths.add(maxrfpath)
                self.rfpaths_cache[(D, 1)] = rfpaths
                    
            else:
                raise RuntimeError("Unsupported path abstraction level: %s" %
                    str(abstraction_level))
                
            
        return self.rfpaths_cache[(D, abstraction_level)]

    def prepare_rfpaths(self, D):
        """Compute request functions up to D without abstractions."""
        timer = time.time()
        
        # Explore task graph.
        store = self.rf_store_class(D)
        path_creator = lambda jlist: self.rf_path_class(jlist, max_horizon = D)
        ee = engine.ExplorationEngine(self, store, path_creator)
        ee.run()
        
        
        timer = time.time()
        maxpaths = set()
        add_to_maxpaths(store.get_paths(), maxpaths, D)
        
        for i, p in enumerate(maxpaths):
            p.name = "%s_D%d_p%d" % (self.name, D, i)

        self.rfpaths_cache[(D, None)] = maxpaths
                

    def get_dbf(self, t):
        """Return the value dbf(<t>) for this Task. (The dbf must have been
        computed first with prepare_dbf().)
        """
        if self.dbf is None:
            raise RuntimeError('The dbf is not computed!')
        return self.dbf[t]

    def get_utilization(self):
        """Return the utilization of this Task. (The utilization must have
        been computed first with prepare_utilization(). If this was not
        done before, do it right away.)"""
        if self.U is None:
#            raise RuntimeError('The utilization is not computed!')
            self.prepare_utilization()
        return self.U

    def prepare_utilization(self):
        """Compute the utilization of this Task."""
        timer = time.time()
        
        # Explore task graph.
        
        store = self.utilization_store_class(len(self.jobs))
        ee = engine.ExplorationEngine(self, store, self.utilization_path_class)
        ee.run()
        e2_max = 0 # Start with dummy cycle (U = 0).
        p_max = 1
        witness = None

        for path in store.get_paths():
            # Check only cycles.
            if path.get_start() == path.get_end() and len(path) > 0:
                e2 = path.get_e2()
                p = path.get_p()
        
                # Test if a new most dense cycle is found (avoid rounding).
                if e2 * p_max > e2_max * p:
                    e2_max = e2
                    p_max = p
                    witness = path
                    
        
        if witness is None:
            self.U = float(0)            
        else:
            self.U = float(e2_max) / p_max
            
    
    def __str__(self):
        return str(self.name)


def add_to_maxpaths(paths, maxpaths, D):
    for path in paths:

        
        # Check whether too short
        if path.get_horizon() < D:
            continue

        # Check whether we are dominated by some processed path
        dominated = False
        new_rf = path.get_rf()
        for p in maxpaths:
            rf = p.get_rf()
            dominated = all([rf[i] >= new_rf[i] for i in range(D)])
            if dominated:                
                break
        if dominated:
            continue
        

        # Remove all paths dominated by us
        remove_paths = set()
        for p in maxpaths:
            rf = p.get_rf()
            dominated = all([rf[i] <= new_rf[i] for i in range(D)])
            if dominated:                
                remove_paths.add(p)
        maxpaths.difference_update(remove_paths)

        # Done. Add finally.
        maxpaths.add(path)

