package abstraction;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dialog;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.LinkedList;

import javax.swing.AbstractCellEditor;
import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JColorChooser;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JTable;
import javax.swing.SwingUtilities;
import javax.swing.border.Border;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableCellRenderer;

import util.Dialogs;
import util.GUIUtil;

public class TasksTable extends JTable {
	
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	AbstractionToolbar control;

	public TasksTable(AbstractionToolbar toolbar) {
		control = toolbar;
		setTableModel(new String[]{}, new Color[]{});
		setDefaultRenderer(Color.class, new ColorRenderer());
		setDefaultEditor(Color.class, new ColorEditor());
	}

	public void setData(String[] task_names, Color[]colors) {
		setTableModel(task_names, colors);
	}
	
	private void setTableModel(String[] task_names, Color[]colors) {
		setModel( new RfTaskTableModel(task_names, colors, control) );
		getColumnModel().getColumn(0).setPreferredWidth(45);
        getColumnModel().getColumn(1).setPreferredWidth(90);
        getColumnModel().getColumn(2).setPreferredWidth(45);
        setFillsViewportHeight(true);
        GUIUtil.alignHeader(this);
	}
}

class RfTaskTableModel extends AbstractTableModel {
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	private AbstractionToolbar	parent;	
    private final	String[] columnNames = {"Show", "Task Name", "Color"};
    private Object[][] data; //  = { {""} };	// default: an empty array
    
    public RfTaskTableModel(String []names, Color []colors, AbstractionToolbar toolbar) {
    	parent = toolbar;
    	data = new Object[names.length][3];
    	for (int i = 0; i < data.length; i++) {
    		data[i][0] = true;
    		data[i][1] = names[i];
    		data[i][2] = colors[i];
    	}
    }
    
    public int getColumnCount() {
        return columnNames.length;
    }

    public int getRowCount() {
        return data.length;
    }

    public String getColumnName(int col) {
        return columnNames[col];
    }

    public Object getValueAt(int row, int col) {
        return data[row][col];
    }

    /*
     * JTable uses this method to determine the default renderer/
     * editor for each cell.  If we didn't implement this method,
     * then the last column would contain text ("true"/"false"),
     * rather than a check box.
     */
    public Class getColumnClass(int c) {
        return getValueAt(0, c).getClass();
    }

    /*
     * Don't need to implement this method unless your table's
     * editable.
     */
    public boolean isCellEditable(int row, int col) {
        //Note that the data/cell address is constant,
        //no matter where the cell appears on screen.
        if (0 < col && col < 2) {
            return false;
        } else {
            return true;
        }
    }

    /*
     * Don't need to implement this method unless your table's
     * data can change.
     */
    public void setValueAt(Object value, int row, int col) {
    	data[row][col] = value;
        if (col == 0)
        	parent.selectedTasksChanged(selectedRows());
        else if (col == 2) {
        	parent.colorChanged(row, (Color)value);
        	parent.selectedTasksChanged(selectedRows());
        }
        
        fireTableCellUpdated(row, col);
    }
    
    private Integer[] selectedRows() {
		LinkedList<Integer> selectedInd = new LinkedList<Integer>();
		for(int row = 0; row < getRowCount(); row++)
			if((Boolean)getValueAt(row, 0) == true)
				selectedInd.add(row);
		return selectedInd.toArray(new Integer[0]);
	}    
}


class ColorEditor extends AbstractCellEditor
                         implements TableCellEditor,
                        ActionListener {
    /**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	Color   currentColor;
    JButton button;        
 
    public ColorEditor() {
        //Set up the editor (from the table's point of view),
        //which is a button.
        //This button brings up the color chooser dialog,
        //which is the editor from the user's point of view.
        button = new JButton();
        button.addActionListener(this);
        button.setBorderPainted(false); 
    }
 
    /**
     * Handles events from the editor button and from
     * the dialog's OK button.
     */
    public void actionPerformed(ActionEvent e) {
        //The user has clicked the cell, so
        //bring up the dialog.
        button.setBackground(currentColor);
        Color color = Dialogs.showColorDialog(currentColor);
        currentColor = color; 
        
        //Make the renderer reappear.
        fireEditingStopped();
       }
 
    //Implement the one CellEditor method that AbstractCellEditor doesn't.
    public Object getCellEditorValue() {
        return currentColor;
    }
 
    //Implement the one method defined by TableCellEditor.
    public Component getTableCellEditorComponent(JTable table,
                                                 Object value,
                                                 boolean isSelected,
                                                 int row,
                                                 int column) {
        currentColor = (Color)value;
        return button;
    }
}

class ColorRenderer extends JLabel implements TableCellRenderer {
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	Border unselectedBorder = null;
	Border selectedBorder = null;
	
	public ColorRenderer() {
		setOpaque(true); 
	}
	
	public Component getTableCellRendererComponent(JTable table, Object color,
			boolean isSelected, boolean hasFocus,
			int row, int column) {
		Color newColor = (Color)color;
		setBackground(newColor);
		if (isSelected) {
			if (selectedBorder == null) {
				selectedBorder = BorderFactory.createMatteBorder(2,5,2,5,
			                   table.getSelectionBackground());
			}
			setBorder(selectedBorder);
		} else {
			if (unselectedBorder == null) {
			unselectedBorder = BorderFactory.createMatteBorder(2,5,2,5,
	                   table.getBackground());
			}
			setBorder(unselectedBorder);
		}
	
		setToolTipText("RGB value: " + newColor.getRed() + ", "
		          + newColor.getGreen() + ", "
		          + newColor.getBlue());
		return this;
	}
}

