package analysis;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.util.LinkedList;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JToolBar;
import javax.swing.border.Border;
import javax.swing.border.EmptyBorder;

import enums.TaskModel;

import analysis.model.AnalysisMgr;
import analysis.model.Settings;
import analysis.model.enums.*;
import analysis.view.CoreTablesPanel;
import analysis.view.ResultsPanel;

import services.RunResult;
import util.Dialogs;
import util.GUIUtil;
import util.SchedPolicy;

import main.JManager;
import main.Manager;
import main.NotesPanel;
import main.Util;

public class AnalysisTab extends JPanel {
	
	private		AnalysisToolBar	toolBar;
	private		SettingsPanel	settingsPanel;
	private		ResultsPanel	resultsPanel;
	private		CoreTablesPanel	coreTablesPanel;
	private		NotesPanel		notes;
	private		AnalysisMgr 	amgr;
	
	public AnalysisTab (Manager m) {
		init(m);
		createPanels();
		createTB(); 
				
		add(toolBar, BorderLayout.PAGE_START);
		add(createMainPanel(), BorderLayout.CENTER);
	}

	private void init(Manager m) {
		amgr  = new AnalysisMgr(m);
		notes = m.notesPanel();
		setLayout(new BorderLayout());
		int R = (int) (245);
		setBackground(new Color(R, R, R));		
	}
	
	/*
	 * GUI-related methods:
	 */
	private void createPanels() {
		settingsPanel   = new SettingsPanel();
		resultsPanel    = new ResultsPanel();						    
		coreTablesPanel = new CoreTablesPanel();		
	}

	private void createTB() {
		toolBar = new AnalysisToolBar(this);		
	}

	private JScrollPane createMainPanel() {
		JPanel left = createVertPanel(); 
		left.add(settingsPanel);
		left.add(getVStrut(10));		
		left.add(resultsPanel);
		left.add(getVStrut(10));		
		// GUIUtil.color(left, Color.LIGHT_GRAY);
		
		JPanel right = createVertPanel();
		right.add(coreTablesPanel);		
		right.setBackground(Color.WHITE);
				
		JPanel bottom = new JPanel();
		bottom.setLayout(new FlowLayout( FlowLayout.LEADING, 10, 10));
		bottom.setBorder(GUIUtil.getBlackTitleBorder(""));
		bottom.add(left);
		bottom.add(right);
		pack(left, right);
		
		JScrollPane scr = GUIUtil.customScroll(bottom);
		return scr;
	}
	
	// creates a vertical strut
	private Component getVStrut(int slack) {
		return Box.createVerticalStrut(slack);		
	}

	private void pack(JPanel left, JPanel right) {
		int lh = left.getPreferredSize().height;
		int rh = right.getPreferredSize().height;
		int space = Math.abs(lh - rh);
		if (lh < rh)
			left.add(Box.createRigidArea(new Dimension(5, space)));
		else
			right.add(Box.createRigidArea(new Dimension(5, space)));
	}

	private JPanel createVertPanel() {
		JPanel p = new JPanel();
		p.setLayout(new BoxLayout(p, BoxLayout.Y_AXIS));
		
		Border b1 = GUIUtil.getGrayTitleBorder("");
		Border b2 = new EmptyBorder( 5, 5, 5, 5);
		p.setBorder(BorderFactory.createCompoundBorder(b1, b2));
		
		return p;
	}

	/*
	 * Analysis-related methods:
	 */
	public void run() {
		if (amgr.hasNonEmptyTask() == false) {
			toolBar.finished();
			Dialogs.showInfoMsg("Currently no non-empty active task set is loaded");
			return;
		}
		
		LinkedList<String> errors = amgr.analysisSyntaxErrors();
		if (errors.size() > 0) {
			toolBar.finished();
			Dialogs.showDefaultSyntaxErrDialog(errors);
			return;
		}
		
		new Thread(new Runnable() {
			public void run() {
				try {
					Settings cfg = getCfg();					
					RunResult res = amgr.runParitioning(cfg);
					// res = Util.runAnalysis(mgr, sched);					
					toolBar.finished();
					if (res != null)
						if (res.succeeded()) {
							res = processResult(res);
							repaint();
							revalidate();
						}
					showRunMsg(res);							
				} catch (Exception ex) {
					ex.printStackTrace();
					toolBar.finished();
				}
			}
		}).start();		
	}

	protected Settings getCfg() {
		int 		  n		= numOfCores();
		BinPackPolicy bp 	= binPMethod();
		SortCriteria  sortC = sortCriter();
		Processor	  proc  = uniProc(); 
		SchedPolicy   sched = schPolicy();
		TaskModel	  model	= taskModel();	
		return new Settings(n, bp, sortC, proc, sched, model);
	}

	private void showRunMsg(RunResult res) {
		Util.showRunMsg(res, notes);
	}
	
	public void stop() {
		amgr.stop();		
	}

	public RunResult processResult(RunResult res) {
		if(res == null)
			return res;
		
		boolean isSchedulable = processAndShowResults(res.linesArr());

		res.setSchedulable(isSchedulable);
		if(isSchedulable)
			res.setMessage("Task set is schedulable.");
		else
			res.setMessage("Task set is not schedulable.");
		return res;
	}

	/*
	 *  an example of the input:
	 *   	Core 1: RandomTask1:0.0173913043478,RandomTask3:0.0178571428571
	 *   	Unalloc :
	 *   returns: total load and unallocated load obtained by processing the input data
	 */	
	private boolean processAndShowResults(String[] res) {
		String [][]data;
		double totU = 0;	// utilization
		double coreU = 0;
		double totD = 0;	// density
		double coreD = 0;
		int taskNo = 0;
		
		coreTablesPanel.removeAll();		
		// each item of res is related to a core (except the the last one, which is for unallocated tasks)
		for (String s : res) {
			coreU = 0;
			coreD = 0;
			taskNo = 0;
			s = s.replaceAll(" ", "");		// remove all blank spaces
			String[] parts = s.split(":", 2);
			String coreName = parts[0];
			String []tasks = parts[1].split(",");
			data = new String [tasks.length][3];
			for (int i = 0; i < tasks.length; i++) {
				String[] task_info = tasks[i].split(":");
				if (task_info.length < 3)
					continue;
				data[i][0] = task_info[0];				// task name
				data[i][1] = substr(task_info[1], 5);	// task utilization
				data[i][2] = substr(task_info[2], 5);	// task density
				coreU += getVal(task_info[1]);
				coreD += getVal(task_info[2]);
				taskNo++;			// when exiting the loop, this shows the number of unallocated tasks
			}
			coreTablesPanel.addCore(data, coreName, coreU, coreD);
			if (schPolicy() == SchedPolicy.SP) {
				String[] names = Util.firstCol(data);
				amgr.setPrio(names, coreName);
			}
			totU += coreU;
			totD += coreD;
		}
		coreTablesPanel.repaint();
		
		double[] loads = new double[]{totU, totD, coreU, coreD};
		loads = Util.round(loads, 3);
		resultsPanel.update_data(loads);		
		
		boolean isSchedulable = (taskNo > 0) ? false : true; 
		return isSchedulable;
	}

	private String substr(String s, int i) {
		int len = s.length();
		return s.substring(0, Math.min(len, i));
	}

	public void clear() {
		resultsPanel.clear();
		coreTablesPanel.clear();
	}
	
	/* 
	 * Getters:
	 */
	public int numOfCores() {
		return settingsPanel.getCoreNo();		
	}

	public SchedPolicy schPolicy() {
		return settingsPanel.getSchPolicy();
	}

	public BinPackPolicy binPMethod() {
		return settingsPanel.getBinPMethod();
	}

	public SortCriteria sortCriter() {
		return settingsPanel.getSortCriteria();
	}
	
	private TaskModel taskModel() {
		return settingsPanel.getTaskModel();
	}

	public Processor uniProc() {
		if (settingsPanel.uniProc())
			return Processor.unicore;
		else
			return Processor.multicore;
	}

	private double getVal(String s) {
		try {
			return Double.parseDouble(s);
		} catch(Exception e) {
			return 0;
		}
	}
}
