package main;
/*
 * Each edge connects a src job to a dst 
 *  and is labeled with the minimum inter-release
 *  separation (release) time between them.
 */
import java.awt.Color;
import java.awt.Point;
import java.util.Vector;


import taskeditor.DrawingPane;
import taskeditor.changetrace.Change;
import taskeditor.changetrace.EdgeChange;
import taskeditor.movables.EdgeLabel;
import taskeditor.movables.Nail;
import taskeditor.movables.Nails;
import util.Dialogs;
import util.EdgeGUIUtil;
import util.GUIUtil;
import enums.ActionType;
import enums.SendSemantic;

public class Edge implements Comparable<Edge> {

	protected	JobType 	source;
	protected	JobType 	destin;
	protected 	EdgeParam	params;
	private		String		guard;	  
	private 	String  	code; 	
	private		EdgeLabel	label;
	private		Nails		nails;  // Excluding start and end points 	
	private		int 		index;  // To make it unique
	
	public Edge() {
		params = new EdgeParam();
		label  = new EdgeLabel();
		guard  = "True";
		code   = "";
	}
	
	public Edge(int indx, JobType src, JobType dst, int interRel, Vector<Point> nailz) {
		this();
		index = indx;
		source = src;
		destin = dst;
		params.interRel = interRel;
		nails = new Nails(nailz, this);
	}
	
	public Edge(int indx, JobType src, JobType dst, EdgeParam data, Vector<Point> nailz) {
		this(indx, src, dst, data.interRel, nailz);
		params.action = data.action;
		params.aType  = data.aType;
		params.send   = data.send;
	}
	
	private void setLabelLocation(int x, int y) {
		label.adjustSizeAndLocation(x, y);
	}
	
	public void addNail(int x, int y) {
		nails.addNail(x, y);
	}

	public Nail[] nails() {
		if (nails == null)
			return new Nail[]{};
		return nails.toArray();
	}
	
	public Point[] getAugmentedNails() {
		return nails.getAugmentedPoints();
	}
	
	public boolean testPoint(Point p, int margin) {
		return nails.testPoint(p, margin);
	}
	
	// called by DrawingPane.drawTaskGraph()
	public void updateLabel(DrawingPane drawPane) {
		if (label.listenersSet() == false) {
			label.setListener(drawPane, this);
			if (label.hasValidPosition() == false)
				setDefaultLabelLoc();
		} 
		label.setText(params.toString());		
	}

	private void setDefaultLabelLoc() {
        int x1 = (int)source.x;
        int y1 = (int)source.y;
        int x2 = (int)destin.x;
        int y2 = (int)destin.y;
        Point p = new Point((x1+x2)/2, (y1+y2)/2 + 10); // - 15);
    	if (source == destin) {   // this is an edge from a vertex to itself
        	p.x += 120;
        }
        if (nails != null && nails.size() > 0) {
        	p = new Point(nails.get(0).x(), nails.get(0).y());	
        }	 
        setLabelLocation(p.x, p.y);
    }
	
	public void editRequest() {
		EdgeParam data = EdgeGUIUtil.getEdgeParamFromUser(params);        	
		if (data == null)
			return;

    	if (check(data.interRel) == false) { 
    		editRequest(); 		
    		return;
    	}    	
    	setLabelText(data.toString());
        setParams(data);
	}

	private boolean check(int i) {
		if (i < 0) {
			Dialogs.showErrorMsg("Inter-release time cannot be negative.");
			return false;
		}
		return true;		
	}

	public void setLabelText(String txt) {
		label.setText(txt);
	}

	public void deselect() {
		GUIUtil.color(label, Color.black);
	}

	public void selected() {
		GUIUtil.color(label, Color.blue);
	}

	public void setLabelLoc(Point p) {
		label.adjustSizeAndLocation(p.x, p.y);
	}

	public int interRel() {
		return params.interRel;
	}
	
	public String action() {
		return params.action;
	}

	public ActionType actionType() {
		return params.aType;
	}
	
	public SendSemantic sendSemantic() {
		return params.send;
	}
	
	public String srcName() {
		return source.getName();
	}
	
	public String dstName() {
		return destin.getName();
	}
	
	public String toString() {
		return "(" + srcName() + "," + dstName() + ")";
	}
	
	public int compareTo(Edge e) {		
		int i = interRel();
		int o = e.interRel();
		if (i > o) {
			return 1;
		} else if (i == o) {
			return 0;
		} else {
			return -1;
		}
	}
		
	/***********************************
	 ****** Change track methods  ******
	 ***********************************/	
	public Edge cloneMe() {
		Vector<Point> ns = nails.clonePoints();
		Edge cloned = new Edge(index, source, destin, params, ns);
		if (label.hasValidPosition()) {
			Point p = label.getXY();
			cloned.label.adjustSizeAndLocation(p.x, p.y);
		}
		cloned.code = this.code;
		cloned.guard = this.guard;		
		return cloned;
	}

	public Change getModifyChange() {
		EdgeChange ch = createChange();
		ch.beModify();
		return ch;
	}

	public Change getCreateChange() {
		EdgeChange ch = createChange();
		ch.beCreate();
		return ch;
	}
	
	public Change getRemoveChange() {
		EdgeChange ch = createChange();
		ch.beRemove();
		return ch;
	}
	
	private EdgeChange createChange() {
		EdgeChange ch = new EdgeChange();
		ch.setRef(this);
		ch.setOrig(this.cloneMe());		
		return ch;
	}
	
	// called for undoing some operation 
	public void loadData(Edge orig) {
		this.code = orig.code;
		this.guard = orig.guard;
		this.nails = orig.nails;
		this.nails.setEdge(this);
		this.params.loadData(orig.params);
		if (orig.label.hasValidPosition()) {
			Point p = orig.label.getXY();
			setLabelLocation(p.x, p.y);			
		}
	}

	// two edges are parallel if they have the same source and destination
	public boolean isParallel(Edge o) {
		if (source != o.source)
			return false;
		if (destin != o.destin)
			return false;
		
		return true;
	}
	
	public boolean hasAction() {
		return actionType() == ActionType.None ? false : true;
	}

	public boolean receives() {
		 return params.aType == ActionType.Receive;
	}

	public boolean sends() {
		 return params.aType == ActionType.Send;
	}
	
	public boolean blockingSend() {
		return params.send == SendSemantic.Blocking;
	}
	
	/*
	 * Getters and Setters
	 */
	public String code() {
		return code;
	}
	
	public void setCode(String s) {
		code = s;		
	}

	public JobType destin() {
		return destin;
	}
		
	public String guard() {
		return guard;
	}
	
	public void setGuard(String g) {
		guard = g;		
	}

	public int index(){
	    return index;
	}
    
	public EdgeLabel getLabel() {
		return label;
	}
	
	public EdgeParam params() {
		return params;
	}
	
	public void setParams(EdgeParam p) {
		params = p;
	}

	public JobType source() {
		return source;
	}
}
