/*
 * Front-end application of MASI.
 * 
 * Written 2015-2016 by Morteza Mohaqeqi, Jakaria Abdullah.
 * 
 * Uppsala University.
 */

package main;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.WindowStateListener;

import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JSplitPane;
import javax.swing.JTabbedPane;
import javax.swing.UIManager;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import abstraction.AbstractionTab;
import analysis.AnalysisTab;

import simulator.SimTab;
import taskeditor.EditorTab;
import util.GUIUtil;

/*
 * This class is the main frame of the application,
 * which is inter-connected with the Manager class
 * and contains the visual aspects.
 */

public class JManager extends JFrame {

	private JPanel 			tabbedPanel; // the main panel which contains the tabs
	private NotesPanel 		notesPanel; // the panel at the bottom which shows notes
	private String			lookAndFeel = "";		 
	
	private   	  String 	current_tab_name = "";
	private final String 	EDITOR_TAB_TITLE = "Task Editor";
	private final String 	SIM_TAB_TITLE    = "Simulator";
	private final String 	DBF_TAB_TITLE 	 = "Abstraction";
	private final String 	AN_TAB_TITLE 	 = "Analysis";
	private final String 	SYN_TAB_TITLE 	 = "Synthesis";	

	public JManager() {
		loadLookAndFeel();
		createNotePanel();
		setBackground(Color.gray);
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		UIManager.put("ToolTip.background", Color.WHITE);
	}
		
	public void init(final Manager m, EditorTab eTab, SimTab sTab, AbstractionTab aTab, AnalysisTab mTab, JPanel syTab) {
		
		JMenuBar menuBar = createMenuBar(m); // menu bar contains: File, Edit,...
		setJMenuBar(menuBar);
		
		JPanel topPanel = new JPanel();
		topPanel.setLayout(new BorderLayout());
		add(topPanel);
		
		// Create the panels
		createTabbedPanel(eTab, sTab, aTab, mTab, syTab);		
		
		// Create a splitter pane
		JSplitPane splitPaneV = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
		topPanel.add(splitPaneV, BorderLayout.CENTER);
		
		splitPaneV.setLeftComponent(tabbedPanel);	// sets the component left or above
		splitPaneV.setRightComponent(notesPanel);
		Util.decorateSplit(splitPaneV, 500);
		splitPaneV.setResizeWeight(0.8);
		
		setPreferredSize(new Dimension(900, 700));
		setLocation(100, 50);
		setStateListener(sTab);
		
		addWindowListener(new WindowAdapter() {
            @Override
            public void windowClosing(WindowEvent e) {
                m.stop();
                e.getWindow().dispose();
            }
        });
	}

	
	public void createTabbedPanel(JPanel editor, final SimTab simTab, final AbstractionTab abstctnTab, AnalysisTab analysisTab, JPanel synthesisTab) {
		JTabbedPane tabbedPane = new JTabbedPane();		
		tabbedPane.addTab(EDITOR_TAB_TITLE, editor);
		tabbedPane.addTab(SIM_TAB_TITLE, simTab);
		tabbedPane.addTab(DBF_TAB_TITLE, abstctnTab);
		tabbedPane.addTab(AN_TAB_TITLE,  analysisTab);
		tabbedPane.addTab(SYN_TAB_TITLE,  synthesisTab);
		
		tabbedPane.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent changeEvent) {
				JTabbedPane sourceTabbedPane = (JTabbedPane) changeEvent
						.getSource();
				current_tab_name = Util.tabTitle(sourceTabbedPane);
				if (current_tab_name.equals(SIM_TAB_TITLE)) {
					simTab.selected();					
				} else if (current_tab_name.equals(DBF_TAB_TITLE)) {
					abstctnTab.selected();
				} else {
					simTab.stopSim();					
				}
			}
		});

		current_tab_name = EDITOR_TAB_TITLE;	// TO OD: it can be removed, and checked wherever it is used
		tabbedPanel = new JPanel(new BorderLayout());		
		tabbedPanel.add(tabbedPane, BorderLayout.CENTER);
	}
	

	private JMenuBar createMenuBar(Manager mgr) {
		JMenuBar menuBar = new JMenuBar();

		JMenu fileMenu = new JMenu("File");
		JMenu editMenu = new JMenu("Edit");
		JMenu analMenu  = new JMenu("Analysis");
		JMenu sntxMenu = new JMenu("Check");
		JMenu stngMenu = new JMenu("Settings");
		JMenu helpMenu = new JMenu("Help");

		ImageIcon iconExit = Util.getIcon("/exit_small.png");

		// Create and add simple menu item to one of the drop down menu
		JMenuItem exitAction   = new JMenuItem("Exit", iconExit);
		JMenuItem stopAction   = new JMenuItem("Stop");
		JMenuItem cutAction    = new JMenuItem("Cut");
		JMenuItem copyAction   = new JMenuItem("Copy");
		JMenuItem pasteAction  = new JMenuItem("Paste");
		JMenuItem respTimeAction = new JMenuItem("SP Response Time");
		JMenuItem analSynxAction = new JMenuItem("Syntax For Analysis");
		JMenuItem cgenSynxAction = new JMenuItem("Syntax For Code Generation");
		JMenuItem dvlpCfgAction = new JMenuItem("Tool Config.");
		JMenuItem guideAction  = new JMenuItem("User Guide");
		JMenuItem aboutAction  = new JMenuItem("About");

		// Set Mnemonic to the menu items
		fileMenu.setMnemonic(KeyEvent.VK_F);
		editMenu.setMnemonic(KeyEvent.VK_E);
		analMenu.setMnemonic(KeyEvent.VK_A);
		
		sntxMenu.setMnemonic(KeyEvent.VK_S);
		analSynxAction.setMnemonic(KeyEvent.VK_A);
		cgenSynxAction.setMnemonic(KeyEvent.VK_C);
		
		stngMenu.setMnemonic(KeyEvent.VK_T);
		helpMenu.setMnemonic(KeyEvent.VK_H);
		exitAction.setMnemonic(KeyEvent.VK_E);
		cutAction.setMnemonic(KeyEvent.VK_U);
		copyAction.setMnemonic(KeyEvent.VK_C);
		pasteAction.setMnemonic(KeyEvent.VK_P);
		respTimeAction.setMnemonic(KeyEvent.VK_R);
		aboutAction.setMnemonic(KeyEvent.VK_A);

		// Add a listener to the menu items. actionPerformed() method will
		// invoked, if user triggred this menu item
		cutAction.	  addActionListener(Util.createUnavailableFeatureListener());
		copyAction.	  addActionListener(Util.createUnavailableFeatureListener());
		pasteAction.  addActionListener(Util.createUnavailableFeatureListener());
		respTimeAction.addActionListener(Util.createRespTimeListner( mgr ));
		dvlpCfgAction.addActionListener(new ConfigPanel( mgr ));
		analSynxAction.addActionListener(Util.createAnalSyntaxListner( mgr ));
		cgenSynxAction.addActionListener(Util.createUnavailableFeatureListener());
		
		guideAction.  addActionListener(Util.createGuideActListener());
		aboutAction.  addActionListener(Util.createHelpActListener());		
		exitAction.	  addActionListener(Util.createExitActListener( mgr ));
		
		// Add the sub-items to the menu items		
		fileMenu.add(exitAction);
		editMenu.add(cutAction);
		editMenu.add(copyAction);
		editMenu.add(pasteAction);
		analMenu.add(respTimeAction);
		editMenu.addSeparator();
		sntxMenu.add(analSynxAction);
		sntxMenu.add(cgenSynxAction);
		stngMenu.add(dvlpCfgAction);
		helpMenu.add(guideAction);
		helpMenu.add(aboutAction);

		// Add the items to the menu bar
		menuBar.add(fileMenu);
		menuBar.add(analMenu);
		menuBar.add(sntxMenu);
		menuBar.add(stngMenu);
		menuBar.add(helpMenu);

		return menuBar;
	}


	private void loadLookAndFeel() {
		String[] conf = Util.load_conf();
		if (conf.length >= 4) {
			lookAndFeel = conf[3];
			setLookAndFeel(lookAndFeel);
		} else {
			Util.warning("Warning: no look and feel specified in the config file.");
		}
	}

	public void setLookAndFeel(String lafName) {
		GUIUtil.setLookAndFeel(lafName);
	}
	
	/*
	 * resizes the components (specially splits) whenever the frame resized
	 */
	private void setStateListener(final SimTab simTab) {
		addWindowStateListener(new WindowStateListener() {
			   public void windowStateChanged(WindowEvent e) {
				   // maximized
				   if ((e.getNewState() & Frame.MAXIMIZED_BOTH) == Frame.MAXIMIZED_BOTH){
					   simTab.repaintViewer();
				   }
				   simTab.rearrange();
			   }
			});		
	}

	// creates the 'Notes' panel (placed at the bottom of the frame) 
	public void createNotePanel() {
		notesPanel = new NotesPanel();
	}

	public boolean tabsActivated() {
		if (current_tab_name == null)
			return false;
		else
			return true;
	}

	/*
	 * Print out a message to the user in the notes panel
	 */
	public void printNote(String msg) {
		notesPanel.println(msg);		
	}
	
	public void printNote(String msg, Color c) {
		notesPanel.println(msg, c);
	}
	
	public void printDashedLine() {
		notesPanel.printDashedLine();
	}
	
	public NotesPanel notesPanel() {
		return notesPanel;
	}

	public String getLookAndFeel() {
		return lookAndFeel;
	}	
}
