/*
 * Front-end application of MASI.
 * 
 * Written 2015-2016 by Morteza Mohaqeqi, Jakaria Abdullah.
 * 
 * Uppsala University.
 */

package main;

import java.awt.Color;
import java.awt.KeyEventDispatcher;
import java.awt.KeyboardFocusManager;
import java.awt.event.*;
import java.util.LinkedList;

import enums.TaskModel;

import abstraction.AbstractionTab;
import analysis.AnalysisTab;
import analysis.model.Settings;

import services.DRTStub;
import services.RunResult;
import simulator.SimTab;
import synthesis.SynthesisTab;
import taskeditor.EditorTab;
import main.storage.TSSaver;
import util.GUIUtil;
import util.SchedPolicy;
import util.SyntaxChecker;

public class Manager {

	private JManager 		frame;
	private EditorTab 		editorTab;	
	private SimTab 			simTab;
	private AbstractionTab 	absTab;
	private AnalysisTab		analysisTab;
	private SynthesisTab	synthesisTab;
	private DRTStub			drtlib;
	private final String 	TOOL_NAME 		  = "TIMES-Pro";
			
	public Manager() {
		init();
		
		frame = new JManager();
		createTabs();
		addKeyEventDispatcher();
		
		frame.init(this, editorTab, simTab, absTab, analysisTab, synthesisTab);
		frame.pack();
	}
	
	private void init() {
		drtlib = new DRTStub();
		drtlib.loadPaths();				
	}

	private void createTabs() {
		editorTab    = new EditorTab(this);
		simTab       = new SimTab(this);
		absTab 	     = new AbstractionTab(this); 
		analysisTab  = new AnalysisTab(this);
		synthesisTab = new SynthesisTab(this);
	}
	
	public void setTiltle(String taskSet) {
		frame.setTitle(TOOL_NAME);
	}
		
	public void newTaskSetLoaded(String taskSet) {
		if (analysisTab != null)
			analysisTab.clear();
		if (frame != null)
			frame.setTitle("TIMES-Pro");		
	}
	
	public void setPriorities(String[] ordered_names) {
		editorTab.setPriorityAndRefresh(ordered_names);		
	}
	
	public void save(String filePath) {
		editorTab.save(filePath);
	}

	public void stop() {
		if (drtlib.running()) {
			hideProgressBar();			
			drtlib.stop();
			String msg = "Process terminated by the user.";
			Util.log(msg);
			printNote(msg);
			System.out.println(msg);
		}
	}

	public void terminate() {
		stop();
		System.out.println("exiting");
		System.exit(0);		
	}
		
	public String getTaskSetPath() {
		return editorTab.getTaskSetPath();
	}	
		
	/*************************************
	 ******   Running DRTlib (Python) *****
	 *************************************/
	public RunResult runParitioning(Settings cfg) throws Exception {		
		printNote("Running analysis ... ");
		RunResult res = drtlib.runAnalysis(cfg, getTaskSet());
		printNote("Analysis Finished. ");
		return res;
	}
	
	public RunResult createRandomTaskSet(String type, String crit, String Ubound, String TNum, String TSParam) {
		RunResult res = drtlib.runRandomTasks(type, crit, Ubound, TNum, TSParam, this);
		hideProgressBar();
		printRunRes(res);
		return res;
	}
	
	private void printRunRes(RunResult res) {
		if (res.succeeded()) {
			printNote(res.message());
		} else {
			printNote(res.message(), Color.RED);
			Util.error(res.getException());
			if (res.exitValue() != 0) {
				printNote("See the log file for details.", Color.RED);				
				Util.error(res.getLog());
			}
		}		
	}
	
	public RunResult runMrfDBFAnalysis(String metric, int maxD) {
		if (drtlib.validatePaths() == false)
			return null;
		printNote("Running analysis ... ");
		
		TaskSet ts = getTaskSet();
		RunResult res = drtlib.workload(metric, ts, maxD);
		if (res.succeeded())
			printNote("Analysis finished.");
		else
			printNote("Analysis finished unsuccessfully.", Color.RED);
		printDashLine();
		hideProgressBar();
		return res;
	}
		
	public RunResult[] runRespTimeAnalysis() {

		if (drtlib.validatePaths() == false)
			return null;
		
		printNote("Running analysis ... ");
		
		Task[] selectedTasks = getSelectedTasks();
		RunResult[] results = drtlib.runRespTime(selectedTasks, getTaskSet(), notesPanel());
		
		hideProgressBar();
		printNote("Analysis finished.");
		return results;
	}
	
	
	/*************************************
	 ******   Getters and Setters   ******
	 *************************************/
	public Task getSelectedTask() {
		return editorTab.getSelectedTask();
	}
	
	private Task[] getSelectedTasks() {
		return editorTab.getSelectedTasks();
	}
	
	public String[] getSelectedTasksNames() {
		Task[] tasks = getSelectedTasks();
		String[] names = new String[tasks.length];
		for(int i = 0; i < tasks.length; i++)
			names[i] = tasks[i].getName(); 
		
		return names;
	}
	
	public LinkedList<String> analysisSyntaxErrors() {
		return SyntaxChecker.checkForAnalysis(editorTab.tasks()); 
	}
	
	public Task[] taskList() {
		return editorTab.tasks();
	}
	
	public TaskSet getTaskSet() {
		return editorTab.getTaskSet();
	}	

	public NotesPanel notesPanel() {
		return frame().notesPanel();
	}
	
	/*************************************
	 ******    Getting a Status     ******
	 *************************************/
	public boolean tabsActivated() {
		return frame.tabsActivated();
	}
	
	/*
	 * Called from TaskPanel.handleTableDataChange() to see if there is an
	 * actual task set loaded currently
	 */
	public boolean hasActiveTaskSet() {
		return editorTab.hasActiveTaskSet();		
	}

	/*
	 * Called by analyzing function to avoid an error when calling the engine with an empty task 
	 */
	public boolean hasNonEmptyTask() {
		return editorTab.hasNonEmptyTask();
	}
	
	public JManager frame() {
		return frame;
	}
	
	public DRTStub drtlib() {
		return drtlib;		
	}
	
	/*************************************
	 ****** GUI Output Presentation ******
	 *************************************/
	public void showProgressBar() {
		editorTab.showProgressBar();		
	}	
	
	public void hideProgressBar() {
		editorTab.hideProgressBar();		
	}	
			
	public void printNote(String msg) {
		frame.printNote(msg);
	}
	
	public void printNote(String msg, Color c) {
		frame.printNote(msg, c);
	}
	public void printDashLine() {
		frame.printDashedLine();
	}

	public void show() {
		frame.setVisible( true );
	}
	
	/***********************************
	 ****** GUI Settings methods  ******
	 ***********************************/
	public void addKeyEventDispatcher() {
		KeyboardFocusManager.getCurrentKeyboardFocusManager()
		.addKeyEventDispatcher(new CustomKeyEventDispatcher());		
	}
	
	class CustomKeyEventDispatcher implements KeyEventDispatcher {
		public boolean dispatchKeyEvent(KeyEvent e) {
			if (e.getID() != KeyEvent.KEY_PRESSED)
				return false;
			
			if (e.getKeyCode() == KeyEvent.VK_ESCAPE) {
				editorTab.escapePressed();
				//e.consume();
			} 
			else if (e.getKeyCode() == KeyEvent.VK_DELETE) {			
				// editorTab.deleteKeyPressed();
				// e.consume();					
			} 			
			return false;
		}
	}	
}