package main;

import main.storage.StorageUtil;
import main.storage.XMLUtil;

import org.w3c.dom.NodeList;
import org.w3c.dom.Node;

import enums.ActionType;
import enums.SendSemantic;

import taskeditor.changetrace.Change;
import taskeditor.changetrace.EdgeChange;
import taskeditor.changetrace.GroupChange;
import taskeditor.changetrace.JobChange;
import taskeditor.changetrace.TaskChange;
import util.Dialogs;
import util.GUIUtil;
import util.SyntaxChecker;

import java.util.Hashtable;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Vector;
import java.awt.Point;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * This class is responsible for loading the task properties from the user-specified XML file 
 */
public class Task {
	private String              name;
	private double				utilization;			// the utilization of the most dense cycle
	private	int					priority;	
	private	static	int			priorityCounter = 1;	// for assigning a default unique priority
	private	String				localVariables;			// for code generation
	private	JobType				startJob = null;
	private AtomicInteger       max_edge_index;         // for indexing edge in code generation

	private LinkedList<JobType> jobs;
	private LinkedList<Edge>    edges;
	private LinkedList<Change>  changeTrace = new LinkedList<Change>();
	// private Hashtable<JobType, LinkedList<Edge>>    branchOrders;

	// This constructor is used to load the task set from an xml file
	public Task(Node taskNode, String path) {
		init();
		name = attrVal(taskNode, "name"); 
		validate_name();
		setPriority(taskNode);
		utilization	= Util.str2num(attrVal(taskNode, "utilization"));		
	}

	private String attrVal(Node node, String attr) {
		return XMLUtil.attrVal(node, attr);
	}

	// This constructor is used when the task set is created from scratch
	public Task(String name) {
		init();
		this.name = name;
		this.max_edge_index = new AtomicInteger();   // indexed from 0
		validate_name();
		setPriority();
	}

	// used when cloning
	public Task(String namn, double utlz, int prio, String localVar, int max_edge_index) {
		init();
		this.name = namn;
		this.utilization = utlz;
		this.priority = prio;
		this.localVariables = localVar;
		this.max_edge_index = new AtomicInteger(max_edge_index); // edge indexing counter
		// validate_name();
	}

	private void setPriority(Node taskNode) {
		String pr = attrVal(taskNode, "priority");
		if(pr == null || pr.length() == 0)
			setPriority();	// default method
		else
			priority = Util.str2int(pr);        		
	}

	private boolean validate_name() {
		if (name.contains("'")) {
			Dialogs.showErrorMsg("Error: task name cannot contain this character: " + "'");
			return false;
		}	
		return true;
	}

	private void init() {
		jobs = new LinkedList<JobType>();
		edges = new LinkedList<Edge>();
		//        branchOrders = new Hashtable<JobType, LinkedList<Edge>>();
	}

	public String getName() {
		return name;
	}

	public void addNewJob(JobType job) {
		jobs.addLast(job);
	}

	//////////////////////////////////////////////////////////////////////////////////
	public int get_edge_Index() {
		return this.max_edge_index.incrementAndGet();
	}

	public int current_max_edge_Index() {
		return this.max_edge_index.get();
	}
	//////////////////////////////////////////////////////////////////////////////////

	public void addEdge(Edge new_edge) {
		edges.add(new_edge);
		JobType src = new_edge.source;
		src.addEdge(new_edge);
		/* if(branchOrders.containsKey(src))
			branchOrders.get(src).add(new_edge); */
	}

	public void removeEdge(Edge e) {
		edges.remove(e);
		JobType src = e.source;
		src.removeEdge(e);
		/* if(branchOrders.containsKey(src))
			branchOrders.get(src).remove(e); */	
	}

	public boolean contains_edge(JobType src, JobType dst) {
		for (Edge edge : edges) {
			if (edge.source == src && edge.destin == dst)
				return true;
		}
		return false;
	}
	
	public boolean contains(JobType j) {
		if (jobs == null)
			return false;
		return (jobs.contains(j));
	}

	/*
	 * This function is called by manager.jobPropValidity() to avoid two jobs with the same name
	 * If job not found, returns -1
	 */
	public int getJobNo(String job_name) {
		for (int i = 0; i < jobs.size(); i++)
			if (jobs.get(i).getName().equalsIgnoreCase(job_name))
				return i;
		return -1;
	}
	
	public int getJobIndex(JobType job){
		return jobs.indexOf(job);
	}

	public void setPriority(int priority) {
		this.priority = priority;
	}

	/*
	 * Sets prio. automatically
	 */
	public void setPriority() {
		this.priority = priorityCounter;
		priorityCounter++;
	}

	public int getPriority() {
		return priority;
	}

	// Called when creating or loading a new task set
	// Accordingly, it is called from Manager.createNewTaskSet() or TaskLoader 
	public static void resetPriorityCounter() {
		priorityCounter = 1;
	}

	/* 
	 * Called when user is modifying the task's name 
	 */
	public void setName(String name) {
		this.name = name;		
	}

	public void setStartJob(JobType j) {
		startJob = j;
	}

	public JobType getStartJob() {
		return startJob;
	}

	public boolean startJobTest(JobType job) {
		if (startJob == null)
			return false;
		return job == startJob;
	}

	public boolean receives(String act) {
		for (Edge e : edges) 
			if (e.params.aType == ActionType.Receive) 
				if (e.params.action.equals(act))
					return true;
		return false;
	}

	/* 
	 * Called by DrawingPane.getTargetEdge()	
	 */
	public Edge getTargetEdge(Point p, int margin) {
		if (jobs.size() == 0)
			return null;

		for (Edge edge : edges)
			if (edge.testPoint(p, margin))
				return edge;

		return null;
	}

	public ListIterator<Edge> edgeIterator() {
		return edges.listIterator();
	}

	public ListIterator<JobType> jobIterator() {
		if (jobs == null)
			return null;
		return jobs.listIterator();
	}

	public boolean hasAnyEdges() {
		if(edges == null)
			return false;
		if(edges.size() == 0)
			return false;
		return true;
	}

	public JobType getJob(int index) {    	
		return jobs.get(index);
	}

	public boolean isEmpty() {
		if(jobs == null)
			return true;
		if(jobs.size()  == 0)
			return true;
		return false;
	}

	public int jobNo() {
		if( isEmpty() )
			return 0;
		return jobs.size();
	}


	public void removeJob(JobType job) {
		LinkedList<Edge> victims = getInOutEdges(job);				
		for (Edge e : victims)
			removeEdge(e);

		jobs.remove(job);
	}

	public LinkedList<Edge> getInOutEdges(JobType job)  {
		LinkedList<Edge> el = new LinkedList<Edge>();
		for (Edge e : edges)
			if (e.source.equals(job) || e.destin.equals(job))
				el.add(e);
		return el;
	}

	public int maxD() {
		if(jobNo() == 0)
			return 0;
		int maxD = 0;
		for(JobType j : jobs)
			maxD = Math.max(maxD, j.deadline);
		return maxD;
	}

	public int getMaxD() {
		if(jobNo() == 0)
			return 0;
		int maxD = 0;
		for(JobType j : jobs)
			if(j.deadline > maxD)
				maxD = j.deadline;
		return maxD;
	}

	public Change getRemoveGroupChange(JobType j) {
		GroupChange gc = new GroupChange();
		LinkedList<Edge> victims = getInOutEdges(j);
		for (Edge e : victims)
			gc.add(e.getRemoveChange());

		gc.add(j.getRemoveChange());
		return gc;	
	}

	public Task cloneMe() {
		Task clone = new Task(name, utilization, priority, localVariables, max_edge_index.get());
		Hashtable<JobType, JobType> jMap = new Hashtable<JobType, JobType>();  
		Hashtable<Edge, Edge> eMap = new Hashtable<Edge, Edge>();
		for (JobType j : jobs) {
			JobType newJ = j.cloneMe();
			clone.addNewJob(newJ);
			jMap.put(j, newJ);
		}
		for (Edge e : edges) {
			Edge newE = e.cloneMe(); // new Edge(src, dst, e.inter_rel, e.action_value, nails);
			eMap.put(e, newE);
			JobType src = jMap.get(e.source);
			JobType dst = jMap.get(e.destin);
			newE.source = src;
			newE.destin = dst;
			clone.addEdge(newE);
		}
		// constructing the branch orders
		for (JobType j : jobs) {
			LinkedList<Edge> outEdges = j.getBranchOrders();
			JobType clonedJ = jMap.get(j);
			clonedJ.clearBranchOrders();
			for (int i = 0; i < outEdges.size(); i++) {
				Edge e = eMap.get(outEdges.get(i));
				clonedJ.addEdge(e);
				if(clonedJ != e.source)
					System.out.println("Warning: the source of the " +
							"cloned edge does not match with the cloned job");
			}			
		}
		if (startJob != null)
			clone.startJob = jMap.get(startJob);
		return clone;
	}

	public void trackChange(Change ch) {
		changeTrace.add(ch);
	}

	// called by DrawingPane.undo()
	public void undo() {
		if (changeTrace.size() == 0)
			return;

		Change ch = changeTrace.pollLast();
		undoChange(ch);
	}

	private void undoChange(Change ch) {
		if(ch.isJobChange())
			undoJobChange((JobChange)ch);
		else if(ch.isEdgeChange())
			undoEdgeChange((EdgeChange)ch);	
		else if(ch.isTaskChange())
			undoTaskChange((TaskChange)ch);	
		else if(ch.isGroupChange()) {
			GroupChange gc = (GroupChange)ch;
			ListIterator<Change> it = gc.getReverseIter();

			// Iterate in reverse.
			while(it.hasPrevious()) 
				undoChange(it.previous());	// recursive call
		} else
			System.out.println("Warning: unknwon change to undo" + ch);
	}

	private void undoTaskChange(TaskChange ch) {
		if (ch.isModify()) {
			ch.undoModify();
		} else
			System.out.println("Warning: unimplemented cundo change in undoTaskChange");
	}


	private void undoEdgeChange(EdgeChange ch) {
		if (ch.isModify()) {
			ch.undoModify();
		} else if (ch.isCreate()) {
			removeEdge(ch.getReference());
		} else if (ch.isRemove()) {
			addEdge(ch.getReference());
		}
	}

	private void undoJobChange(JobChange ch) {
		if (ch.isModify()) {
			ch.undoModify();
		} else if (ch.isCreate()) {
			removeJob(ch.getReference());
		} else if (ch.isRemove()) {
			addNewJob(ch.getReference());
		}
	}

	// used when undoing a recent change
	public void loadData(Task orig) {
		this.name = orig.name;
		this.utilization = orig.utilization;
		this.priority = orig.priority;
		this.localVariables = orig.localVariables;
		// this.startJob = orig.startJob;
	}


	public TaskChange getModifyChange() {
		TaskChange ch = createChange();
		ch.setStartJob(startJob);	// this is an exception in creating the changes 
		// since we cannot copy the startJob		
		ch.beModify();
		return ch;
	}

	public TaskChange getDeleteChange() {
		TaskChange ch = createChange();
		ch.beRemove();
		return ch;
	}

	public TaskChange getCreateChange() {
		TaskChange ch = createChange();
		ch.beCreate();
		return ch;
	}

	private TaskChange createChange() {
		TaskChange ch = new TaskChange();
		ch.setRef(this);
		ch.setOrig(this.cloneMe());		
		return ch;
	}

	public int lastChange() {
		if(changeTrace.size() <= 0)
			return -1;
		return changeTrace.getLast().seqNo();
	}

	public String localVars() {
		return localVariables;
	}
	
	public void setLocalVars(String v) {
		localVariables = v;
	}

	public void setmaxEdgeInd(AtomicInteger i) {
		max_edge_index = i;		
	}
}
