package main;


import java.util.LinkedList;
import java.util.ListIterator;

import javax.swing.JFileChooser;
import javax.xml.bind.annotation.adapters.XmlJavaTypeAdapter.DEFAULT;

import taskeditor.changetrace.Change;
import taskeditor.changetrace.GroupChange;
import taskeditor.changetrace.TaskChange;
import util.Dialogs;
import util.GUIUtil;

public class TaskSet {
	
	private String				name;		// task set's name
	private String 				globalVars   = ""; // declaration of global variables
	private String 				contextClaus = ""; // declaration of Ada context clauses
	private String 				mainProcCode = ""; 
	private LinkedList<Task> 	taskList 	 = null;
	private LinkedList<Change>  changeTrace  = new LinkedList<Change>();	
	private	final	String		DEFAULT_NAME = "TaskSet_1";
	
	/*** Constructors ***/
	public TaskSet() {
		taskList = new LinkedList<Task>();
		// simjp.reset();
		this.name = DEFAULT_NAME;
	}
		
	public TaskSet(LinkedList<Task> tasks, String ctxCl, String g, String p, String n) {
		taskList 	= tasks;
		// simjp.reset();
		contextClaus= ctxCl;
		globalVars 	= g;
		mainProcCode= p;
		name 		= n.isEmpty() ? DEFAULT_NAME : n;		
	}
	
	public Task createNewTask(String name) {
		Task newT = new Task(name);
		taskList.add(newT);
		return newT;
	}

	public boolean hasNonEmptyTask() {
		if (taskList == null)
			return false;
		
		for (Task t : taskList)
			if (t.isEmpty() == false)
				return true;
		
		return false;
	
	}

	public void updateTaskName(String oldName, String newName) {
		Task t = getTask(oldName);
		if (t != null)
			t.setName(newName);
		else
			Util.debug("Task with the given name " + oldName + " not found");
	}

	public void setPriority(String tName, int priority) {
		Task task = getTask(tName);
		if (task != null) {
			task.setPriority(priority);
			return;
		}
		
		String s = "Task with the given name " + tName + " not found (when updating the priority)";
		Util.debug(s);
		Util.log(s);
	}

	public boolean remove(String taskName) {
		int ind = getTaskIndex(taskName);
		if (ind < 0)
			return false;
		
		taskList.remove(ind);
		return true;		
	}

	public void assignPriority(String[] orderedNames) {		
		for (int i = 0; i < orderedNames.length; i++)
			setPriority(orderedNames[i], i + 1);
	}

	public void add(Task t) {
		taskList.add(t);		
	}

	public int size() {
		if (taskList == null)
			return 0;
		return taskList.size();
	}

	public Task[] toArray() {
		return taskList.toArray(new Task[]{});
	}
	
	/*** Getters ***/
	public Task get(int taskIndex) {
		if (taskList == null)
			return null;

		if (taskList.size() == 0)
			return null;

		if (taskIndex < 0 || taskIndex >= taskList.size())
			return null;
		
		return taskList.get(taskIndex);
	}

	public int getTaskIndex(String name) {
		if (taskList.size() == 0)
			return -1;
		
		for (int i = 0; i < taskList.size(); i++)
			if (taskList.get(i).getName().equals(name))
				return i;
		
		return -1;
	}

	public Task getTask(String name) {
		for (Task t : taskList)
			if (t.getName().equals(name)) 
				return t;
		
		return null;
	}
	
	public LinkedList<Task> taskList() {		
		return taskList;
	}

	/***************************************
	 * Methods related to Code generation  *
	 ***************************************/
	public String contextClause() {
		return contextClaus;
	}
	
	public void setContextClause(String code) {
		contextClaus = code;
	}
	
	public String mainProcCode() {
		return mainProcCode;		
	}

	public void mainProcCode(String code) {
		mainProcCode = code;		
	}

	public String globalVars() {
		return globalVars;
	}

	public void setGlobalVars(String code) {
		globalVars = code;
	}

	
	/*****************************
	 * Change tracking (for Undo) *
	 ******************************/
	public int latestChange() {
		if (changeTrace.size() <= 0)
			return -1;
		return changeTrace.getLast().seqNo();
	}

	public void undo() {
		if (changeTrace.size() == 0)
			return;
		
		Change ch = changeTrace.pollLast();
		undoChange(ch);			
	}

	private void undoChange(Change ch) {
		if (ch.isGroupChange()) {
			GroupChange gc = (GroupChange)ch;
			ListIterator<Change> it = gc.getReverseIter();

			// Iterate in reverse.
			while (it.hasPrevious()) 
				undoChange(it.previous());	// recursive call
		} else if (ch.isTaskChange())
			undoTaskChange((TaskChange)ch);	
	}

	private void undoTaskChange(TaskChange ch) {
		if (ch.isModify()) {
			ch.undoModify();
		} else if (ch.isRemove()) {
			taskList.add(ch.getReference());
		} else if (ch.isCreate()) {
			remove(ch.getReference().getName());
		} else {
			String msg = "Feature not implemented: Undo a task change";
			Dialogs.showWarnMsg(msg);
			Util.log(msg);
			Util.log(ch.toString());
			System.out.println(ch.toString());
		}	
	}

	public void trackChange(Change ch) {
		changeTrace.addLast(ch);		
	}

	public void setName(String name) {
		this.name = name;
	}
	
	public String getName() {		
		return name;
	}	
}