package main.storage;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ListIterator;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import main.Edge;
import main.JobType;
import main.Task;
import main.TaskSet;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import taskeditor.movables.Nail;


public class TSSaver {

	public static void save(Task[] tasks, TaskSet taskSet, String path) {
		try {
			DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder docBuilder;
	
			docBuilder = docFactory.newDocumentBuilder();
	
			// root elements
			Document doc = docBuilder.newDocument();
			Element rootElement = getRoot(doc, taskSet); 
			doc.appendChild(rootElement);
			rootElement.appendChild(ctxClauseNode(doc, taskSet));
			rootElement.appendChild(globalVarNode(doc, taskSet));
			rootElement.appendChild(mainProcNode(doc, taskSet));
	
			for (Task task : tasks) {
				// task elements
				Element taskNode = doc.createElement("task");
				taskNode.setAttribute("name", task.getName());
				taskNode.setAttribute("priority", "" + task.getPriority());
				taskNode.setAttribute("max_edge_index","" + task.current_max_edge_Index());   // added new attribute
				if (task.getStartJob() != null) 
					taskNode.setAttribute("startJob", task.getStartJob().getName());
				if (task.localVars() != null)
					taskNode.appendChild(taskVarNode(doc, task));
	
				ListIterator<JobType> jItr = task.jobIterator();		    	
				while (jItr.hasNext()) {
					JobType job = jItr.next();
					Element jNode = jobNode(job, doc, task);
					taskNode.appendChild(jNode);
					saveCodeSegment(jNode, task, job, StorageUtil.extractPath(path));					
				}
				
				ListIterator<Edge> itr = task.edgeIterator();  
				while (itr.hasNext()) {
					Edge e = itr.next();
					taskNode.appendChild(edgeNode(e, doc));
				}
				
				rootElement.appendChild(taskNode);
			}
			// write the content into xml file
			TransformerFactory transformerFactory = TransformerFactory.newInstance();
			Transformer transformer = transformerFactory.newTransformer();
			DOMSource source = new DOMSource(doc);
			StreamResult result = new StreamResult(new File(path));
	
			// Output to console for testing
			// StreamResult result = new StreamResult(System.out);
	
			transformer.transform(source, result);
		} catch (TransformerException e) {			
			e.printStackTrace();
		} catch (ParserConfigurationException e1) {
			e1.printStackTrace();
		}
	}

	private static Element getRoot(Document doc, TaskSet taskSet) {
		Element root = doc.createElement("taskSet");
		root.setAttribute("name", taskSet.getName());
		return root;
	}

	private static Node taskVarNode(Document doc, Task t) {
		Element node = doc.createElement("variables");
		node.appendChild( doc.createTextNode(t.localVars()) );
		return node;
	}

	private static Element jobNode(JobType job, Document doc, Task t) {
		Element jobNode = doc.createElement("jobType");
		setJobAttr(jobNode, job);
		
		Element nameNode = doc.createElement("name");
		nameNode.appendChild(doc.createTextNode(job.getName()));
		jobNode.appendChild(nameNode);
		
		Element WCETNode = doc.createElement("WCET");
		WCETNode.appendChild( doc.createTextNode(num2str(job.WCET)) );
		jobNode.appendChild(WCETNode);
		
		Element dlNode = doc.createElement("deadline");
		dlNode.appendChild( doc.createTextNode(num2str(job.deadline)) );
		jobNode.appendChild(dlNode);
		return jobNode;
	}

	private static Node edgeNode(Edge e, Document doc) {
		Element depNode = doc.createElement("dependency");
		Element srcNode = doc.createElement("src");
		Element dstNode = doc.createElement("dst");
		Element indNode = doc.createElement("edge_index");
		Element irlNode = doc.createElement("interRelease");
		Element actNode = doc.createElement("action");
		Element grdNode = doc.createElement("guard");
		Element codNode = doc.createElement("code");
		Element nlsNode = createNailsNode(e, doc);               
		srcNode.appendChild( doc.createTextNode(e.srcName()) );
		dstNode.appendChild( doc.createTextNode(e.dstName()) );
		indNode.appendChild( doc.createTextNode(num2str(e.index())) );        // added new
		irlNode.appendChild( doc.createTextNode(num2str(e.interRel())) );
		actNode.appendChild(doc.createTextNode(e.action()));    // added new
		grdNode.appendChild(doc.createTextNode(e.guard()));
		codNode.appendChild(doc.createTextNode(e.code()));					
		
		actNode.setAttribute("type", String.valueOf(e.actionType()));
		if (e.sends())
			actNode.setAttribute("semantics", String.valueOf(e.sendSemantic()));
		if (e.getLabel().hasValidPosition()) {
			irlNode.setAttribute("x", num2str(e.getLabel().getXY().x)); 
			irlNode.setAttribute("y", num2str(e.getLabel().getXY().y));
		}

		depNode.appendChild(srcNode);
		depNode.appendChild(dstNode);
		depNode.appendChild(indNode);
		depNode.appendChild(irlNode);
		depNode.appendChild(grdNode);
		depNode.appendChild(codNode);
		if (e.hasAction())
			depNode.appendChild(actNode);
		if (nlsNode != null)
			depNode.appendChild(nlsNode);	
		return depNode;
	}

	private static Element ctxClauseNode(Document doc, TaskSet ts) {
		Element node = doc.createElement("contextclause");
		node.appendChild( doc.createTextNode(ts.contextClause()) );
		return node;
	}

	private static Element globalVarNode(Document doc, TaskSet ts) {
		Element node = doc.createElement("globals");
		node.appendChild( doc.createTextNode(ts.globalVars()) );
		return node;
	}
	
	private static Element mainProcNode(Document doc, TaskSet ts) {
		Element node = doc.createElement("procedure");
		node.appendChild( doc.createTextNode(ts.mainProcCode()) );
		return node;
	}
	
	static void setJobAttr(Element jobNode, JobType job) {
		jobNode.setAttribute("x", num2str(job.x));
		jobNode.setAttribute("y", num2str(job.y));
		jobNode.setAttribute("labelx", num2str(job.getLabelXY().x));
		jobNode.setAttribute("labely", num2str(job.getLabelXY().y));
	}

	static void saveCodeSegment(Element jobNode, Task task, JobType job, String path) {
		if (job.getCode() == null)
			return;
		
		String file = getFileName(task, job);
		jobNode.setAttribute("path", file);
		String jpath = path + file;
		save(jpath, job.getCode());				
	}

	static Element createNailsNode(Edge e, Document doc) {
		Nail[] nails = e.nails();
		if (nails.length == 0)
			return null;
		String xs = "";
		String ys = "";
		int len = nails.length;
		for (int i = 0; i < len - 1; i++) {
			xs += nails[i].x() + ",";
			ys += nails[i].y() + ",";
		}
		xs += nails[len-1].x();
		ys += nails[len-1].y();
		Element nlsNode = doc.createElement("nails");
		nlsNode.setAttribute("xs", xs);
		nlsNode.setAttribute("ys", ys);
		return nlsNode;
	}

	// writes the given string (code) to the specified file (i.e., jpath)
	private static void save(String jpath, String code) {
		File file = new File(jpath);
		FileOutputStream fos;
		try {
			fos = new FileOutputStream(file, false); // false to overwrite.
			fos.write(code.getBytes());
			fos.close();
		} catch (FileNotFoundException e1) {
			e1.printStackTrace();
		} catch (IOException e1) {
			e1.printStackTrace();
		}		
	}

	static String getFileName(Task task, JobType job) {
		String path =  task.getName() + "." + job.getName() + ".adb"; //  for saving the code segment associated to this job 
		return path;
	}
	
	private static String num2str(int num) {
		return String.valueOf(num);
	}
}
