package simulator;

import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Timer;
import java.util.TimerTask;

import main.Edge;
import main.Manager;
import main.Task;
import main.Util;
import simulator.view.SimView;
import simulator.engine.Execution;
import simulator.engine.SimEngine;
import util.SchedPolicy;
import util.SyntaxChecker;

public class SimManager {
	private SimEngine	engine;
	private SimView		viewer;
	private Manager		manager;

	private	Timer		timer;
	private	RunMode		runMode;
	private	boolean		stopped;
	private	boolean		resetFlag;		// indicates whether the simulator must be reset in the next resumption
	private double		sim_time_dstep;	// distance between running two simulation steps (milliseconds)
		
	public SimManager(Manager m, SimView v, int speed) {
		manager = m;
		viewer  = v;
		engine  = new SimEngine();
		stopped = true;
		resetFlag = false;
		sim_time_dstep = 1000;
		schedule_sim(speed);
	}

	public void tabSelected(SimSettings settings) {
		stopSim();
		reset(settings);
	}

	protected void reset_on_demand(SimSettings settings) {
		if (resetFlag == true)
			reset(settings);
		resetFlag = false;		
	}

	public void stop(SimSettings settings) {
		stopSim();	
		reset(settings);	// called to erase the existing simulation area
		resetFlag = true;		
	}

	public void run(SimSettings settings) {
		if (checkSyntax()) {
			resetRunModeP( settings.policy );	
			startSim(settings);
		}
	}

	public void Run2DLmiss(SimSettings settings) {
		if (checkSyntax()) {				
			runMode = RunMode.TO_DL_MISS;	
			engine.setSchPolicy(settings.policy);
			startSim(settings);
		}
	}

	public void step(SimSettings settings) {
		/*if (engine.hasAnyEvent() == false)
			return;
		*/
		if (checkSyntax()) {
			reset_on_demand(settings);
			runOneStep();
			stopSim();
		}
	}
	
	private void reset(SimSettings settings) {
		Task[] tasks = manager.taskList();
		engine.reset(tasks, settings);
		viewer.reset(getTaskNames());
	}
	
	private boolean checkSyntax() {
		Task[] tasks = manager.taskList();
		LinkedList<String> errors = SyntaxChecker.checkForSimulation(tasks);
		if (errors.size() > 0) {
			Util.showSyntaxErrWarningDialog(manager.frame(), errors);
			return false;
		}
		return true;
	}
	
    private String[] getTaskNames() {
    	return engine.getTaskNames();
	}
    

	public void resetRunModeP(SchedPolicy policy) {
		runMode = RunMode.NONE;	
		engine.setSchPolicy(policy);
	}
	
	public void resetRunMode() {
		runMode = RunMode.NONE;	
	}
   	
	public void stopSim() {
		resetRunMode();
		stopped = true;		
	}
	
	private void startSim(SimSettings settings) {
		reset_on_demand(settings);		
		stopped = false;
	}
	
	public void setSpeed(int sim_speed) {
		schedule_sim(sim_speed);
	}
	
	public boolean runOneStep() {
		LinkedList<Execution> new_execs = engine.run_one_step();
		if (new_execs.size() > 0) {
			double wl = engine.currentWrkLoad();
			int time = engine.getCurrTime();
			viewer.addNewExecs(new_execs, wl, time);			
		}
		return allDLMeet(new_execs);
	}
	
	private boolean allDLMeet(LinkedList<Execution> new_execs) {
		for(int i = 0; i < new_execs.size(); i++)
			if(new_execs.get(i).type == EXEC_TYPE.DLMiss)
				return false;
		return true;
	}


	/*
	 * (re)schedules the simulation
	 * This is called in two situations: 
	 * 		- when initializing in constructor
	 * 		- when changing the simulation speed by the user
	 */
	public void schedule_sim(int sim_speed) {
		if (timer != null)
			timer.cancel();
		timer = new Timer();
		TimerTask task = new TimerTask() {
			  public void run() {					
				  if (stopped == false) {
						if (engine.hasAnyFutureEvent() == false) {
							stopSim();
							manager.printNote("No progress possible, simulation stopped");
							engine.printWaitingTasks();
						}							
						else {
							boolean allDLMeet = runOneStep();
							if (allDLMeet == false && runMode == RunMode.TO_DL_MISS) {
								resetRunMode();
								stopSim();
							}
						}							  
				  }
				  viewer.refresh();
			  }
			}; 
		timer.schedule(task, 0, (int)(sim_time_dstep / sim_speed));
	}
}

enum RunMode {
	TO_DL_MISS, NONE
}