package simulator;

import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.plaf.SliderUI;

import main.Manager;
import main.Util;

import simulator.simenums.SimInitMethod;
import simulator.simenums.SimMethod;
import simulator.view.SimView;
import simulator.view.WorkloadView;
import util.Dialogs;
import util.SchedPolicy;

import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;


/**
 * For controlling and showing the simulation
 */
public class SimTab extends JPanel implements ComponentListener {
	
	private SimManager		mgr;
	private SimView			viewer;
	private JSplitPane 		splitV;
	private SimSettings		settings;
	
	private	final int 		INIT_SPEED = 2;	
		
	public SimTab(Manager manager) {
		settings = defaultSetting();
		WorkloadView wlView = new WorkloadView();
		viewer = new SimView(wlView);
		mgr    = new SimManager(manager, viewer, INIT_SPEED);
		
		JToolBar toolbar = new SimToolBar(this);
		splitV = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
		splitV.setLeftComponent(viewer.getScroll());
		splitV.setRightComponent(wlView.getScroll());
		
		setLayout(new BorderLayout());
		this.add(toolbar, BorderLayout.PAGE_START);
		this.add(splitV, BorderLayout.CENTER);		
		addComponentListener(this);
	}

	private SimSettings defaultSetting() {
		return new SimSettings(SchedPolicy.EDF, SimInitMethod.Zero, 
				SimMethod.MaxProgress, INIT_SPEED);
	}

	/*
     * called when the simulator tab is selected
     */
	public void selected() {
		mgr.tabSelected(settings);
		rearrange();
	}
	
	/*
	 * called when:
	 * 1. resized;
	 * 2. the whole frame minimized/maximized (called by JManager)
	 */
	public void rearrange() {
		int h = splitV.getHeight();
		int h2 = viewer.getWLViewMinHeight();
		Util.decorateSplit(splitV, h - h2);
	}
	
	public void speedChanged(int s) {
        mgr.setSpeed(s);
	}
	
	public void repaintViewer() {
		
	}
	
	@Override
	public void componentHidden(ComponentEvent e) {
		// TODO Auto-generated method stub		
	}

	@Override
	public void componentMoved(ComponentEvent e) {
		// TODO Auto-generated method stub		
	}

	@Override
	public void componentResized(ComponentEvent e) {
		rearrange();
	}

	@Override
	public void componentShown(ComponentEvent e) {
		// TODO Auto-generated method stub		
	}
	
	public void stopSim() {
		mgr.stopSim();		
	}	
	
	/*
	 * This class stops the simulation
	 */
	public void pause() {
		stopSim();
	}
	
	public void run() {
		mgr.run(settings);		
	}
	
	/*
	 * For stopping and clearing the simulation
	 */
	public void stop() {
		mgr.stop(settings);		
	}		
	
	/*
	 * This class is used for running the simulation for one step
	 */
	public void step() {
		mgr.step(settings);
	}		
	
	/*
	 * This class is used for running the simulation until the first next deadline miss
	 */
	public void Run2DLmiss() {
		mgr.Run2DLmiss(settings);
	}

	public void getUserSetting() {
		settings = SettingsPanel.getUserSettings(settings, this);
		speedChanged(settings.simSpeed);
	}	
}   

class SettingsPanel extends JPanel implements ChangeListener {
	
	private	JComboBox<SimInitMethod>simInitCB;
	private	JComboBox<SchedPolicy>  schdPolCB;
	private JSlider					slider;
	private SimTab 					parent;
	private	final int 				MIN_SPEED  = 1;
	private	final int 				MAX_SPEED  = 10;
	
	public static SimSettings getUserSettings(SimSettings cfg, SimTab simTab) {
		SettingsPanel p = new SettingsPanel(cfg, simTab); 
		String title = "Simulation Settings"; 
		int res = Dialogs.showOkCancelPanel(p, title);		
        if (res == JOptionPane.OK_OPTION) 
        	return p.getSettings();
        else
        	return cfg;
	}
	
	private SettingsPanel(SimSettings cfg, SimTab simTab) {
		parent = simTab;
		
		Box up = constructCombosBox(cfg);
		Box dn = constructSliderBox(cfg);		
		
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
		add(up);
		add(Box.createVerticalStrut(20));
		add(new JSeparator(SwingConstants.HORIZONTAL));
		add(Box.createVerticalStrut(20));
		add(dn);
		add(Box.createVerticalStrut(10));
	}
	
	private Box constructCombosBox(SimSettings cfg) {
		simInitCB = new JComboBox<SimInitMethod>();
		schdPolCB = new JComboBox<SchedPolicy>();		
		simInitCB.addItem(SimInitMethod.Zero);
		simInitCB.addItem(SimInitMethod.Random);
		schdPolCB.addItem(SchedPolicy.SP);
		schdPolCB.addItem(SchedPolicy.EDF);
		
		simInitCB.setSelectedItem(cfg.initMethod);
		schdPolCB.setSelectedItem(cfg.policy);
		
		JPanel p = new JPanel(new GridLayout(2, 2, 5, 5));
		p.add(new JLabel("Initial Phase:"));
		p.add(simInitCB);		
		p.add(new JLabel("Scheduling Policy: "));
		p.add(schdPolCB);
		
		Box b = Box.createHorizontalBox();
	    b.add(p);
	    b.add(Box.createHorizontalGlue());
		return b;
	}
	
	private Box constructSliderBox(SimSettings cfg) {		
        slider = new JSlider( JSlider.HORIZONTAL,
							MIN_SPEED, MAX_SPEED, cfg.simSpeed);
        slider.addChangeListener(this);		
        slider.setMajorTickSpacing(1);
        slider.setMinorTickSpacing(1);
        slider.setPaintTicks(true);
        slider.setPaintLabels(true);
        
		Box b = Box.createHorizontalBox();
	    b.add( new JLabel("Simulation Speed:  "));
	    b.add( slider );
	    return b;
	}

	public SimSettings getSettings() {
		SimSettings s = new SimSettings();
		s.policy 	  = (SchedPolicy) schdPolCB.getSelectedItem();
		s.initMethod  = (SimInitMethod) simInitCB.getSelectedItem();
		s.simSpeed    = (int)slider.getValue();
		s.method      = SimMethod.MaxProgress;
		return s;
	}

	@Override
	public void stateChanged(ChangeEvent e) {
		JSlider source = (JSlider)e.getSource();
	    if (!source.getValueIsAdjusting()) {
	        int s = (int)source.getValue();	// sim speed
	        parent.speedChanged(s);
	    }		
	}
}	