package simulator.engine;

import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Random;


import enums.ActionType;

import main.Edge;
import main.EdgeParam;
import main.JobType;
import main.Task;
import main.Util;

import simulator.simenums.SimInitMethod;
import simulator.EXEC_TYPE;
import simulator.engine.SimEngine.EventType;
import simulator.engine.SimEngine.Status;

class STask implements Comparable<STask> {
	private	Status 	state;
	private	SJob 	currJob;		//  the last/next released job
	private	SJob 	prevJob;		//  the last/next released job
	private	Edge 	currEdge;	//  the last/next released edge
	private	SEvent	nextEvent;
	private	int		prio;			// a unique number
	private	Task	myTask;
	private	int		lastUpdate;		// the last time the task event was processed
	private	boolean	is_running;
	private	Random	rng;			// random generator
	protected int	id;				// a unique number (for painting purposes)
	private LinkedList<Edge> candidEdges;
	
	public STask(Task master, int id, Random randGen, SimInitMethod initMethod) {
		myTask  = master;
		rng 	= randGen;
		prio 	= master.getPriority();
		this.id = id;
		init(initMethod);
	}

	private void init(SimInitMethod initMethod) {
		candidEdges = new LinkedList<Edge>(); 
		lastUpdate  = 0;
		if ( myTask.isEmpty() )	{ // this task has no job		
			nextEvent = createNullEvent();
		} else {
			currJob	    = getFirstJob(lastUpdate);
			int phase	= offSet(initMethod, currJob.absDLine());
			currJob.arrival = lastUpdate + phase;
			nextEvent	= createArrivEvent(currJob.arrival);
		}
		state = Status.FINISHED;
		is_running = false;
	}
	
	private int offSet(SimInitMethod initMethod, int max) {
		int offset = 0;
		switch (initMethod) {
		case Zero:
			offset = 0;
			break;

		case Random:
			offset = ((int)(Math.random() * max)); // a random release time for the first job;
			break;
		}
		return offset;
	}

	private SJob getFirstJob(int arrival) {
		int numOfJobs = myTask.jobNo();
		if (numOfJobs == 0)
			return null;
		
		JobType nextJ = myTask.getStartJob();
		if (nextJ == null) {
			int jId = rng.nextInt(numOfJobs);
			nextJ = myTask.getJob(jId);
		}			
		return new SJob(nextJ , arrival);
	}
	
	private SJob getNextJob(Edge e, int prevArriv) {
		return new SJob(e.destin(), prevArriv + e.interRel());
	}

	private Edge getRandEdge(SJob job) {
		LinkedList<Edge> out_edges = outgoingEdges(job);
		int out_degree = out_edges.size();
		if (out_degree == 0)	// the edge has not successor 
			return null;
		int randInd = rng.nextInt(out_edges.size()); 
		return out_edges.get(randInd);
	}
	
	private LinkedList<Edge> outgoingEdges(SJob j) {
		JobType src = j.masterJob;
		ListIterator<Edge> itr = myTask.edgeIterator();		
		LinkedList<Edge> res = new LinkedList<Edge>();
		
		while (itr.hasNext()) {
			Edge e = itr.next();
			if (e.source().equals(src))
				res.add(e);
		}
		return res;
	}

	// progresses the time by handling the current event
	// always called at an event time, namely we have curr_time = nextEvent.time,
	// therefore, a new event has to be generated after progressing the time and handling the current event
	public LinkedList<Execution> progress(int curr_time, Boolean maxProg) {
		if(curr_time != nextEvent.time)
			System.out.println("Warning! current time: " + curr_time + 
					" different than the event time: " + nextEvent.time);
		
		LinkedList<Execution> execs = new LinkedList<Execution>(); 
		int elapsed = curr_time - lastUpdate;
		if (is_running)
			currJob.progress(elapsed);
			
		switch (nextEvent.type) {
			case ARRIVAL:
				String label = arrivalLabel();
				execs.add(new Execution(getId(), curr_time, elapsed, label, EXEC_TYPE.Arrival));
				currJob.arrival = curr_time;
				nextEvent = new SEvent(EventType.DLMISS, this, currJob.absDLine());
				state = Status.READY;
				break;
			
			case DLMISS:
				if (is_running)
					execs.add(newExecution(elapsed));
				if (currJob.remainedExecution > 0)		// to avoid deadline miss of jobs with 0 execution time!
					execs.add(new Execution(getId(), curr_time, elapsed, currJobName(), EXEC_TYPE.DLMiss));			
				setNextJobAndEvent(maxProg, curr_time);
				state = Status.FINISHED;
				break;
			
			case FINISH:
				test1();
				execs.add(newExecution(elapsed));
				setNextJobAndEvent(maxProg, curr_time);					
				state = Status.FINISHED;
				break;
		}
		lastUpdate = curr_time;
		return execs;
	}
	
	private Execution newExecution(int elapsed) {
		return new Execution(getId(), lastUpdate, elapsed, currJobName(), EXEC_TYPE.Execution);
	}

	protected String arrivalLabel() {
		String s = "";
		if (currJob != null)
			s = currJobName();
		else
			s = prevJob.masterJob.getName();
		if (currEdge != null) {
			String a = currEdge.action();
			if (currEdge.sends()) 
				s += "[" + a + "!]";
			else if (currEdge.receives())
				s += "[" + a + "?]";
		}
		return s; 
	}

	private void setNextJobAndEvent(Boolean maxProg, int now) {
		prevJob  = currJob;
		currJob  = null;
		currEdge = null;
		candidEdges.clear();
		if (hasNextJob(prevJob)) {						
			if (maxProg) {	// choose the first enabled
				nextEvent = createArrivEvent(now);
				LinkedList<Edge>newCandids = eligibleEdges(prevJob, now);
				addNewCandids(newCandids, now);								
			} else {
				takeEdge(getRandEdge(prevJob), prevJob.arrival);
			}			 
		} else { 
			nextEvent = createNullEvent(); 
		}		
	}

	private void takeEdge(Edge e, int arrival) {
		currEdge  = e;
		currJob   = getNextJob(e, arrival);
		nextEvent = createArrivEvent(currJob.arrival);			
	}

	private Edge firstFutureEdge(SJob job, int now) {
		Edge first = null;
		LinkedList<Edge> edges = outgoingEdges(job);
		for (Edge e : edges) {
			if (job.arrival + e.interRel() > now) {
				if (first == null)
					first = e;
				else
					first = e.interRel() < first.interRel() ? e : first;
			}
		}
		return first;
	}

	private Edge firstNonReceiving(LinkedList<Edge> edges) {
		Edge first = null;
		for (Edge e : edges) {
			if (e.receives() == false) {
				if (first == null)
					first = e;
				else
					first = e.interRel() < first.interRel() ? e : first;
			}
		}
		return first;
	}

	private LinkedList<Edge> eligibleEdges(SJob job, int now) {
		LinkedList<Edge> res = new LinkedList<Edge>();		
		LinkedList<Edge> out_edges = outgoingEdges(job);
		for (Edge e : out_edges) {
			if (job.arrival + e.interRel() <= now) 
				res.add(e);			
		}
		return res;
	}
	
	public void updateCandids(int now) {
		if (hasCurrJob())
			return;
		if (hasNextJob(prevJob) == false)
			return;

		if (now < nextEvent.time)	// no new edge has been eligible
			return;
		if (now > nextEvent.time)	// all edges already checked
			return;
		
		LinkedList<Edge>newCandids = eligibledNow(prevJob, now); 				
		if (newCandids.size() == 0) {
			Util.simDebug("*** Warning: There should be somethind eligible " + 
					myTask.getName() + " " + now + " " + prevJob.masterJob.getName());
			return;
		}			
		
		addNewCandids(newCandids, now);
	}
	
	private void addNewCandids(LinkedList<Edge> newCandids, int now) {
		Edge first = firstNonReceiving(newCandids);
		if (first != null)
			takeEdge(first, prevJob.arrival);
		else {
			candidEdges.addAll(newCandids);
			Edge firstFuture = firstFutureEdge(prevJob, now);
			if (firstFuture != null)
				nextEvent = createArrivEvent(prevJob.arrival + firstFuture.interRel());
		}
	}

	private LinkedList<Edge> eligibledNow(SJob job, int now) {
		LinkedList<Edge> eligibles = eligibleEdges(job, now);
		LinkedList<Edge> res = new LinkedList<Edge>();
		for (Edge e : eligibles)
			if (job.arrival + e.interRel() == now)
				res.add(e);
		return res;		
	}

	private boolean hasNextJob(SJob j) {
		if (j == null)
			return false;
		if (outgoingEdges(j).size() > 0)	 
			return true;		
		return false;
	}

	/*
	 * Null event means no event (there is no job to be released)
	 */	
	private SEvent createNullEvent() {
		return new SEvent(EventType.NONE, this, Integer.MAX_VALUE);
	}

	private SEvent createArrivEvent(int arrival) {
		return new SEvent(EventType.ARRIVAL, this, arrival);
	}

	/*
	 * This is called when this task has been scheduled to be executed
	 */
	protected void start_running(int now) {
		is_running = true;
		state = Status.READY;
		int finish_time = now + currJob.remainedExecution;
		if(finish_time <= currJob.absDLine())
			nextEvent = new SEvent(EventType.FINISH, this, finish_time);
		else
			nextEvent = new SEvent(EventType.DLMISS, this, currJob.absDLine());		
		lastUpdate = now;
	}
	
	/*
	 * This is called when a higher priority task is preempting this one
	 */
	protected LinkedList<Execution> stop_running(int now) {
		if (is_running == false)
			System.out.println("Warning : task not running asked to be stopped");
		is_running = false;
		
		LinkedList<Execution> result = new LinkedList<Execution>();
		if (state == Status.FINISHED)
			return result;		// already stopped (via function progress)		
				
		int elapsed = now - lastUpdate;
		result.add(new Execution(id, lastUpdate, elapsed, currJobName(), EXEC_TYPE.Execution));
		lastUpdate = now;		
		
		if (currJob.remainedExecution > 0) {
			nextEvent = new SEvent(EventType.DLMISS, this, currJob.absDLine());
			state = Status.READY;
			currJob.progress(elapsed);
		}
		else
			state = Status.FINISHED;
		return result;
	}

	// checks some property of correctness of the implementation
	private void test1() {
		if (is_running == false)
			System.out.println("Warning in Simulation: job finished while not running");
	}

	public void setId(int id) {
		this.id = id;
	}

	public int getId() {
		return id;
	}
	
	public int getPrio() {
		return prio;
	}

	/* 
	 * returns a string describing the currentJob
	 */
	public String jobStr() {
		return currJob.toString();
	}

	public boolean isReady() {
		return state == Status.READY;
	}

	public String name() {
		return myTask.getName();
	}
	
	public String currJobName() {
		return currJob.masterJob.getName();
	}

	public int compareTo(STask other) {
		if (this.prio > other.prio)
			return 1;
		else if (this.prio == other.prio)
			return 0;
		else
			return -1;
	}
	
	public int getCurrDeadLine() {
		if (currJob == null)
			return Integer.MAX_VALUE;
		return currJob.absDLine();
	}

	public double currentWL(int currTime) {
		if (isReady() == false || currJob == null)
			return 0;
		
		double e = currJob.remainedExecution;
		double relDL = currJob.absDLine() - currTime;
		
		if(relDL == 0)
			return 0;
		return e / relDL;
	}
	
	public double getUrgency(int currTime) {
		if (isReady() == false || currJob == null)
			return 0;
		
		double e = currJob.remainedExecution;
		double relDL = currJob.absDLine() - currTime;
		
		if (relDL == 0)
			return 0;
		if (e >= relDL)
			return 1;
		
		return e / relDL;
	}

	public EdgeParam edgeProps() {
		if (currEdge == null)
			return null;
		return currEdge.params();		
	}

	public boolean receives(String action) {
		return myTask.receives(action);
	}

	public boolean waitingForReceive(String act, int time) {
		if (!readyToSynch(time))
			return false;
	
		if (currEdge != null) { 
			EdgeParam p = currEdge.params();		
			if (p.aType == ActionType.Receive && p.action.equals(act))
				return true;
			else
				return false;			
		} else {
			for (Edge e : candidEdges) {
				if (canSync(e, act, time, ActionType.Receive))
					return true;
			}
		}
		return false;		
	}
	
	private boolean canSync(Edge e, String act, int time, ActionType type) {
		EdgeParam p = e.params();		
		if (p.aType == type && p.action.equals(act) && 
				prevJob.arrival + e.interRel() <= time)
			return true;
		return false;
	}

	private boolean readyToSynch(int time) {
		if (nextEvent.type != EventType.ARRIVAL)
			return false;
		if (hasPrevJob() == false)
			return false;
		
		if (currEdge != null) { // nondeterministic method				
			if (nextEvent.time > time)
				return false;
			if (currEdge.params().aType == ActionType.None)
				return false;
			return true;
		} else {
			if (candidEdges.size() > 0)
				return true;
			else
				return false;
		}
	}

	public boolean hasCurrJob() {
		return currJob != null;
	}

	public boolean hasPrevJob() {
		return prevJob != null;
	}

	public void receive(String a) {
		for (Edge e : candidEdges)
			if (e.receives())
				if (e.params().action.equals(a)) {
					takeEdge(e, prevJob.arrival);					
					candidEdges.clear();
					return;
				}					
	}
	
	public SEvent nextEvent() {
		return nextEvent;
	}
}
