package simulator.engine;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.PriorityQueue;
import java.util.Random;

import enums.ActionType;

import main.EdgeParam;
import main.Task;
import main.Util;

import simulator.simenums.SimInitMethod;
import simulator.simenums.SimMethod;
import simulator.SimSettings;
import util.SchedPolicy;

public class SimEngine {
	
	private	Random 				rng;
	private	LinkedList<STask> 	tasks;
	private	STask 				runningTask;
	private	int 				now;
	private SchedPolicy 		schPolicy;
	private	SimMethod			method;	
	
	public SimEngine() {
		now = -1;
		int SEED = 100;		// random generation seed
		rng = new Random(SEED);
	}
	
	/*
	 * runs one step of simulation
	 */
	public LinkedList<Execution> run_one_step() {
		Util.simDebug("Running one step " + now);
		
		LinkedList<Execution> runs = new LinkedList<Execution>(); 
		if (noTask())  
			return runs;
		
		if (hasAnyFutureEvent() == false)  
			return runs;
		
		now = earliestFutureEvent().time;
		updateAll(now);	// update tasks' candidate list		
		LinkedList<SEvent> events = collectDoableEvents(now);
		Util.simDebug("Doable events: " + events.size());
		while (events.size() > 0) {
			for (SEvent e : events) {
				runs.addAll( handle(e) );
			}
			runs.addAll( schedule() );
			events = collectDoableEvents(now);	// collect possibly newly generated event
		}			
		return runs;
	}

	private LinkedList<Execution> schedule() {
		LinkedList<Execution> runs = new LinkedList<Execution>(); 
		if (runningTask != null) 
			runs.addAll(runningTask.stop_running(now));
		
		// schedule
		STask highestPrio = getEligibleTask(); 
		if (highestPrio != null) {
			highestPrio.start_running(now);
			Util.simDebug("Task " + highestPrio.name() + " scheduled to run at " + now);
		}
		runningTask = highestPrio;
		return runs;
	}

	private void updateAll(int now) {
		if (method == SimMethod.MaxProgress) { 
			for (STask t : tasks) 
				t.updateCandids(now);
		}
	}

	private Collection<Execution> handle(SEvent e) {
		Util.simDebug("Handling event: " + e.toString());
		boolean maxProg = (method == SimMethod.MaxProgress);
		return e.handle(now, maxProg);
	}
	
	private void ackReceivers(String a) {
		for (STask st : tasks)
			st.receive(a);
	}	

	/*
	 * Among the tasks which have non-complete work, returns the task which is of highest priority  
	 */
	private STask getEligibleTask() {
		STask elig = null;
		for (STask task : tasks) {
			if (task.isReady()) {
				if (elig == null)
					elig = task;
				else 
					elig = higherPriority(elig, task);				
			}
		}
		return elig;
	}

	private STask higherPriority(STask t1, STask t2) {
		STask higher = t1;	// a default task
		if (schPolicy == SchedPolicy.EDF) { // earliest deadline first
			if (t1.getCurrDeadLine() > t2.getCurrDeadLine())  // earlier deadline means higher priority
				higher = t2;
			else
				higher = t1;	
		} else {	// default is SP (static priority)		
			if (t1.getPrio() > t2.getPrio())  // lower value means higher priority
				higher = t2;
			else
				higher = t1;
		}
				
		return higher;
	}

	private SEvent earliestFutureEvent() {
		PriorityQueue<SEvent> eventQ = new PriorityQueue<SEvent>();
		for (STask task : tasks) {
			if (task.nextEvent().time > now)
				eventQ.add(task.nextEvent());
		}
		SEvent pop = eventQ.peek();
		return pop;
	}

	/*
	 * Returns all available events happending at time "time"
	 */
	private LinkedList<SEvent> collectDoableEvents(int time) {
		LinkedList<SEvent> result = new LinkedList<SEvent>(); 
		for (STask task : tasks) {	
			SEvent e = task.nextEvent();
			if (isDoable(e, time))
				result.add(e);
		}		
		return result;
	}

	private boolean isDoable(SEvent e, int time) {
		if (e.isNull())
			return false;
		if (e.time > time)
			return false;
		if (e.type != EventType.ARRIVAL)
			return true;
		if (e.hasCurrJob() == false)
			return false;
		if (e.hasPrevJob() == false)
			return true;	// this is the first job
		EdgeParam p = e.currEdgeProps();
		
		switch (p.aType) {
		case None:
			return true;
			
		case Send:
			switch (p.send) {
			case Non_Blocking:
				ackReceivers(p.action);
				return true;
			case Blocking:
				if (allReceiversReady(p.action, time)) {
					ackReceivers(p.action);
					return true;
				} else
					return false;
			}
			break;
			
		case Receive:
			// A receiver with determined currJob has been already selected by the sender 
			return true;
		}
		System.out.println("Warning: This code is not supposed to be reachable.");
		return true;
	}

	private boolean allReceiversReady(String action, int time) {
		for (STask s : tasks) {
			if (s.receives(action) &&
				!s.waitingForReceive(action, time))
					return false;
		}
		return true;
	}

	public void setTaskSet(Task[] taskSet, SimInitMethod initMethod) {
		tasks = new LinkedList<STask>();
		int id = 0;
		for (Task task : taskSet)
			tasks.add(new STask(task, id, rng, initMethod));
		Collections.sort(tasks);
		// assign id according to priority order
		for (STask task : tasks)
			task.id = id++;		
	}
	
	enum Status  {
		READY,		// having a non complete job  
		FINISHED	// currently no job for execution (due to execution completion or deadline miss) 
	}
	
	enum EventType {
		ARRIVAL, DLMISS, FINISH, NONE;
	}

	public int taskNo() {
		if (noTask())
			return 0;
		return tasks.size();
	}

	public String[] getTaskNames() {
		if (tasks == null)
			return new String[0];
		
		String[] names = new String[tasks.size()];
		for (int i = 0; i < tasks.size(); i++)
			names[i] = tasks.get(i).name();
		
		return names;
	}

	public int taskSetSize() {
		if (tasks == null)
			return 0;
		return tasks.size();
		
	}

	public void setSchPolicy(SchedPolicy policy) {
		schPolicy = policy;		
	}
	
	public int getCurrTime() {
		return now;
	}

	
	public double currentWrkLoad() {
		double utilz = 0; // utilization
		if (noTask())
			return utilz;
		
		for (STask t : tasks) {	
			if (t.isReady())
			//	utilz += t.currentWL(current_time);
				utilz += (t.currentWL(now) > 0) ? 1 : 0;			
		}
		return utilz / tasks.size();
	}
	
	public double currentLoad() {
		double utilz = 0; // utilization
		if (noTask())
			return utilz;
		
		for (STask t : tasks) {	
			if (t.isReady())
				utilz = Math.max(utilz, t.getUrgency(now) );			
		}
		return utilz;
	}
	
	private boolean noTask() {
		if (tasks == null)
			return true;
		if (tasks.size() == 0)
			return true;
		return false;
	}

	// to detect the end of simulation (i.e. no further job can be released)
	public boolean hasAnyFutureEvent() {
		if (tasks == null)
			return false;
		
		for (STask task : tasks) {	
			SEvent e = task.nextEvent();
			if (e.notNull() && e.time > now)
				return true;
		}
		return false;
	}

	public void reset(Task[] tasks, SimSettings settings) {
		setTaskSet( tasks, settings.initMethod );
		setSchPolicy( settings.policy );
		method = settings.method; 
		now = -1;
	}

	public void printWaitingTasks() {
		System.out.println("---------------");
		for (STask task : tasks) {	
			SEvent e = task.nextEvent();
			if (e.notNull())
				e.print();
		}
		System.out.println("---------------\r\n");
	}	
}
