package synthesis;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dialog;

import javax.swing.JOptionPane;
import javax.swing.JPanel;


import services.RunResult;
import synthesis.model.Settings;
import synthesis.model.SynthesisMgr;
import synthesis.view.CodePane;
import util.Dialogs;
import util.GUIUtil;

import main.Manager;
import main.NotesPanel;
import main.Util;

public class SynthesisTab extends JPanel {
	
	private	ToolBar			toolbar;
	private	CodePane		codePane;
	private	NotesPanel		notePane;
	private	SettingsGUI		confPane;
	
	private SynthesisMgr	synthMgr;
	private	String			savePath;
	
	public SynthesisTab(Manager m) {
		setNotePane(m);
		createToolBar();
		createCodePane();
		createConfPane();
		createSynthsizer(m);		
		setLanguage();
		
		setLayout();
		add(toolbar, BorderLayout.PAGE_START);
		add(codePane, BorderLayout.CENTER);
	}

	private void setNotePane(Manager m) {
		notePane = m.frame().notesPanel();
	}

	private void createToolBar() {
		toolbar = new ToolBar(this); 
	}
	
	private void createCodePane() {
		codePane = new CodePane();
	}	
	
	private void createConfPane() {
		confPane = new SettingsGUI();
	}
	

	private void createSynthsizer(Manager m) {
		synthMgr = new SynthesisMgr(m);		 
	}

	/* 
	 * Event handlers
	 * Called by toolbar whenever a button is pressed 
	 */
	public void synthesisPressed() {
		savePath = null;
		initializeSynthesis();
		
		new Thread(new Runnable() {
			public void run() {
				Settings cfg = getCfg();
				String code = synthMgr.synthesisCode(cfg);
				codePane.display(code);
				finalizeSynthesis();
			}
		}).start();			
	}
	
	private Settings getCfg() {
		Settings cfg = new Settings();
		// cfg.blockingSend(confPane.blockingBcst());
		return cfg;
	}
	
	private void setLanguage() {
		if (codePane == null || synthMgr == null)
			return;
		synthMgr.init(confPane.getLang());
		codePane.setKWrds(synthMgr.kwords());
		codePane.setSymbs(synthMgr.symbs());
		codePane.setLang(confPane.getLang());
		toolbar.adapt(confPane.getLang());
	}
	
	private void initializeSynthesis() {
		notePane.println("Code generation started ...");
	}

	private void finalizeSynthesis() {
		notePane.println("Code generation finished.");
	}
	
	public void openPressed() {
		String addr = "";
		String s = synthMgr.suffix();
		final String p = Dialogs.showOpenDialog("Open", addr, s);
		if (p != null) {
			new Thread(new Runnable() {
				public void run() {
					String c = Util.readFile(p);
					String d = c.replaceAll("\r\n", "\n");
					codePane.display(d);
					savePath = p;
				}
			}).start();
		}
	}
	
	public int savePressed() {
		String p = savePath;
		String s = synthMgr.suffix();
		 
		if ( !Util.exists(p) ) {
			p = synthMgr.preferredSavePath();
			p = Util.getPathFromUser("Save Code", p, s);
			if (p == null)
				return -1;	// user canceled			
		}			
		
		while (codePane.isBusy()) {
			Util.sleep(100);
		}
		saveCode(p);
		savePath = p;
		return 1;
	}
	
	void saveCode(String path) {
		String code = codePane.getText();
		String c1 = code.replaceAll("\r\n", "\n"); // first unify the text
		String c2 = c1.replaceAll("\n", "\r\n");		
		synthMgr.saveCode(path, c2);		
	}

	public void checkPressed() {
		while (codePane.isBusy()) {
			Util.sleep(100);
		}
		String code = codePane.getText();
		RunResult res = synthMgr.syntaxCheck(code);
		handleRes(res);		
	}

	public void compilePressed() {
		if ( !Util.exists(savePath) ) {
			String msg = "The code is not saved.\r\nPlease save it first.";
			Dialogs.showInfoMsg(msg);			
			int saved = savePressed();
			if (saved < 0)
				return;
		} else
			saveCode(savePath);
		
		while (codePane.isBusy()) {
			Util.sleep(100);
		}		
		RunResult res = synthMgr.compile(savePath);
		handleRes(res);
	}
	
	private void handleRes(RunResult res) {
		if (res.exitValue() != 0) {
			res.addLine("Compilation terminated with a problem.");
			Util.error(res.toStr());
			Util.error(res.getLog());
		}
		if (res.succeeded() == false) {
			Exception e = res.getException();
			if (e != null) {
				Util.error(e);
				String msg = e.getMessage();
				notePane.println(msg, Color.RED);
			}
			Util.showRunErrorMsg(res.getCommand()[0]);
			Util.error(res.getLog());
		}
		notePane.printDashedLine();
		notePane.println(res.toStr(), Color.BLUE);
		notePane.printDashedLine();		
	}	

	public void cntxClausPressed() {
		if (synthMgr.hasTaskSet()) {
			String cc = synthMgr.getContextClause();
			String code = getCode(cc, "Context Clauses");
			
			if (code == null)	// user canceled
				return;
			synthMgr.setContextClause(code);
		} else
			noTaskSetError();				
	}

	public void globDeclPressed() {
		if (synthMgr.hasTaskSet()) {
			String gd = synthMgr.getGlobalDec();
			String code = getCode(gd, "Global Variables");
			if (code == null)	// user canceled
				return;
			synthMgr.setContextClause(code);
		} else
			noTaskSetError();					
	}
	
	public void mainProcPressed() {
		if (synthMgr.hasTaskSet()) {
			String mp = synthMgr.getMainProc();
			String code = getCode(mp, "Main Procedure");
			if (code == null)	// user canceled
				return;
			synthMgr.setMainProc(code);
		} else
			noTaskSetError();					
	}
	
	private String getCode(String cc, String title) {
		return GUIUtil.getCodeFromUser( cc, title, title + ":" );
	}
	
	private void noTaskSetError() {
		Dialogs.showErrorMsg("Currently there is no task set loaded.");
	}
		
	public void undoPressed() {
		codePane.undo();
	}
	
	public void redoPressed() {
		codePane.redo();
	}
	
	public void confPressed() {
		confPane.getSettingsFromUser();
		setLanguage();
	}

	private void setLayout() {
		setLayout(new BorderLayout());				
	}
}
