package synthesis.model;

import java.io.File;

import enums.Language;

import services.Executor;
import services.RunResult;
import synthesis.model.adacoder.AdaSynthesizer;
import synthesis.model.ccoder.CSynthesizer;
import util.Dialogs;

import main.Manager;
import main.TaskSet;
import main.Util;

public class SynthesisMgr {

	private	Manager			manager;
	private	AbsSynthesizer	coder;
	private String			compiler;
	
	private final	String	GNAT = "gnatmake";
	private final	String	GCC  = "gcc";
	
	public SynthesisMgr(Manager m) {
		manager = m;	
	}
	
	public void init (Language lang) {
		switch (lang) {
		case Ada:
			coder 	 = new AdaSynthesizer();
			compiler = GNAT;
			break;

		case C:
			coder = new CSynthesizer();
			compiler = GCC;
			break;
			
		default:
			Dialogs.showErrorMsg("Unimplemented programming language: " + lang);
			break;
		}	
	}

	public String synthesisCode(Settings cfg) {
		 return coder.synthesisCode(taskSet(), cfg);
	}
	
	public String preferredSavePath() {
		String n = taskSetName();
		String p = taskSetPath();
		return Util.extractPath(p) + n + suffix();	
	}
	
	public void saveCode(String path, String code) {
		Util.write2file(path, code);
	}	

	public RunResult syntaxCheck(String code) {
		String file = taskSetName() + "_temp5291.adb";	// a random name
		String[] command = new String[]{compiler, "-gnatv", "-gnats", file};				
		
		saveCode(file, code);	// save to a temporary file		
		RunResult res = runCompiler(command);
		remove(file);			// delete the temporary file
        return res;
	}
	
	public RunResult compile(String file) {
		String path = Util.extractPath(file);		
		String[] command = new String[]{compiler, "-D", path, "-o", 
				path + taskSetName(), "-gnatv", file};		
		return runCompiler(command);		
	}
	
	private RunResult runCompiler(String[] cmd) {
		Executor runner = new Executor();
		RunResult res = runner.run(cmd);		
		return res;
	}
	
	private void remove(String file) {
		File f = new File(file);
        if (f.exists()) {
            f.delete();
        }
	}
	
	public boolean hasTaskSet() {
		if (taskSet() == null)
			return false;
		return true;
	}
	
	/*
	 * Getters and Setters:
	 */

	public String suffix() {
		return coder.suffix();
	}
		
	public String[] kwords() {
		return coder.keywords();
	}
	
	public String[] symbs() {
		return coder.symbols();
	}
	
	public String getContextClause() {
		return taskSet().contextClause();
	}
	
	public void setContextClause(String code) {
		taskSet().setContextClause(code);
	}	
	
	public String getGlobalDec() {
		return taskSet().globalVars();
	}
	
	public void setGlobalDec(String code) {
		taskSet().setGlobalVars(code);
	}	
	
	public String getMainProc() {
		return taskSet().mainProcCode();
	}
	
	public void setMainProc(String code) {
		taskSet().mainProcCode(code);
	}	
	
	/*
	 * Private methods
	 */
	private TaskSet taskSet() {
		return manager.getTaskSet();
	}

	private String taskSetName() {
		return taskSet().getName();
	}
	
	private String taskSetPath() {
		return  manager.getTaskSetPath();
	}

	
}
