package taskeditor;
import javax.swing.*;

import main.Edge;
import main.JobType;
import main.Manager;
import main.Task;
import main.Util;

import taskeditor.changetrace.Change;
import taskeditor.listeners.JobIconListener;
import taskeditor.listeners.JobLabelListener;
import taskeditor.movables.Nail;
import util.Dialogs;
import util.EdgeGUIUtil;
import util.GUIUtil;

import java.awt.*;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.ListIterator;
import java.util.Vector;

/**
 This is a customized JPanel which is used for drawing the task graph
 */
public class DrawingPane extends JPanel {

    private 	EditorTab 	editorTab;
    private 	boolean 	nodeChanged = true;
    private 	DrawStatus 	drawStatus = DrawStatus.NONE; // indicates that user is in which mode of drawing (e.g. putting a new job, drawing an arc, ...) 
    private 	int			tempX = 0;
    private 	int 		tempY = 0;
    private 	JobType 	newEdgeSrcJob = null;	  // used when creating new edge
    private 	Vector<Point>newEdgeNails = null;	  // used when creating new edge
    private		Edge 		selectedEdge  = null;
    private		JobType		selectedJob   = null;
    private 	Rectangle 	selectedArea  = null;	// used to multiple object selection by mouse dragging
	
    private		EditorInfoBar infoBar;			    // shows the name of the current task set and the selected task,
    											    // as well as the cursor position
    private 	boolean 	snapped = true;
    
    public		final		int gridX = 10;
    public		final		int gridY = 10;
    
    /* 
     * constructor
     */
    public DrawingPane(Manager mngr, EditorTab 	editor){
        setLayout( null );
        this.editorTab = editor;
        
        DefaultDragMouseAdapter mouseAdapter = new DefaultDragMouseAdapter();
        addMouseListener(mouseAdapter);
        addMouseMotionListener(mouseAdapter);
        int darkC = 210;
        Color c = new Color(darkC, darkC, darkC);    	
        //setBackground(Color.LIGHT_GRAY);
        setBackground(c);
        setBorder(BorderFactory.createLineBorder(Color.gray));
    }

    public void paintComponent(Graphics g) {

        super.paintComponent(g);
        Task task = selectedTask();
        Graphics2D g2d = (Graphics2D) g;
                
        // GUIUtil.dotPane(g, this, gridX, gridY);
        GUIUtil.gridPane(g2d, this);
        if (task != null)
	        if (task.hasAnyEdges()) {
	    		ListIterator<Edge> itr = task.edgeIterator();  
				while (itr.hasNext()) {
					Edge edge = itr.next();
					drawEdge(g2d, edge);
				}
	    	}            
    		
        if (newEdgeSrcJob != null) {
        	if (selectedTask().contains(newEdgeSrcJob) == false) {
        		escapePressed();
        	} else {        	 
	        	Vector<Point> path_nails = new Vector<Point>(newEdgeNails);
	        	Point start_nail = new Point(newEdgeSrcJob.x, newEdgeSrcJob.y);
	        	Point end_nail = new Point(tempX, tempY);	        	
	        	path_nails.add(0, start_nail);
	        	path_nails.add(end_nail);
	        	Point[] nails = path_nails.toArray(new Point[]{});
	        	GUIUtil.drawSampleEdge(g, nails, true);
        	}
    	}
        
        if (selectedArea != null)
    		drawSelectionArea(g2d);
               	
        if (nodeChanged == true) {
            revalidate();
            repaint();
        	nodeChanged = false;    // to prevent unnecessary repaint
        }
        paintChildren(g);
    }

	private void drawSelectionArea(Graphics2D g2d) {
		g2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.2f));
    	g2d.setColor(Color.blue);
    	Rectangle r = selectedArea;
    	int x = Math.min(r.x, r.x + r.width);
    	int y = Math.min(r.y, r.y + r.height);
        g2d.fillRect(x, y, Math.abs(r.width), Math.abs(r.height));
        g2d.setColor(Color.black);
        g2d.setComposite(AlphaComposite.SrcOver);		
	}

	private void drawEdge(Graphics2D g2d, Edge edge) {
    	Point[] nails = edge.getAugmentedNails();
    	boolean isSeleced  = (selectedEdge == edge);
    	GUIUtil.drawSampleEdge(g2d, nails, isSeleced);
    }
    
    public void drawTaskGraph(Task t) {
    	ListIterator<Edge> 	  eItr = t.edgeIterator();
    	ListIterator<JobType> jItr = t.jobIterator();
    	removeAll();
        while (jItr.hasNext()) {
        	JobType job = jItr.next();
        	
			boolean isStartJob = t.startJobTest(job);
            job.createLabel(this, isStartJob);
            
            add(job.getIcon());
            add(job.getLabel()); 
        }
        
        // ************************ Edge labels: Inter_release time
    	while (eItr.hasNext()) {
			Edge edge = eItr.next();
			edge.updateLabel(this);
        	add(edge.getLabel()); 
        }
    	
    	infoBar.taskName(t.getName());
        nodeChanged();    // to trigger painting the edges
        repaint();
    }
    
    public void updateTaskInfo() {
		editorTab.updateTaskTable( editorTab.taskList() );		
	}
    
    public void redrawSelectedTask(){
		Task t = selectedTask();
		if (t == null)
			return;
		drawTaskGraph(t);
		nodeChanged();
    }
    
    public Task selectedTask() {
		return editorTab.getSelectedTask(); 
	}

	private void deselectEdge() {
    	if (selectedEdge != null) {
			selectedEdge.deselect(); //
			selectedEdge = null;
			nodeChanged();
		}
    }
	
	public void deselectJob() {
    	if (selectedJob != null) {
			selectedJob.deselect(); //
			selectedJob = null;
			nodeChanged();
		}		
    }
    
    public void edgeSelected(Edge edge) {
    	deselectEdge();
    	deselectJob();
    	edge.selected();
    	selectedEdge = edge;
		nodeChanged();    		
	}
    
    public void jobSelected(JobType j) {
    	deselectEdge();
    	deselectJob();
    	j.selected();
		selectedJob = j;
	}
    
    class DefaultDragMouseAdapter extends MouseAdapter implements MouseMotionListener {

    	private	Nail  draggedNail = null;
    	
    	private	static 	final 	int EDGE_MARGINE = 3;
    	private	static 	final 	int NAIL_MARGINE = 7;
    	
        public void mouseEntered(MouseEvent e) {
        	
            if (drawStatus() == DrawStatus.CIRCLE) {
            	Cursor cur = Util.getCustomCursor("/circle.png");
            	setCursor(cur);                
            }
            else if (drawStatus() == DrawStatus.ARC) {
            	setCursorCross();                
            }
            else
            	setCursorDefault();
        }

        public void mousePressed(MouseEvent e) {
        	if ( !GUIUtil.leftClick(e) )
        		return;
        	
        	deselectEdge();
        	deselectJob();
        	if ( drawStatus == DrawStatus.CIRCLE ) {
    			createNewJob(e);    			
    		} else if (drawStatus == DrawStatus.ARC) {
    			if (newEdgeSrcJob != null) {
	    			tempX = e.getX();
	            	tempY = e.getY();
	            	newEdgeNails.add(new Point(tempX, tempY));
	            	nodeChanged();
	         	}
    			// newEdgeSrcJob = null;
    		} else if ( noneDrawStatus() ) {
        		Point p = e.getPoint();
        		Nail nail = getTargetNail(p, NAIL_MARGINE);
        		Edge  edge;
        		boolean doTrack = false;
        		if (nail != null) {
        			if (draggedNail == null)
        				doTrack = true;
        			draggedNail = nail;
        			draggedNail.pressed(p);        			
        			edge = nail.myEdge();
    			} else  // if nail selected, the respective edge has been also selected, otherwise, check the edges 
        			edge = getTargetEdge(p, EDGE_MARGINE);
            
        		if (edge != null) {
            		edgeSelected(edge);
            		if (doTrack)
            			trackEdgeChange(edge);            		
            	} else {
        			selectedArea = new Rectangle (e.getX(), e.getY(), 0, 0);
        		}
            }    
	        repaint();
        }

		public void mouseExited(MouseEvent e) {
        	/*JPanel pnl = (JPanel) manager.getContentPane();
            pnl.*/setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
        }

        public void mouseDragged(MouseEvent e) {
        	updatePointInfo(e.getPoint());
        	
        	Point p = e.getPoint();
        	boolean changed = true; 
        	if (draggedNail != null) {
        		draggedNail.dragged(p, snapped, gridX, gridY);
        	} else if (selectedArea != null) {
        		selectedArea.width  = p.x - selectedArea.x;
        		selectedArea.height = p.y - selectedArea.y;        		
    		} else
        		changed = false;
        	
        	if (changed) {
        		nodeChanged();
            	repaint();
        	}
        }

        public void mouseMoved(MouseEvent e) {
        	updatePointInfo(e.getPoint());
        	
        	if (drawStatus() == DrawStatus.ARC) {
        		tempX = e.getX();
            	tempY = e.getY();
            	nodeChanged();
            	repaint();
    		} else if (noneDrawStatus() && 
            		e.getSource().getClass() == DrawingPane.class) {
            	Point p = e.getPoint();            	
            	if (isNearAnObject(p) == true)
            		setCursorHand();
    			else
    				setCursorDefault();
            }
        }
        
		public void mouseClicked(MouseEvent e) {
        	if (GUIUtil.doubleClick(e) && noneDrawStatus()) {
        		createNewJob(e);
    			repaint();
    		}
        }    

		public void mouseReleased(MouseEvent e) {
        	draggedNail = null;
        	selectedArea = null;
        	Point p = e.getPoint();
            
        	if (noneDrawStatus() && GUIUtil.rightClick(e)) {
        		JComponent c = (JComponent) e.getSource();
                Edge edge = getTargetEdge(p, EDGE_MARGINE);
                if (edge != null) {
                	edgeSelected(edge);
                	JPopupMenu menu = EdgeGUIUtil.createEdgeMenu(edge, DrawingPane.this, p);
                  	menu.show(c, p.x, p.y);                  	
            	} else
                	;// handle right click              	
            }
        	nodeChanged();
        	repaint();
        }

		private void trackEdgeChange(Edge edge) {
			trackChange(edge.getModifyChange());			
		}

        /*
         * This is called when the user double-click on the draw pane, 
         * or after selecting the new node icon, pressing the mouse on this area 
         */
		private void createNewJob(MouseEvent e) {
			Task selectedTask = selectedTask();
			if (selectedTask == null) {
    			Dialogs.showErrorMsg("Please first create a new task");
    			return;
    		}
			String[] jobProps = Util.getDefaultJobProps(selectedTask);
			
			if (jobProps != null) {
            	String name = jobProps[0];
            	String WCET = jobProps[1];
            	String DdLn = jobProps[2];
            	/*String valid = editorTab.jobPropValidity(name, WCET, DdLn, -1);
            	if(valid.compareToIgnoreCase("valid") != 0)
            	{
            		Util.showErrorMsg(valid);
            		return;
            	}*/
            	int x = e.getX();
            	int y = e.getY();
            	JobType job = new JobType(name, WCET, DdLn, x, y);             	
            	
	            // ************ Job text label: (WCET, Deadline) ************            	
            	job.createLabel(DrawingPane.this, false /* this is not the start job by default */);	            
            		
	            add(job.getIcon());
	            add(job.getLabel());	            
	            trackChange(job.getCreateChange());
	            
	            // Finally add the created job to the task
	            editorTab.addJobToCurrentTask(job);
    	        nodeChanged();    // to trigger painting the edges
			}
		}
        
        /*
         * Checks if any object (nail or edge) is currently under the cursor 
         */
        private boolean isNearAnObject(Point p) {
        	Nail nail = getTargetNail(p, NAIL_MARGINE);
        	Edge edge = getTargetEdge(p, EDGE_MARGINE); 
        	if(nail != null || edge != null)
        		return true;
        	return false;
		}
        
        /* 
         * returns the edge which is under the current position of the cursor
         * if no edge is close to the cursor, returns null
         */
        private Edge getTargetEdge(Point p, int margin) {
        	Task task = selectedTask();  // currently loaded task
        	if(task == null)
    			return null;
        	
        	return task.getTargetEdge(p, margin);    		
		}
        
        /* 
         * Returns the nail which is under the current position of the cursor
         * if no nail is close to the cursor, returns null
         */
        private Nail getTargetNail(Point p, int margin) {
        	Edge edge = getTargetEdge(p, margin);
        	if (edge == null)
    			return null;

        	for (Nail nail : edge.nails()) {
        		if (nail.position().distance(p) < margin)
        			return nail;
        	}
			return null;
		}

		private boolean noneDrawStatus() {
			if(drawStatus == DrawStatus.NONE )
				return true;
			return false;
		}       
    }

    public void setDrawingState(DrawStatus status) {
		setDrawStatus(status);
	}

	public void escapePressed() {
		setDrawStatus(DrawStatus.NONE);
		setCursorDefault();
		setNewEdgeSrcJob(null);
		newEdgeNails = null;
		nodeChanged = true;
    	repaint();		
	}
	
	public void nodeChanged() {
		nodeChanged = true;
	}

	public void setDrawStatus(DrawStatus drawStatus) {
		this.drawStatus = drawStatus;
	}

	public DrawStatus drawStatus() {
		return drawStatus;
	}

	
	/*
	 * called by pop up menu, when the user selects to delete an edge
	 */
	public void removeEdge(Edge clickedEdge) {
		deselectEdge();
		editorTab.removeEdge(clickedEdge);
		remove(clickedEdge.getLabel());
	}
	
	// called by jobIconMouseListener
	public void removeJob(JobType j) {
		editorTab.removeJob(j);
		remove(j.getIcon());
		remove(j.getLabel());
	}
	
	public void deleteKeyPressed() {
		if (selectedEdge != null) {
			int option = JOptionPane.showConfirmDialog(null, "Are you sure you want to delete the edge?", "Confirm", JOptionPane.YES_NO_OPTION);
			if (option == JOptionPane.YES_OPTION) {
				removeEdge(selectedEdge);
				redrawSelectedTask();
			}
		}
	}

	public JobType getNewEdgeSrcJob() {
		return newEdgeSrcJob;
	}

	public void setNewEdgeSrcJob(JobType newEdgeSrcJob) {
		this.newEdgeSrcJob = newEdgeSrcJob;
	}

	public void setInfoBar(EditorInfoBar infoBar) {
		this.infoBar = infoBar;  		
	}

	public void setTempX(int x) {
		this.tempX = x;
	}
	
	public void setTempY(int y) {
		this.tempY = y;
	}

	public void setNewEdgeNails(Vector<Point> nails) {
		newEdgeNails = nails;
	}
	
	public Vector<Point> newEdgeNails() {
		return newEdgeNails;
	}
	
	public void setCursorHand() {	
		setCursor(GUIUtil.HAND_CURSOR);
	}
	
	private void setCursorDefault() {	
		setCursor(GUIUtil.DFLT_CURSOR);
	}

	private void setCursorCross() {	
		setCursor(GUIUtil.CROSS_CURSOR);
	}
	
	public void updatePointInfo(Point loc) {
		infoBar.updatePointInfo(loc);		
	}

	public String jobNameValidity(String new_name, int jobNo) {
		return editorTab.jobNameValidity(new_name, jobNo);
	}

	public void emptyPane() {
		removeAll();		
	}

	public void trackChange(Change ch) {
		selectedTask().trackChange(ch);		
	}

	public void undo() {
		Task t = selectedTask();
		t.undo();
		drawTaskGraph(t);
	}

	public int lastTaskChange() {
		Task t = selectedTask();
		if(t == null)
			return -1;
		else
			return t.lastChange();
	}

	public boolean snaped() {
		return snapped;
	}

}
