package taskeditor;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.io.FileNotFoundException;
import java.util.LinkedList;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JSplitPane;
import javax.swing.JToolBar;
import javax.swing.UIManager;

import main.Edge;
import main.JobType;
import main.Manager;
import main.Task;
import main.TaskSet;
import main.Util;
import services.RunResult;
import taskeditor.changetrace.Change;
import util.Dialogs;
import util.GUIUtil;
import main.storage.TSLoader;
import main.storage.TSSaver;

public class EditorTab extends JPanel {

	private Manager 	 	manager;
	private TaskSet 	 	taskSet;
	private String 		 	currTaskSetFilePath = null;
	
	private DrawingPane  	drawPane;
	private TaskPanel	 	taskPanel;
	private JProgressBar 	progressBar;
	private RandTSGenerator randGen; 
	
	public EditorTab(Manager mgr) {
		
		manager	  = mgr; 
		drawPane  = new DrawingPane(mgr, this);
		taskPanel = new TaskPanel(this);
		randGen	  = new RandTSGenerator(this, mgr);
		
		JToolBar drawToolBar = new DrawToolBar(drawPane);
		drawToolBar.setOrientation(1);
		
		EditorInfoBar infoBar = new EditorInfoBar( drawToolBar.getPreferredSize().width );
		drawPane.setInfoBar(infoBar);
		
		JPanel panel_right = createRightPanel(drawToolBar, drawPane, infoBar); 		
		JSplitPane splitPaneH = createSplit(taskPanel, panel_right);
	
		setLayout(new BorderLayout());
		add(createToolBar(), BorderLayout.NORTH);		
		add(splitPaneH, BorderLayout.CENTER);
		
		createAndShowNewTaskSet(mgr); // create an empty new task set when the application is launched
	}

	private JSplitPane createSplit(TaskPanel tables, JPanel panel_right) {
		JSplitPane split = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
		split.setRightComponent( panel_right );
		split.setLeftComponent( tables );
		Util.decorateSplit(split, 200);
		return split;
	}

	private JPanel createRightPanel(JToolBar drawToolBar, DrawingPane drawPn, EditorInfoBar infoBar) {
		
		JPanel panel_right = new JPanel(new BorderLayout());		
		panel_right.add(drawToolBar, BorderLayout.LINE_START);
		panel_right.add(drawPn, BorderLayout.CENTER);
		panel_right.add(infoBar, BorderLayout.PAGE_END);
		return panel_right;
	}
	
	// This is the toolbar of the task editor window
	private JToolBar createToolBar() {
		progressBar = createProgressBar();
		return new ToolBar(this, progressBar);	
	}
	
	private JProgressBar createProgressBar() {
		JProgressBar progress = new JProgressBar();
		progress.setIndeterminate(true);
		progress.setVisible(false);
		Dimension d = new Dimension(150, 15);
		progress.setPreferredSize(d);
		progress.setSize(d);
		progress.setMaximumSize(d);
		UIManager.put("ProgressBar.cycleTime", new Integer(10000));
		
		return progress;
	}	

	public void showProgressBar() {
		progressBar.setVisible(true);		
	}
	
	public void hideProgressBar() {
		progressBar.setVisible(false);		
	}
	
	public void drawTaskGraph(Task task) {
		drawPane.drawTaskGraph(task);
	}
	
	public void displayTaskSet() {
		showTaskSetTable();
		drawTaskGraph(getSelectedTask());
		manager.setTiltle(taskSet.getName());
	}	
	
	public void showTaskSetTable() {
		taskPanel.showTaskSet();		
	}

	public void refreshTaskSetTable(LinkedList<Task> taskList) {
		taskPanel.refreshTaskSetTable(taskList);
	}

	public void escapePressed() {
		drawPane.escapePressed();
	}
	
	public void deleteKeyPressed() {
		drawPane.deleteKeyPressed();
	}

	public void updateTaskTable(LinkedList<Task> taskList) {
		taskPanel.updateTaskTable(taskList);		
	}

	public void emptyPane() {
		drawPane.emptyPane();
	}

	public void undo() {		
		drawPane.undo();
	}

	public void refreshTaskPanel(LinkedList<Task> taskList) {
		taskPanel.updateTaskTable(taskList);
	}

	// called by TaskPanel.handleTableDataChange() and TaskPanel.TaskSetTableModel.setValueAt()
	public void trackChange(Change ch) {
		drawPane.trackChange(ch);		
	}

	// called by task panel when the last change is related to a property in the task set level 
	public void trackGlobalChange(Change ch) {
		if (taskSet == null)	{
			Util.log("Error: tracking a change while the task set is empty:");
			Util.log(ch.toString());
			return;
		}
		
		taskSet.trackChange(ch);
	}
	
	public void getTaskVarFromUser() {
		if ( hasActiveTaskSet() ) {
			Task t       = getSelectedTask();
			String msg   = "Variable Declaration for \"" + t.getName() + "\":";
			String title = "Task Declarations";
			String code  = GUIUtil.getCodeFromUser(t.localVars(), title, msg);
			if (code == null)	// user canceled
				return;
			t.setLocalVars(code);		
			
		} else
			Dialogs.showErrorMsg("Currently there is no task set loaded.");		
	}
	
	public int lastTaskChange() {
		return drawPane.lastTaskChange();
	}

	public void loadFromXML(String xml) throws FileNotFoundException {
		taskSet = TSLoader.load(xml, "STRING");				
	}

	public String taskSetName() {
		return taskSet.getName();
	}

	public Task getSelectedTask() {
		int taskIndex = taskPanel.getSelectedTaskIndex(); 
		Task task = taskIndex == -1 ? null : taskSet.get(taskIndex);
		return task;
	}

	public Task[] getSelectedTasks() {
		String [] names = taskPanel.getSelectedTasksName();
		int n = names.length;
		Task[]tasks = new Task[n];
		for (int i = 0; i < n; i++)
			tasks[i] = taskSet.getTask(names[i]);
		return tasks;
	}
	
	public Task getTask(int index) {
		if (taskSet == null)
			return null;
		return taskSet.get(index);
	}
	
	/*
	 * Assuming task names are unique, returns the task whose name is given.
	 * This is called by TaskPanel, when the user clicks on the TaskSetTable
	 */
	public int getTaskIndex(String name) {
		if (taskSet == null)
			return -1;
		
		return taskSet.getTaskIndex(name);
	}

	public void createNewTaskSet() {
		Task.resetPriorityCounter();
		taskSet = new TaskSet(); 	// create an empty task set
		taskSet.createNewTask("Task_1"); 	// adds a default new task to the created taskset		
	}

	// called by OpenActionListner in task editor
	public void loadTaskSet(String path, Manager manager) throws FileNotFoundException {
		taskSet = TSLoader.load(path, "FILE");
		displayTaskSet();
		newTaskSetLoaded(manager);
		//manager.newTaskSetLoaded(taskSetName());
		currTaskSetFilePath = path;
	}

	public void undoDispatcher() {
		if (taskSet == null)
			return;
		
		int lastTSChange = taskSet.latestChange();
		int lastTaskChange = lastTaskChange();
		if (lastTSChange > lastTaskChange || // more recent change 
				taskSet.size() <= 0) {
			taskSet.undo();
			showTaskSetTable();
		} else 
			undo();
		
		refreshTaskPanel(taskList());
		refreshTaskSetTable(taskList());		
	}
	
	public LinkedList<Task> taskList() {
		return taskSet.taskList();
	}

	public Task[] tasks() {
		return taskSet.toArray();
	}	
	
	public void addJobToCurrentTask(JobType job) {
		getSelectedTask().addNewJob(job);		
		updateTaskTable( taskList() );        
	}

	public boolean hasActiveTaskSet() {
		if (taskSet == null)
			return false;
		if (taskList().size() == 0)
			return false;
		return true;
	}
		
	public boolean hasNonEmptyTask() {
		if(hasActiveTaskSet() == false) 
			return false;
		return taskSet.hasNonEmptyTask();
	}

	public void setPriorityAndRefresh(String[] ordered_names) {
		taskSet.assignPriority( ordered_names );
		refreshTaskSetTable( taskList() );		
	}

	public void setPriority(String taskName, int priority) {
		taskSet.setPriority(taskName, priority);
	}

	/*
	 * Called from setValueAt() in TaskPanel when user is changing the task's name 
	 */	
	public void setTaskName(String oldName, String newName) {
		taskSet.updateTaskName(oldName, newName);		
	}


	/*
	 * returns true if deletion is successful
	 */
	public boolean deleteTask(String taskName) {
		
		return taskSet.remove(taskName);		
	}
	
	// called when copying a task by TaskPanel
	public void addTask(Task t) {
		if(getTaskIndex(t.getName()) >= 0)
			System.out.println("Warning: A task with this name already exists");
		taskSet.add(t);		
	}
	
	// called by TaskPanel
	public Task createNewTask(String task_name) {
		return taskSet.createNewTask(task_name);		
	}
	
	/*
	 * This method is used to check the correctness of the job name specified by
	 * the user. Currently, job name is used as its unique id. So, a name cannot
	 * be repeated for two jobs. There are two cases: 1. A new job is being
	 * created by the user. In this case, the jobNo is equal to -1. The
	 * getJobNo() should also return -1, meaning that no job with this name
	 * already exists. 2. An already existing job is going to be modified. In
	 * this case, the name should not conform to the name of another job
	 */
	public String jobNameValidity(String name, int jobNo) {
		String wordError = Util.checkWord(name);
		if (wordError != null)
			return wordError;
		
		int found_job_no = getSelectedTask().getJobNo(name);

		if (found_job_no == -1) // ok because no job with this
								// name already exists
			return "valid";

		if (jobNo == found_job_no)
			return "valid";

		return "A job with this name already exists.\r\n Please specify another name.";
	}

	public void removeJob(JobType job) {
		Task task = getSelectedTask();
		if (task == null) {
			System.out.println("Warning: no active task found for removing the ndoe" + job.getName());
			return;
		}
		task.removeJob( job );
	}
	

	/* 
	 * called by drawingPane, when the user is going to delete an edge
	 */
	public void removeEdge(Edge victim) {
		Task task = getSelectedTask();
		if (task == null) {
			System.out.println("Warning: no active task found to remove the edge");
			return;
		}
		task.removeEdge(victim);
	}
	
	// called by both "save" and "save as" buttons
	public void save() {
		String path = currTaskSetFilePath; 
		if (path == null) { // first get the file name from user
			JFileChooser fileDialog = Util.getFileChooser("Save As",
					taskSetName() + ".xml", ".xml");				
			path = Util.showFileBrowser(fileDialog);
			if (path == null)
				return;	// user canceled
			else {
				currTaskSetFilePath = path;
				String name = Util.extractFileName(path);
				taskSet.setName(name);
				manager.setTiltle(name);
			}
		}
		save(currTaskSetFilePath); 
	}

	public void save(String filePath) {
		TSSaver.save(tasks(), getTaskSet(), filePath);
	}

	/*
	 * This function is called in two cases: 1. When the application starts 
	 * 2. When the user is going to create a new task set
	 */
	public void createAndShowNewTaskSet(Manager manager) {
		createNewTaskSet();		
		displayTaskSet();
		newTaskSetLoaded(manager);
		//manager.newTaskSetLoaded(taskSetName());
		currTaskSetFilePath = null;
	}
	
	public void newTaskSetLoaded(Manager mgr) {
		mgr.newTaskSetLoaded(taskSetName());
		taskPanel.selectFirstTask();
	}

	public void setTaskSetPath(String path) {
		currTaskSetFilePath = path;		
	}
	
	public String getTaskSetPath() {
		return currTaskSetFilePath;		
	}

	public String globalVars() {
		return taskSet.globalVars();
	}
	
	public String contextClause() {
		return taskSet.contextClause();
	}

	public TaskSet getTaskSet() {
		return taskSet;
	}

	 /***********************************\
	/	Definitions of ToolBar's Commands \
	\									  /
	 \***********************************/
	public void newPressed() {
		createAndShowNewTaskSet(manager);
	}

	public void openPressed() {
		String addr = getTaskSetPath();
		if (addr == null)
			addr = ".";

		String path = Dialogs.showOpenDialog("Open", addr, ".xml");
		if (path != null) {
			try {
				loadTaskSet(path, manager);
			} catch (Exception e) {				
				e.printStackTrace();
				int result = Util.showOpenErrDialog(manager.frame(), e);			
				if (result == JOptionPane.YES_OPTION)
					openPressed();
			}
		}
	}

	public void saveAs() {
		currTaskSetFilePath = null;
		save();
	}

	public void randomButtPressed() {
		randGen.run();		
	}

	public void configPressed() {
		randGen.getUserConfig();		
	}
}


/*
 * This is for connecting to drtlib for random task set generation
 */
class RandTSGenerator {
	Config		config;
	EditorTab	editorTab;
	Manager 	manager;
		
	public RandTSGenerator(EditorTab et, Manager m) {
		editorTab = et;
		manager   = m;
		config	  = new Config();		
	}
	
	public void run() {
		editorTab.showProgressBar();

		new Thread(new Runnable() {
			public void run() {
				try {
					runRandomGenerator();
				} catch (Exception e) {
					editorTab.hideProgressBar();
					Util.error(e);
					Dialogs.showErrorMsg("Random task set generation unsuccessful." +
							"\r\nSee the log file for details.");					
				}
				editorTab.hideProgressBar();
			}
		}).start();
	}

	private void runRandomGenerator() throws Exception {
		String type = getTaskType();
		String crit = getTaskSetCriterion(); // either U (for utilization based) or
												// N (for number of tasks)
		String Ubound = "" + config.utilization;	
		String TNum   = "" + config.numOfTasks;	
		String TSParam = getTSParam();
		
		editorTab.setTaskSetPath(null);
		RunResult res = manager.createRandomTaskSet(type, crit, Ubound, TNum, TSParam);
		if (res.succeeded() == false || res.exitValue() != 0)
			return;	
		
		String xml = res.getLine(0);
		/*** Creating task set according to the obtained XML ***/
		try {
			editorTab.loadFromXML(xml);
			editorTab.displayTaskSet();
			editorTab.newTaskSetLoaded(manager);
		}
		catch(Exception e) {
			throw e;
		}		
	}

	// S: small, M: Medium, ...
	private String getTaskType() {
		String type = config.getTaskType();
		if (type.equalsIgnoreCase("P"))
			type = "C";	// periodic is handled as a special case of customized task set
		return type;
	}

	private String getTaskSetCriterion() {
		return config.getTaskCriterion();
	}

	private String getTSParam() {
		return config.getParamStr();
	}
	
	public void getUserConfig() {
		ConfigPanel p = new ConfigPanel(config, this, editorTab);
		p.appear();
	}

	public void setCfg(Config conf) {
		config = conf;		
	}
}