package util;

import java.util.LinkedList;
import java.util.ListIterator;

import main.Edge;
import main.JobType;
import main.Task;
import main.Util;

public class SyntaxChecker {

	public static LinkedList<String> checkForAnalysis(Task[] tasks) {
		LinkedList<String> errors = new LinkedList<String>();		
		String[] names = Util.extractNames(tasks);
		String[] redund = Util.getRedundants(names);
		if (redund.length > 0)
			for (String s : redund)
				errors.add("Error: Tasks with redundant name: " + s);
		
		for (Task t : tasks)
			errors.addAll(checkForAnalysis(t));
		
		return errors;
	}
	

	private static LinkedList<String> checkForAnalysis(Task t) {
		LinkedList<String> errors = new LinkedList<String>(); 
		ListIterator<JobType> itr = t.jobIterator();
		ListIterator<Edge> eItr = t.edgeIterator();
		while (itr.hasNext()) {
			JobType j = itr.next();
			if (j.WCET < 0)
				errors.add(errPrefix(t, j) + ": Negative job execution time: " + j.WCET + ".");
			if (j.deadline < 0)
				errors.add(errPrefix(t, j) + ": Negative job deadline: " + j.deadline + ".");
			if (j.WCET > j.deadline)
				errors.add(errPrefix(t, j) + ": WCET (i.e., " + j.WCET + ") is larger than the deadline (i.e., " + j.deadline + ").");
			String s = checkNameSyntax(j.getName());
			if (s != null)  // i.e., there is an error
				errors.add(errPrefix(t, j) + ": Error in job name," + s);
		}
		while (eItr.hasNext()) {
			Edge e = eItr.next();
			int d = e.source().deadline;
			int r = e.interRel(); 
			if (r <= 0)
				errors.add(errPrefix(t, e) + ": Non-positive inter-release time: " + r + ".");
			if (d > r)
				errors.add(errPrefix(t, e) + ": Inter-release time (" + r + ") is smaller than the deadline of the source job (" + d + ").");			
		}

		eItr = t.edgeIterator();
		Edge[]redundants = duplicateEdges(eItr);
		for (Edge e : redundants)
			errors.add(errPrefix(t) + ": Duplicate edge: " + e.toString() + ".");

		// check for redundant job name:
		String[] names = Util.extractJobsName(t);
		String[] redund = Util.getRedundants(names);
		if (redund.length > 0)
			for (String s : redund)
				errors.add(errPrefix(t) + ": Jobs with redundant name: " + s);

		return errors;
	}

	// returns null if it is a valid name
	public static String checkNameSyntax(String name) {
		return Util.checkWord(name);
	}


	private static Edge[] duplicateEdges(ListIterator<Edge> eItr) {
		LinkedList<Edge> checked = new LinkedList<Edge>();
		LinkedList<Edge> duplicates = new LinkedList<Edge>();
		while (eItr.hasNext()) {
			Edge e = eItr.next();
			for (Edge p : checked)
				if (e.isParallel(p)) {
					duplicates.add(e);
					break;
				}
			checked.add(e);
		}
		return duplicates.toArray(new Edge[]{});
	}

	public static LinkedList<String> checkForSimulation(Task[] tasks) {
		LinkedList<String> errors = new LinkedList<String>();		
		for (Task t : tasks) {
			ListIterator<Edge> itr = t.edgeIterator();
			while (itr.hasNext()) {
				Edge e = itr.next();
				int d = e.source().deadline;
				int r = e.interRel(); 
				if (d > r)
					errors.add(errPrefix(t, e) + ": Inter-release time (" + r + 
							") is less than the deadline of the source job (" + d + ").");
			}
		}
		return errors;
	}

	public static String errPrefix(Task t) {
		return "Erorr in task \"" + t.getName() + "\"";
	}

	public static String errPrefix(Task task, Edge e) {
		return errPrefix(task) + ", edge \"" + e.toString() + "\"";
	}

	public static String errPrefix(Task task, JobType j) {
		return errPrefix(task) + ", job \"" + j.getName() + "\"";
	}
}
